// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "actgen_direct_pass.h"
#include "world_model_ext.h"
#include "simple_pass_checker.h"
#include <rcsc/math_util.h>
#include <rcsc/common/logger.h>
#include <vector>

//#define DEBUG_PRINT

//static const double SAME_PASSER_POS_THRESHOLD = 0.1;
static const double SAME_PASSER_POS_THRESHOLD = 10.0;

#ifdef DEBUG_PRINT
#include "recursive_action_chain_graph.h"
#endif

ActGen_DirectPass::ActGen_DirectPass()
{
}

ActGen_DirectPass::~ActGen_DirectPass()
{
}

static
double
s_get_ball_speed_for_pass( const double & distance );

void
ActGen_DirectPass::addCandidates( std::vector< rcsc::ActionStatePair > * result,
                                  const rcsc::PredictState & state,
                                  const rcsc::WorldModel & current_wm,
                                  const std::vector< rcsc::ActionStatePair > & path ) const
{
    static const int VALID_PLAYER_THRESHOLD = 10;


    //
    // check same situation in history
    //
    std::vector< bool > old_holder_table( rcsc::ServerParam::i().maxPlayer(), false );

#ifdef DEBUG_PRINT
    rcsc::RecursiveActionChainGraph::writeChainInfoLog
        ( "branch cut checking: history = ",
          current_wm, path, 0.0 );
#endif

    for ( int c = static_cast<int>( path.size() ) - 1; c >= 0; --c )
    {
        int ball_holder_unum;

        if ( c == 0 )
        {
            const rcsc::AbstractPlayerObject * h
                = current_wm.getTeammateNearestToBall
                  ( rcsc::PredictState::VALID_PLAYER_THRESHOLD );

            if ( h && ( h->pos() - current_wm.ball().pos() ).r()
                 < ( current_wm.self().pos() - current_wm.ball().pos() ).r() )
            {
                ball_holder_unum = h->unum();
            }
            else
            {
                ball_holder_unum = current_wm.self().unum();
            }
        }
        else
        {
            ball_holder_unum = path[c - 1].state().ballHolder()->unum();
        }

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            " state %d, ball holder = %d",
                            c, ball_holder_unum );
#endif

        if ( ball_holder_unum == rcsc::Unum_Unknown )
        {
#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                " unknown number ball holder" );
#endif
            continue;
        }

        const rcsc::CooperativeAction & action = path[c].action();

        if ( ! action.supportPlayerMoves().empty() )
        {
#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                " has support player moves" );
#endif
            break;
        }

        if ( action.type() != rcsc::CooperativeAction::ActionType::Pass )
        {
#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                " not pass" );
#endif
            break;
        }

        rcsc::Vector2D pass_target_pos;
        if ( c == 0 )
        {
            pass_target_pos = current_wm.teammate( action.targetPlayerUnum() )->pos();
        }
        else
        {
            pass_target_pos = path[c - 1].state().teammate( action.targetPlayerUnum() )->pos();
        }

        const double receive_point_diff = ( action.targetPoint() - pass_target_pos ).r();
        const bool is_direct_pass = ( receive_point_diff < SAME_PASSER_POS_THRESHOLD );

        if ( ! is_direct_pass )
        {
#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                " not direct pass: receive point diff = %f",
                                receive_point_diff );
#endif
            break;
        }

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            " same situation found in history, holder = %d",
                            ball_holder_unum );
#endif
        old_holder_table[ ball_holder_unum - 1 ] = true;
    }

#ifdef DEBUG_PRINT
    for ( int i = 1; i <= rcsc::ServerParam::i().maxPlayer(); ++i )
    {
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            " num = %d, ignore? = %s",
                            i,
                            ( old_holder_table[ i - 1 ] ? "true" : "false" ) );
    }
#endif



    //
    // add pass candidates
    //
    const rcsc::AbstractPlayerObject * holder = state.ballHolder();
    const rcsc::SimplePassChecker pass_check;

    const rcsc::PredictState::PlayerPtrCont::const_iterator end = state.allTeammates().end();
    for ( rcsc::PredictState::PlayerPtrCont::const_iterator it = state.allTeammates().begin();
          it != end;
          ++it )
    {
#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "checking to %d", (*it)->unum() );
#endif

        if ( (*it)->unum() != rcsc::Unum_Unknown
             && old_holder_table[ (*it)->unum() - 1 ] )
        {
#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                "direct: old holder %d",
                                (*it)->unum() );
#endif
        }

        if ( (*it)->unum() == state.ballHolderUnum() )
        {
#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                "direct: ignored direct pass to self" );
#endif
            continue;
        }


        if ( ! (*it)->valid()
             || (*it)->posCount() > VALID_PLAYER_THRESHOLD
             || (*it)->isGhost()
             || (*it)->unum() == rcsc::Unum_Unknown
             || (*it)->unumCount() > VALID_PLAYER_THRESHOLD
             || (*it)->isTackling() )
        {
#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                "direct: can't pass from %d to %d,"
                                " target accuracy low",
                                holder->unum(), (*it)->unum() );
#endif
            continue;
        }


        //
        // check direct pass
        //
        const double dist = ( (*it)->pos() - holder->pos() ).r();

        const double ball_speed = s_get_ball_speed_for_pass
                                  ( ( (*it)->pos() - holder->pos() ).r() );

        if ( ! pass_check( state, holder, &(**it), (**it).pos(), ball_speed ) )
        {
#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                "direct: can't pass from %d to %d",
                                holder->unum(), (*it)->unum() );
#endif
            continue;
        }

        const unsigned long kick_step = 2;

        const unsigned long spend_time = rcsc::calc_length_geom_series
                                         ( ball_speed,
                                           dist,
                                           rcsc::ServerParam::i().ballDecay() )
                                         + kick_step;

        result->push_back( rcsc::ActionStatePair
                           ( new rcsc::CooperativeAction
                             ( rcsc::CooperativeAction::ActionType::Pass,
                               (*it)->pos(),
                               (*it)->unum(),
                               s_get_ball_speed_for_pass( dist ),
                               0,
                               spend_time,
                               false,
                               "direct" ),
                             new rcsc::PredictState( state,
                                                     spend_time,
                                                     (*it)->unum(), (*it)->pos() ) ) );
    }
}


static
double
s_get_ball_speed_for_pass( const double & distance )
{
    if ( distance >= 20.0 )
    {
        return 3.0;
    }
    else if ( distance >= 8.0 )
    {
        return 2.5;
    }
    else if ( distance >= 5.0 )
    {
        return 1.8;
    }
    else
    {
        return 1.5;
    }
}
