// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA, Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "actgen_through_pass2008.h"

#include <rcsc/common/logger.h>
#include <rcsc/common/audio_memory.h>
#include <rcsc/math_util.h>
#include <utility>

// #define DEBUG_PRINT_SUCCESS_PASS
// #define DEBUG_PRINT_FAILED_PASS

// #define DEBUG_PRINT_SCORE_COLOR

// #define DEBUG_PRINT_THROUGH_PASS_LEVEL_1
// #define DEBUG_PRINT_THROUGH_PASS_LEVEL_2
// #define DEBUG_PRINT_THROUGH_PASS_LEVEL_3

using namespace rcsc;

ActGen_ThroughPass2008::ActGen_ThroughPass2008()
{

}

ActGen_ThroughPass2008::~ActGen_ThroughPass2008()
{

}

static
double
s_distFromHolder( const AbstractPlayerObject * player,
                  const AbstractPlayerObject * holder )
{
    return ( player->pos() - holder->pos() ).r();
}

void
ActGen_ThroughPass2008::addCandidates( std::vector< ActionStatePair > * result,
                                       const PredictState & state,
                                       const WorldModel &,
                                       const std::vector< ActionStatePair > & path ) const
{
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_1
    dlog.addText( Logger::ACTION,
                  __FILE__": ActGen_ThroughPass2008::addCandidates()" );
#endif

    // loop candidate teammates
    const PredictState::PlayerPtrCont::const_iterator t_end = state.allTeammates().end();
    for ( PredictState::PlayerPtrCont::const_iterator it = state.allTeammates().begin();
          it != t_end;
          ++it )
    {
        if ( ! (*it)->valid()
             || (*it)->unum() == state.ballHolderUnum()
             || (*it)->unum() == rcsc::Unum_Unknown )
        {
            continue;
        }

        if ( (*it)->goalie() && (*it)->pos().x < -22.0 )
        {
            // goalie is rejected.
            continue;
        }
        if ( (*it)->posCount() > 3 )
        {
            // low confidence players are rejected.
            continue;
        }
        if ( (*it)->isTackling() )
        {
            // player is freezed
            continue;
        }

        if ( (*it)->pos().x > state.offsideLineX() )
        {
            // offside players are rejected.
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_2
            dlog.addText( Logger::PASS,
                          __FILE__" receiver %d is offside",
                          (*it)->unum() );
#endif
            continue;
        }
        if ( (*it)->pos().x < state.ball().pos().x - 25.0 )
        {
            // too back
            continue;
        }

        // create & verify each route
        if ( state.ballHolder()->pos().x > state.offsideLineX() - 20.0 )
        {
            create_through_pass( result, state, path, &(**it) );
        }
    }
}


/*-------------------------------------------------------------------*/
/*!
  static method
*/
void
ActGen_ThroughPass2008::create_through_pass( std::vector< ActionStatePair > * result,
                                             const PredictState & world,
                                             const std::vector< ActionStatePair > & path,
                                             const AbstractPlayerObject * receiver )
{
    static const double MAX_THROUGH_PASS_DIST
        = 0.9 * inertia_final_distance( ServerParam::i().ballSpeedMax(),
                                        ServerParam::i().ballDecay() );

    ////////////////////////////////////////////////////////////////
    static const
        Rect2D shrinked_pitch( Vector2D( -ServerParam::i().pitchHalfLength() + 5.0,
                                         -ServerParam::i().pitchHalfWidth() + 5.0 ),
                               Size2D( ServerParam::i().pitchLength() - 10.0,
                                       ServerParam::i().pitchWidth() - 10.0 ) );

    //static const double receiver_dash_speed = 1.0;
    //static const double receiver_dash_speed = 0.85;

    static const double min_first_speed = 1.5; // 1.8 (2009-06-09 akiyama)

    static const double S_min_dash = 5.5; //8.0;
    static const double S_max_dash = 25.0;
    static const double S_dash_range = S_max_dash - S_min_dash;
    const double S_dash_inc = ( path.size() >= 1 ? 5.0 : 2.5 );
    static const int S_dash_loop
        = static_cast< int >( std::ceil( S_dash_range / S_dash_inc ) ) + 1;

    const AngleDeg S_min_angle = ( path.size() >= 1 ? -24.0 : -32.0 );
    const AngleDeg S_max_angle = ( path.size() >= 1 ? +24.0 : +32.0 );
    static const double S_angle_range = ( S_min_angle - S_max_angle ).abs();
    static const double S_angle_inc = 8.0;
    static const int S_angle_loop
        = static_cast< int >( std::ceil( S_angle_range / S_angle_inc ) ) + 1;

    const AbstractPlayerObject * holder = world.ballHolder();

#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_1
    dlog.addText( Logger::PASS,
                  "Create_through_pass() to %d(%.1f %.1f)",
                  receiver->unum(),
                  receiver->pos().x, receiver->pos().y );
#endif

    if ( ( receiver->pos() - holder->pos() ).th().abs() > 135.0 )
    {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_2
        dlog.addText( Logger::PASS,
                      "__ receiver angle is too back" );
#endif
        return;
    }

    const PlayerType * player_type = world.teammatePlayerType( receiver->unum() );

    Vector2D receiver_pos = receiver->pos();
    if ( receiver->velCount() < 3 )
    {
        Vector2D fvel = receiver->vel();
        fvel /= player_type->playerDecay();
        fvel *= std::min( receiver->velCount() + 1, 2 );
        receiver_pos += fvel;
    }

#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_2
    dlog.addText( Logger::PASS,
                  "__ receiver predict pos(%.2f %.2f)",
                  receiver_pos.x, receiver_pos.y );
#endif

    if ( receiver_pos.x < holder->pos().x - 10.0 )
    {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_2
        dlog.addText( Logger::PASS,
                      "__ receiver is back" );
#endif
        return;
    }
    if ( std::fabs( receiver_pos.y - holder->pos().y ) > 35.0 )
    {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_2
        dlog.addText( Logger::PASS,
                      "__ receiver Y diff is big" );
#endif
        return;
    }
    if ( world.defenseLineX() < 0.0
         && receiver_pos.x < world.defenseLineX() - 15.0 )
    {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_2
        dlog.addText( Logger::PASS,
                      "__ receiver is near to defense line" );
#endif
        return;
    }
    if ( world.offsideLineX() < 30.0
         && receiver_pos.x < world.offsideLineX() - 15.0 )
    {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_2
        dlog.addText( Logger::PASS,
                      "__ receiver is far from offside line" );
#endif
        return;
    }

    const int seen_opponent_count = world.opponents().size();

    const bool pass_request = ( world.audioMemory().passRequestTime().cycle()
                                > world.time().cycle() - 2
                                && ! world.audioMemory().passRequest().empty()
                                && world.audioMemory().passRequest().front().sender_
                                == receiver->unum() );

    // angle loop
    AngleDeg dash_angle = S_min_angle;
    for ( int i = 0; i < S_angle_loop; ++i, dash_angle += S_angle_inc )
    {
        const Vector2D base_dash = Vector2D::polar2vector( 1.0, dash_angle );

        // dash dist loop
        double dash_dist = S_min_dash;
        for ( int j = 0; j < S_dash_loop; ++j, dash_dist += S_dash_inc )
        {
            Vector2D target_point = base_dash;
            target_point *= dash_dist;
            target_point += receiver->pos();

            if ( ! shrinked_pitch.contains( target_point ) )
            {
                // out of pitch
                continue;
            }
#if 0
            if ( target_point.x < 35.0
                 && target_point.x < receiver->pos().x + 10.0 )
            {
                // no good run
                continue;
            }
#endif
            if ( target_point.x < holder->pos().x + 3.0
                 && target_point.x < 40.0 )
            {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_2
                dlog.addText( Logger::PASS,
                              "cc__ through unum=%d pos=(%.1f %.1f) < holder.x + 3",
                              receiver->unum(),
                              target_point.x, target_point.y );
#endif
                continue;
            }

            const Vector2D target_rel = target_point - world.ball().pos();
            const double target_dist = target_rel.r();
            const AngleDeg target_angle = target_rel.th();

            if ( seen_opponent_count < 11 )
            {
                // check dir confidence
                int max_count = 100, ave_count = 100;
                world.dirRangeCount( target_angle, 20.0,
                                     &max_count, NULL, &ave_count );
                //if ( max_count > 9 || ave_count > 3 )
                if ( max_count > 9 || ave_count > 4 )
                {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_2
                    dlog.addText( Logger::PASS,
                                  "cc__ through unum=%d pos=(%.1f %.1f)  low dir accuracy max=%d ave=%d",
                                  receiver->unum(),
                                  target_point.x, target_point.y,
                                  max_count, ave_count );
#endif
                    continue;
                }
            }

            if ( target_dist > MAX_THROUGH_PASS_DIST ) // dist range over
            {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_2
                dlog.addText( Logger::PASS,
                              "cc__ through unum=%d pos=(%.1f %.1f) over max dist %.1f",
                              receiver->unum(),
                              target_point.x, target_point.y,
                              target_dist );
#endif
                continue;
            }

            if ( target_dist < dash_dist ) // I am closer than receiver
            {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_2
                dlog.addText( Logger::PASS,
                              "cc__ through unum=%d pos=(%.1f %.1f) target_dist=%.1f < dash_dist=%.1f",
                              receiver->unum(),
                              target_point.x, target_point.y,
                              target_dist, dash_dist );
#endif
                continue;
            }

            const double dash_step = player_type->cyclesToReachDistance( dash_dist );
            double dash_step_buf = 0.0;

            if ( pass_request )
            {
                //dash_step_buf = -0.5 - 0.01 * dash_dist;
                dash_step_buf = -0.5;
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_1
                dlog.addText( Logger::PASS,
                              "__ through unum=%d pos=(%.1f %.1f) dash_step_buf=%.2f pass_request",
                              receiver->unum(),
                              target_point.x, target_point.y,
                              dash_step_buf );
#endif
            }
            else if ( receiver->velCount() == 0
                      && ( receiver->vel().th() - dash_angle ).abs() < 15.0
                      && receiver->vel().r() > 0.2 )
            {
                //dash_step_buf = -0.5 - 0.01 * dash_dist;
                //dash_step_buf = -0.5 - 0.005 * dash_dist;
                //dash_step_buf = -0.5 - 0.003 * dash_dist; // 2008-07-18
                dash_step_buf = -0.5 - 0.0035 * dash_dist; // 2008-07-18
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_1
                dlog.addText( Logger::PASS,
                              "__ through unum=%d pos=(%.1f %.1f) dash_step_buf=%.2f running",
                              receiver->unum(),
                              target_point.x, target_point.y,
                              dash_step_buf );
#endif
            }
            else if ( receiver->bodyCount() <= 1
                      && ( receiver->body() - dash_angle ).abs() < 15.0 )
            {
                //dash_step_buf = -0.01 * dash_dist;
                //dash_step_buf = -0.005 * dash_dist;
                //dash_step_buf = -0.003 * dash_dist; // 2008-07-18
                //dash_step_buf = -0.0035 * dash_dist; // 2008-07-18
                dash_step_buf = 0.5 - 0.0035 * dash_dist; // 2009-05-09
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_1
                dlog.addText( Logger::PASS,
                              "__ through unum=%d pos=(%.1f %.1f) dash_step_buf=%.2f body_facing",
                              receiver->unum(),
                              target_point.x, target_point.y,
                              dash_step_buf );
#endif
            }
            else if ( dash_dist > 15.0 )
            {
                //dash_step_buf = 0.5 - 0.01 * dash_dist;
                dash_step_buf = 0.5 - 0.005 * dash_dist; // 2008-07-18
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_1
                dlog.addText( Logger::PASS,
                              "____ through unum=%d pos=(%.1f %.1f) dash_step_buf=%.2f long_run",
                              receiver->unum(),
                              target_point.x, target_point.y,
                              dash_step_buf );
#endif
            }
            else
            {
                dash_step_buf = 3.0;
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_1
                dlog.addText( Logger::PASS,
                              "__ through unum=%d pos=(%.1f %.1f) dash_step_buf=%.2f default",
                              receiver->unum(),
                              target_point.x, target_point.y,
                              dash_step_buf );
#endif
            }

            dash_step_buf += bound( 0, receiver->posCount() - 1, 3 );

            //double end_speed = 0.81;//0.65
            double end_speed = 1.75; //1.5; //2.0;
            double first_speed = 100.0;
            double ball_steps_to_target = 100.0;

            const double max_ball_speed = ( world.gameMode().type() == GameMode::PlayOn
                                            ? ServerParam::i().ballSpeedMax()
                                            : holder->kickRate() * ServerParam::i().maxPower() );

            do
            {
                first_speed
                    = calc_first_term_geom_series_last
                    ( end_speed,
                      target_dist,
                      ServerParam::i().ballDecay() );
                if ( first_speed > max_ball_speed + 0.01 )
                {
                    end_speed -= 0.1;
                    continue;
                }

                if ( first_speed < min_first_speed ) break;

                ball_steps_to_target
                    = calc_length_geom_series( first_speed,
                                               target_dist,
                                               ServerParam::i().ballDecay() );
                if ( dash_step + dash_step_buf < ball_steps_to_target )
                {
                    break;
                }

                end_speed -= 0.075;
            }
            while ( end_speed > 0.1 );

            if ( first_speed > max_ball_speed + 0.01
                 || first_speed < min_first_speed
                 || dash_step + dash_step_buf > ball_steps_to_target )
            {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_1
                dlog.addText( Logger::PASS,
                              "cc__ through unum=%d pos=(%.1f %.1f) first_speed=%.1f"
                              " end_speed=%.1f dash_step=%.1f ball_step=%.1f",
                              receiver->unum(),
                              target_point.x, target_point.y,
                              first_speed,
                              end_speed,
                              dash_step,
                              ball_steps_to_target );
#endif
                continue;
            }

            if ( verify_through_pass( world,
                                      receiver,
                                      receiver->pos(),
                                      target_point, target_dist, target_angle,
                                      first_speed,
                                      ball_steps_to_target,
                                      true ) )
            {
                result->push_back( ActionStatePair
                                   ( new CooperativeAction
                                     ( CooperativeAction::ActionType::Pass,
                                       target_point,
                                       receiver->unum(),
                                       first_speed,
                                       dash_step,
                                       static_cast<int>( ball_steps_to_target ),
                                       ( target_point.x < 30.0 ),
                                       "through2008" ),
                                     new PredictState
                                     ( world,
                                       ball_steps_to_target,
                                       receiver->unum(), target_point ) ) );
#ifdef DEBUG_PRINT_SUCCESS_PASS
                dlog.addText( Logger::PASS,
                              "ok__%d/%d through unum=%d pos=(%.1f %.1f) angle=%.1f first_speed=%.1f end_speed=%.1f dash_step=%.1f ball_step=%.1f",
                              0, 0, //g_success_pass_counter, g_pass_counter,
                              receiver->unum(),
                              target_point.x, target_point.y,
                              target_angle.degree(),
                              first_speed,
                              end_speed,
                              dash_step,
                              ball_steps_to_target );
#ifndef DEBUG_PRINT_SCORE_COLOR
                dlog.addCircle( Logger::PASS,
                                target_point, 0.2,
                                "#FF0000" );
//                 char num[8];
//                 snprintf( num, 8, "%d", g_pass_counter );
//                 dlog.addMessage( Logger::PASS,
//                                  target_point, num );
#endif
#endif
            }
#ifdef DEBUG_PRINT_FAILED_PASS
            else
            {
                dlog.addText( Logger::PASS,
                              "xx__%d through unum=%d pos=(%.1f %.1f) angle=%.1f first_speed=%.1f end_speed=%.1f dash_step=%.1f ball_step=%.1f",
                              0, //g_pass_counter,
                              receiver->unum(),
                              target_point.x, target_point.y,
                              target_angle.degree(),
                              first_speed,
                              end_speed,
                              dash_step,
                              ball_steps_to_target );
                dlog.addRect( Logger::PASS,
                              target_point.x - 0.2, target_point.y - 0.2, 0.4, 0.4,
                              "#0000FF" );
                // char num[8];
//                 snprintf( num, 8, "%d", g_pass_counter );
//                 dlog.addMessage( Logger::PASS,
//                                  target_point, num );
            }
#endif
        }
    }
}

/*-------------------------------------------------------------------*/
/*!
  static method
*/
bool
ActGen_ThroughPass2008::verify_through_pass( const PredictState & world,
                                             const AbstractPlayerObject * receiver,
                                             const Vector2D & receiver_pos,
                                             const Vector2D & target_point,
                                             const double & target_dist,
                                             const AngleDeg & target_angle,
                                             const double & first_speed,
                                             const double & /*reach_step*/,
                                             const bool is_through_pass )
{
    const AbstractPlayerObject * holder = world.ballHolder();

    static const double player_dash_speed = ( is_through_pass ? 0.9 : 1.0 );

    const Vector2D first_vel = Vector2D::polar2vector( first_speed, target_angle );
    const AngleDeg minus_target_angle = -target_angle;
    const double next_speed = first_speed * ServerParam::i().ballDecay();

    const double receiver_to_target = receiver_pos.dist( target_point );

    bool very_aggressive = false;
    if ( is_through_pass )
    {
        if ( target_point.x > 28.0
             && target_point.x > world.ball().pos().x + 20.0 )
        {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
            dlog.addText( Logger::PASS,
                          "____ aggressive(1)." );
#endif
            very_aggressive = true;
        }
        else if ( target_point.x > 28.0
                  && target_point.x > world.offsideLineX() + 15.0
                  && target_point.x > world.ball().pos().x + 15.0 )
        {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
            dlog.addText( Logger::PASS,
                          "____ aggressive(2)." );
#endif
            very_aggressive = true;
        }
        else if ( target_point.x > 38.0
                  && target_point.x > world.offsideLineX()
                  && target_point.absY() < 14.0 )
        {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
            dlog.addText( Logger::PASS,
                          "____ aggressive(3)." );
#endif
            very_aggressive = true;
        }
        else if ( target_point.x > receiver->pos().x + 10.0
                  && receiver->pos().x > world.offsideLineX() - 2.5 )
        {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
            dlog.addText( Logger::PASS,
                          "____ aggressive(4)." );
#endif
            very_aggressive = true;
        }

#if 1
        if ( ( target_point - receiver_pos ).th().abs() > 16.0 )
        {
            very_aggressive = false;
        }
#endif
    }

    bool pass_request = false;
    if ( is_through_pass
         && target_point.x > receiver->pos().x + 8.0 )
    {
        if ( world.audioMemory().passRequestTime().cycle() >= world.time().cycle() - 1 )
        {
            for ( std::vector< AudioMemory::PassRequest >::const_iterator it = world.audioMemory().passRequest().begin();
                  it != world.audioMemory().passRequest().end();
                  ++it )
            {
                if ( it->sender_ == receiver->unum() )
                {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
                    dlog.addText( Logger::PASS,
                                  "____ pass request. time(%ld) sender=%d pos=(%.1f %.1f)",
                                  world.audioMemory().passRequestTime().cycle(),
                                  it->sender_,
                                  it->pos_.x, it->pos_.y );
#endif
                    pass_request = true;
                    very_aggressive = true;
                    break;
                }
            }
        }
    }

    const PlayerPtrCont::const_iterator o_end = world.opponentsFromSelf().end();
    for ( PlayerPtrCont::const_iterator it = world.opponentsFromSelf().begin();
          it != o_end;
          ++it )
    {
        if ( (*it)->posCount() > 20 ) continue;
        if ( (*it)->isGhost() && (*it)->posCount() > 10 ) continue;
        if ( (*it)->isTackling() ) continue;

        if ( ( ( (*it)->pos() - holder->pos() ).th() - target_angle ).abs() > 100.0 )
        {
            continue;
        }

        const PlayerType * ptype = (*it)->playerTypePtr();

        const double virtual_dash
            = player_dash_speed * bound( 0.0, (*it)->posCount() - 0.5, 2.0 );
        const double opp_to_target = (*it)->pos().dist( target_point );
        double dist_rate = 1.0;
        //double dist_rate = ( very_aggressive ? 0.8 : 1.0 );
        double dist_buf = ( is_through_pass
                            ? ( very_aggressive ? 2.5 : 1.0 )
                            : -2.0 );
        if ( pass_request ) dist_buf += 2.0;

        if ( (*it)->pos().x > target_point.x - 3.0 )
        {
            //dist_rate = 1.08;
            dist_rate = 1.05;
            dist_buf -= 2.0;
        }

        if ( (*it)->goalie()
             && (*it)->pos().x > target_point.x - 3.0 )
        {
            //dist_rate = 1.025;
            //dist_buf -= 1.5;
            dist_buf -= 0.5;
            //if ( is_through_pass ) dist_buf -= 0.5;
            //if ( very_aggressive ) dist_buf -= 0.5;
        }

#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
        dlog.addText( Logger::PASS,
                      "______ opp%d(%.1f %.1f) dist_rate=%.2f, dist_buf=%.2f"
                      " opp_dist=%.2f receiver_dist=%.2f",
                      (*it)->unum(),
                      (*it)->pos().x, (*it)->pos().y,
                      dist_rate, dist_buf,
                      opp_to_target - virtual_dash,
                      receiver_to_target * dist_rate - dist_buf );
#endif

        if ( opp_to_target - virtual_dash
             < receiver_to_target * dist_rate - dist_buf )
        {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
            dlog.addText( Logger::PASS,
                          "______ opp%d(%.1f %.1f) is closer than receiver. %.2f < %.2f %s",
                          (*it)->unum(),
                          (*it)->pos().x, (*it)->pos().y,
                          opp_to_target, receiver_to_target,
                          very_aggressive ? "aggressive" : "" );
#endif
            return false;
        }

        Vector2D ball_to_opp = (*it)->inertiaPoint( 3 ); //(*it)->pos() + (*it)->vel();
        ball_to_opp -= world.ball().pos();
        ball_to_opp -= first_vel;
        ball_to_opp.rotate( minus_target_angle );

        if ( 0.0 < ball_to_opp.x && ball_to_opp.x < target_dist )
        {
            ///
            /// 2008-04-29 akiyama new algorithm
            ///
            const double ball_steps_to_project
                = calc_length_geom_series( next_speed,
                                           ball_to_opp.x,
                                           ServerParam::i().ballDecay() );

            double opp2line_dist = ball_to_opp.absY();
            //opp2line_dist -= ServerParam::i().defaultKickableArea();
            opp2line_dist -= ptype->kickableArea();
            opp2line_dist -= 0.2;
            opp2line_dist -= s_distFromHolder( *it, holder ) * 0.02;

            if ( (*it)->velCount() <= 1 )
            {
                opp2line_dist -= ( (*it)->vel().rotatedVector( minus_target_angle ).x
                                   * ( 1.0 - std::pow( ptype->playerDecay(), ball_steps_to_project ) )
                                   / ( 1.0 - ptype->playerDecay() ) );
            }

            if ( opp2line_dist < 0.0 )
            {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
                dlog.addText( Logger::PASS,
                              "______ opp%d(%.1f %.1f) is already on pass line.",
                              (*it)->unum(),
                              (*it)->pos().x, (*it)->pos().y );
#endif
                return false;
            }

            int opp_line_reach_cycle = ptype->cyclesToReachDistance( opp2line_dist );

#if 0
            if ( (*it)->bodyCount() <= 3 )
            {
                if ( ( ( target_point - (*it)->pos() ).th() - (*it)->body() ).abs()
                     > 95.0 )
                {
                    opp_line_reach_cycle += 1;
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
                    dlog.addText( Logger::PASS,
                                  "________ 1 step penalty to body direction" );
#endif
                }
            }
#else
            if ( (*it)->bodyCount() == 0
                 && (*it)->velCount() == 0 )
            {
                if ( ( ( target_point - (*it)->pos() ).th() - (*it)->body() ).abs() > 120.0
                     && (*it)->vel().r2() > std::pow( 0.2, 2 ) )
                {
                    opp_line_reach_cycle += 1;
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
                    dlog.addText( Logger::PASS,
                                  "________ 1 step penalty to body direction" );
#endif
                }
            }
#endif
            opp_line_reach_cycle -= bound( 0, (*it)->posCount(), 5 );

            if ( ball_steps_to_project < 0.0
                 || opp_line_reach_cycle < ball_steps_to_project * 1.05 + 0.5 ) // 20080821
                //|| opp_line_reach_cycle < ball_steps_to_project * 1.02 + 0.5 ) // 20080718
                //|| opp_line_reach_cycle < ball_steps_to_project + 0.5 ) // 20080713
                //|| opp_line_reach_cycle < ball_steps_to_project * 1.05 + 0.5 ) // old
            {
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
                dlog.addText( Logger::PASS,
                              "______ xx opp can reach pass line."
                              " project_x=%.1f opp_y=%.1f"
                              " opp_reach_cycle=%d"
                              " ball_reach_step_to_project=%.1f",
                              ball_to_opp.x, ball_to_opp.y,
                              opp_line_reach_cycle,
                              ball_steps_to_project );
#endif
                return false;
            }
#ifdef DEBUG_PRINT_THROUGH_PASS_LEVEL_3
            dlog.addText( Logger::PASS,
                          "______ ok "
                          " project_x=%.1f opp_y=%.1f"
                          " opp_reach_cycle=%d"
                          " ball_reach_step_to_project=%.1f",
                          ball_to_opp.x, ball_to_opp.y,
                          opp_line_reach_cycle,
                          ball_steps_to_project );
#endif
        }
    }

    return true;
}
