// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

//#define DEBUG_PRINT

#include "actgen_through_pass_2.h"
#include "simple_pass_checker.h"
#include "world_model_ext.h"
#include <rcsc/common/logger.h>
#include <rcsc/math_util.h>
#include <set>
#include <utility>

ActGen_ThroughPass2::ActGen_ThroughPass2()
{
}

ActGen_ThroughPass2::~ActGen_ThroughPass2()
{
}


// XXX: should unify with bhv_basic_kick.cpp
static
double
s_get_ball_speed_for_pass( const double & distance );

void
ActGen_ThroughPass2::addCandidates( std::vector< rcsc::ActionStatePair > * result,
                                    const rcsc::PredictState & state,
                                    const rcsc::WorldModel &,
                                    const std::vector< rcsc::ActionStatePair > & path ) const
{
#ifdef DEBUG_PRINT
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        __FILE__": ActGen_ThroughPass2::addCandidates()" );
#endif

    static const int VALID_PLAYER_THRESHOLD = 10;

    std::set< double > opp_y_set;

    const rcsc::PlayerCont::const_iterator opp_end = state.opponents().end();
    for ( rcsc::PlayerCont::const_iterator opp = state.opponents().begin();
          opp != opp_end;
          ++opp )
    {
        if ( (*opp).posCount() > VALID_PLAYER_THRESHOLD
             || (*opp).isGhost() )
        {
            continue;
        }

        if ( ! (*opp).pos().x < state.offsideLineX() - 5.0 )
        {
            continue;
        }

        opp_y_set.insert( (*opp).pos().y );
    }
    opp_y_set.insert( - rcsc::ServerParam::i().pitchHalfWidth() );
    opp_y_set.insert( + rcsc::ServerParam::i().pitchHalfWidth() );

    std::vector< double > y_candidates;
    {
        const std::set< double >::const_iterator opp_y_set_end = opp_y_set.end();
        const std::set< double >::const_iterator opp_y_set_begin = opp_y_set.begin();
        double old_y = 0.0;
        for ( std::set< double >::const_iterator opp_y = opp_y_set.begin();
              opp_y != opp_y_set_end;
              ++opp_y )
        {
            if ( opp_y != opp_y_set_begin )
            {
                y_candidates.push_back( ( *opp_y + old_y ) / 2.0 );
            }

            old_y = *opp_y;
        }
    }


    std::vector<double> length_candidates;

    for ( double d = -2.0; d <= 30.0 + rcsc::EPS; )
    {
        length_candidates.push_back( d );

        if ( path.size() >= 1 )
        {
            d += 10.0;
        }
        else if ( d >= 5.0 + rcsc::EPS )
        {
            d += 2.0;
        }
        else if ( d >= 10.0 + rcsc::EPS )
        {
            d += 5.0;
        }
        else
        {
            d += 1.0;
        }
    }

    const rcsc::AbstractPlayerObject * holder = state.ballHolder();
    const rcsc::Vector2D ball_pos = state.ballPos();

    //
    // add pass candidates
    //
    const rcsc::SimplePassChecker pass_check;

    const rcsc::PredictState::PlayerPtrCont::const_iterator end = state.allTeammates().end();
    for ( rcsc::PredictState::PlayerPtrCont::const_iterator pl = state.allTeammates().begin();
          pl != end;
          ++pl )
    {
        if ( ! (*pl)->valid()
             || (*pl)->posCount() > VALID_PLAYER_THRESHOLD
             || (*pl)->isGhost()
             || (*pl)->unum() == rcsc::Unum_Unknown )
        {
            continue;
        }

        if ( (*pl)->pos().x < state.offsideLineX() - 10.0 )
        {
            continue;
        }

        rcsc::Vector2D teammate_pos = (*pl)->pos();
        if ( (*pl)->velCount() == 0 )
        {
            teammate_pos += (*pl)->vel() * 2.0;
        }

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            __FILE__": checking through pass to player %d",
                            (*pl)->unum() );
#endif

        for ( size_t y_index = 0; y_index < y_candidates.size(); ++y_index )
        {
            const rcsc::Vector2D on_line_pos( state.offsideLineX(),
                                              y_candidates[y_index] );

            for ( size_t len_index = 0; len_index < length_candidates.size(); ++len_index )
            {
                const rcsc::Vector2D diff( on_line_pos - ball_pos );

                rcsc::AngleDeg th = diff.th();
                double r = diff.r();
                if ( r < rcsc::EPS * 2 )
                {
                    continue;
                }

                rcsc::Vector2D receive_point = ball_pos + diff.setLengthVector( r + length_candidates[len_index] );

                const double ball_dist = ( state.ballPos() - receive_point ).r();
                const double receiver_dist = ( teammate_pos - receive_point ).r();
                const double ball_first_speed = s_get_ball_speed_for_pass( ball_dist );
                long ball_step = rcsc::calc_length_geom_series
                    ( ball_first_speed,
                      ball_dist,
                      rcsc::ServerParam::i().ballDecay() );

                double receiver_step
                    = (*pl)->playerTypePtr()->cyclesToReachDistance( receiver_dist );

#if 0
                if ( receiver_step + 3 > ball_step )
#else
                if ( receiver_step + 2 >= ball_step )
#endif
                {
                    continue;
                }

                if ( ! pass_check( state, holder, &(**pl), receive_point, ball_first_speed ) )
                {
                    continue;
                }

                result->push_back( rcsc::ActionStatePair
                                   ( new rcsc::CooperativeAction
                                     ( rcsc::CooperativeAction::ActionType::Pass,
                                       receive_point,
                                       (*pl)->unum(),
                                       ball_first_speed,
                                       receiver_step,
                                       ball_step,
                                       true,
                                       "through 2" ),
                                     new rcsc::PredictState
                                     ( state, ball_step,
                                       (*pl)->unum(), receive_point ) ) );
            }
        }
    }
}


static
double
s_get_ball_speed_for_pass( const double & distance )
{
    if ( distance >= 20.0 )
    {
        return 3.0;
    }
    else if ( distance >= 8.0 )
    {
        return 2.5;
    }
    else if ( distance >= 5.0 )
    {
        return 1.8;
    }
    else
    {
        return 1.5;
    }
}
