// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "actgen_voronoi_pass.h"

#include "simple_pass_checker.h"
#include "world_model_ext.h"

#include <rcsc/player/field_analyzer.h>

#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/geom/voronoi_diagram.h>
#include <rcsc/timer.h>

#include <boost/shared_ptr.hpp>

//#define DEBUG_PRINT

using rcsc::dlog;

ActGen_VoronoiPass::ActGen_VoronoiPass()
{
}

ActGen_VoronoiPass::~ActGen_VoronoiPass()
{
}

// XXX: should unify with bhv_basic_kick.cpp
static
double
s_get_ball_speed_for_pass( const double & distance );

void
ActGen_VoronoiPass::addCandidates( std::vector< rcsc::ActionStatePair > * result,
                                   const rcsc::PredictState & state,
                                   const rcsc::WorldModel &,
                                   const std::vector< rcsc::ActionStatePair > & ) const
{
    static const int VALID_PLAYER_THRESHOLD = 10;

    //
    // cache of pass target candidates
    //
    static std::vector< rcsc::Vector2D > cand;
    static rcsc::GameTime cache_time;

    rcsc::Timer timer;

    if ( cache_time != state.time() )
    {
        // update cache time
        cache_time = state.time();

        // clear old candidates
        cand.clear();


        //
        // get voronoi diagram
        //
        const rcsc::VoronoiDiagram & voronoi = FieldAnalyzer::i().passVoronoiDiagram();
        const double MIN_LEN = 2.0;
        const int MAX_DIV = 10;

        voronoi.getPointsOnSegments( MIN_LEN,
                                     MAX_DIV,
                                     &cand );

#ifdef DEBUG_PRINT
        const rcsc::VoronoiDiagram::Segment2DCont::const_iterator s_end = voronoi.resultSegments().end();
        for ( rcsc::VoronoiDiagram::Segment2DCont::const_iterator it = voronoi.resultSegments().begin();
              it != s_end;
              ++it )
        {
            dlog.addLine( rcsc::Logger::RECURSIVE,
                          (*it).origin(), (*it).terminal(),
                          "#0000ff" );
        }

        const std::vector< rcsc::Vector2D >::const_iterator p_end = cand.end();
        for ( std::vector< rcsc::Vector2D >::const_iterator p = cand.begin();
              p != p_end;
              ++p )
        {
            dlog.addRect( rcsc::Logger::RECURSIVE,
                          p->x - 0.15,  p->y - 0.15,
                          0.3, 0.3,
                          "#ff00ff" );
        }
#endif

        dlog.addText( rcsc::Logger::RECURSIVE,
                      "voronoi_pass: generate candidates, elapsed %f [ms]",
                      timer.elapsedReal() );
    }



    //
    // check pass to each pass target
    //
    dlog.addText( rcsc::Logger::RECURSIVE,
                  "voronoi_pass: ball = (%f, %f)",
                  state.ballPos().x, state.ballPos().y );

    const rcsc::AbstractPlayerObject * holder = state.ballHolder();

    const rcsc::SimplePassChecker pass_check;

    const rcsc::PredictState::PlayerPtrCont::const_iterator pl_end = state.allTeammates().end();
    for ( rcsc::PredictState::PlayerPtrCont::const_iterator pl = state.allTeammates().begin();
          pl != pl_end;
          ++pl )
    {
        if ( ! (*pl)->valid()
             || (*pl)->posCount() > VALID_PLAYER_THRESHOLD
             || (*pl)->isGhost()
             || (*pl)->unum() == state.ballHolderUnum()
             || (*pl)->unum() == rcsc::Unum_Unknown
             || (*pl)->unumCount() > VALID_PLAYER_THRESHOLD
             || (*pl)->isTackling() )
        {
            continue;
        }

#ifdef DEBUG_PRINT
        dlog.addText( rcsc::Logger::RECURSIVE,
                      "voronoi_pass: checking teammate %d, pos = (%f, %f)",
                      (*pl)->unum(), (*pl)->pos().x, (*pl)->pos().y );
#endif

        const std::vector< rcsc::Vector2D >::const_iterator end = cand.end();
        for ( std::vector< rcsc::Vector2D >::const_iterator it = cand.begin();
              it != end;
              ++it )
        {
            const rcsc::Vector2D & receive_point = *it;

            //if ( receive_point.absX() >= rcsc::ServerParam::i().pitchHalfLength()
            //     || receive_point.absY() >= rcsc::ServerParam::i().pitchHalfWidth() )
            //{
            //    continue;
            //}

#ifdef DEBUG_PRINT
            dlog.addText( rcsc::Logger::RECURSIVE,
                          "voronoi_pass: receive_point = (%f, %f)",
                          receive_point.x, receive_point.y );
#endif

            rcsc::Vector2D teammate_pos = (*pl)->pos();

            if ( (*pl)->velCount() == 0 )
            {
                teammate_pos += (*pl)->vel() * 2.0;
            }


            const double ball_dist = ( state.ballPos() - receive_point ).r();
            const double ball_first_speed = s_get_ball_speed_for_pass( ball_dist );
            const double receiver_dist = ( teammate_pos - receive_point ).r();

            long ball_step = rcsc::calc_length_geom_series
                             ( ball_first_speed,
                               ball_dist,
                               rcsc::ServerParam::i().ballDecay() );

            double receiver_step = (*pl)->playerTypePtr()->cyclesToReachDistance( receiver_dist );

            if ( receiver_step + 3 > ball_step )
            {
                continue;
            }

            if ( ! pass_check( state, holder, &(**pl), receive_point, ball_first_speed ) )
            {
                continue;
            }

            result->push_back( rcsc::ActionStatePair
                               ( new rcsc::CooperativeAction
                                 ( rcsc::CooperativeAction::ActionType::Pass,
                                   receive_point,
                                   (*pl)->unum(),
                                   ball_first_speed,
                                   receiver_step,
                                   ball_step,
                                   false,
                                   "voronoi" ),
                                 new rcsc::PredictState
                                 ( state, ball_step,
                                   (*pl)->unum(), receive_point ) ) );
        }
    }


    dlog.addText( rcsc::Logger::RECURSIVE,
                  "voronoi_pass: done, elapsed %f [ms]",
                  timer.elapsedReal() );
}


static
double
s_get_ball_speed_for_pass( const double & distance )
{
    if ( distance >= 20.0 )
    {
        return 3.0;
    }
    else if ( distance >= 8.0 )
    {
        return 2.5;
    }
    else if ( distance >= 5.0 )
    {
        return 1.8;
    }
    else
    {
        return 1.5;
    }
}
