// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_goalie_free_kick.h"

#include "strategy.h"

#include "bhv_savior.h"
#include "bhv_goalie_basic_move.h"
#include "bhv_set_play.h"

#include <rcsc/action/body_clear_ball.h>
#include "bhv_pass_test.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/neck_scan_field.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/rect_2d.h>

using namespace rcsc;

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_GoalieFreeKick::execute( PlayerAgent * agent )
{
    static bool s_first_move = false;
    static bool s_second_move = false;
    static int s_second_wait_count = 0;

    const WorldModel & wm = agent->world();

    dlog.addText( Logger::TEAM,
                  __FILE__": Bhv_GoalieFreeKick" );

    if ( wm.gameMode().type() != GameMode::GoalieCatch_
         || wm.gameMode().side() != agent->world().ourSide()
         || ! wm.self().isKickable() )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": Not a goalie catch mode" );
        if ( Strategy::i().savior() )
        {
            Bhv_Savior().execute( agent );
        }
        else
        {
            Bhv_GoalieBasicMove().execute( agent );
        }
        return true;
    }


    const long time_diff
        = wm.time().cycle()
        - agent->effector().getCatchTime().cycle();
    //- M_catch_time.cycle();

    // reset flags & wait
    if ( time_diff <= 2 )
    {
        s_first_move = false;
        s_second_move = false;
        s_second_wait_count = 0;

        doWait( agent );
        return true;
    }

    // first move
    if ( ! s_first_move )
    {
        Vector2D move_target( ServerParam::i().ourPenaltyAreaLineX() - 1.5,
                              wm.ball().pos().y > 0.0 ? -13.0 : 13.0 );
        s_first_move = true;
        s_second_move = false;
        s_second_wait_count = 0;
        agent->doMove( move_target.x, move_target.y );
        agent->setNeckAction( new Neck_ScanField );
        return true;
    }

    // after first move
    // check stamina recovery or wait teammate
    Rect2D our_pen( Vector2D( -52.5, -40.0 ),
                    Vector2D( -36.0, 40.0 ) );
    if ( time_diff < 50
         || wm.setplayCount() < 3
         || ( time_diff < ServerParam::i().dropBallTime() - 15
              && ( wm.self().stamina() < ServerParam::i().staminaMax() * 0.9
                   || wm.existTeammateIn( our_pen, 20, true )
                   )
              )
         )
    {
        doWait( agent );
        return true;
    }

    // second move
    if ( ! s_second_move )
    {
        Vector2D kick_point = getKickPoint( agent );
        agent->doMove( kick_point.x, kick_point.y );
        agent->setNeckAction( new Neck_ScanField );
        s_second_move = true;
        s_second_wait_count = 0;
        return true;
    }

    s_second_wait_count++;

    // after second move
    // wait see info
    if ( s_second_wait_count < 5
         || wm.seeTime() != wm.time() )
    {
        doWait( agent );
        return true;
    }

    if ( Bhv_SetPlay::is_delaying_tactics_situation( agent ) )
    {
        agent->debugClient().addMessage( "GoalieKick:Delaying" );
        dlog.addText( Logger::TEAM,
                      __FILE__":  delaying" );

        doWait( agent );
        return true;
    }

    s_first_move = false;
    s_second_move = false;
    s_second_wait_count = 0;

    doKick( agent );

    return true;
}

/*-------------------------------------------------------------------*/
/*!

 */
Vector2D
Bhv_GoalieFreeKick::getKickPoint( const PlayerAgent * agent )
{
    static const double base_x = -43.0;
    static const double base_y = 10.0;

    const WorldModel & wm = agent->world();

    std::vector< std::pair< Vector2D, double > > candidates;
    candidates.reserve( 4 );
    candidates.push_back( std::make_pair( Vector2D( base_x, base_y ), 0.0 ) );
    candidates.push_back( std::make_pair( Vector2D( base_x, -base_y ), 0.0 ) );
    candidates.push_back( std::make_pair( Vector2D( base_x, 0.0 ), 0.0 ) );

    const PlayerPtrCont::const_iterator opps_end = wm.opponentsFromSelf().end();
    for ( PlayerPtrCont::const_iterator o = wm.opponentsFromSelf().begin();
          o != opps_end;
          ++o )
    {
        for ( std::vector< std::pair< Vector2D, double > >::iterator it = candidates.begin();
              it != candidates.end();
              ++it )
        {
            it->second += 1.0 / (*o)->pos().dist2( it->first );
        }
    }

    Vector2D best_pos = candidates.front().first;
    double min_cong = 10000.0;
    for ( std::vector< std::pair< Vector2D, double > >::iterator it = candidates.begin();
          it != candidates.end();
          ++it )
    {
        if ( it->second < min_cong )
        {
            best_pos = it->first;
            min_cong = it->second;
        }
    }

    return best_pos;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
Bhv_GoalieFreeKick::doKick( PlayerAgent * agent )
{
    const WorldModel & wm = agent->world();

    const Bhv_PassTest::PassRoute * pass
        = static_cast< const Bhv_PassTest::PassRoute * >( 0 );
    double best_score = -100000.0;

    const std::vector< Bhv_PassTest::PassRoute > & passes
        = Bhv_PassTest::get_pass_route_cont( wm );
    const std::vector< Bhv_PassTest::PassRoute >::const_iterator end = passes.end();
    for ( std::vector< Bhv_PassTest::PassRoute >::const_iterator p = passes.begin();
          p != end;
          ++p )
    {
        double opp_dist = 0.0;
        const PlayerObject * opp = wm.getOpponentNearestTo( p->receive_point_, 20, &opp_dist );
        if ( opp && opp_dist < 3.0 )
        {
            continue;
        }

        double score = p->score_;
        double dist =  wm.ball().pos().dist( p->receive_point_ );
        score *= std::max( 0.0, 30.0 - dist ) / 30.0;

        if ( p->receive_point_.absY() < 7.0 )
        {
            score *= 0.5;
        }

        if ( score > best_score )
        {
            best_score = score;
            pass = &(*p);
        }
    }

    if ( pass )
    {
        Body_KickOneStep( pass->receive_point_,
                          pass->first_speed_ ).execute( agent );
        dlog.addText( Logger::TEAM,
                      __FILE__": kick to (%.1f, %.1f)",
                      pass->receive_point_.x,
                      pass->receive_point_.y );
        agent->setNeckAction( new Neck_ScanField() );
        return;
    }

    Body_ClearBall().execute( agent );
    agent->setNeckAction( new Neck_ScanField() );
}

/*-------------------------------------------------------------------*/
/*!

 */
void
Bhv_GoalieFreeKick::doWait( PlayerAgent * agent )
{
    const WorldModel & wm = agent->world();
    Vector2D face_target( 0.0, 0.0 );

    if ( wm.self().pos().x > ServerParam::i().ourPenaltyAreaLineX()
         - ServerParam::i().ballSize()
         - wm.self().playerType().playerSize()
         - 0.5 )
    {
        face_target.assign( ServerParam::i().ourPenaltyAreaLineX() - 1.0,
                            0.0 );
    }

    Body_TurnToPoint( face_target ).execute( agent );
    agent->setNeckAction( new Neck_ScanField() );
}
