// -*-c++-*-

/*!
  \file bhv_pass_test2008.h
  \brief advanced pass planning & behavior.
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA, Hiroki SHIMORA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef RCSC_ACTION_BHV_PASS_TEST_2008_H
#define RCSC_ACTION_BHV_PASS_TEST_2008_H

#include <rcsc/player/soccer_action.h>
#include "recursive_action_chain_graph.h"
#include <rcsc/geom/vector_2d.h>

#include <vector>
#include <boost/shared_ptr.hpp>
#include <functional>
#include <vector>

namespace rcsc {

class WorldModel;
class AbstractPlayerObject;
class PlayerObject;
class RecursiveActionChainGraph;

/*!
  \class Bhv_PassTest2008
  \brief advanced pass planning & behavior.
 */
class Bhv_PassTest2008
    : public SoccerBehavior {
public:
    enum PassType {
        DIRECT  = 1,
        LEAD    = 2,
        THROUGH = 3,
        RECURSIVE = 4
    };

    /*!
      \struct PassRoute
      \brief pass route information object, that contains type,
      receiver info, receive point and ball first speed.
     */
    struct PassRoute {
        int index_;
        PassType type_;
        const AbstractPlayerObject * receiver_;
        Vector2D receive_point_;
        double first_speed_;
        bool one_step_kick_;
        boost::shared_ptr< const std::vector< ActionStatePair > > action_chain_;
        double score_;

        PassRoute( PassType type,
                   const AbstractPlayerObject * receiver,
                   const Vector2D & point,
                   const double & speed,
                   const bool one_step_kick,
                   const boost::shared_ptr< const std::vector< ActionStatePair > > & action_chain )
            : index_( -1 )
            , type_( type )
            , receiver_( receiver )
            , receive_point_( point )
            , first_speed_( speed )
            , one_step_kick_( one_step_kick )
            , action_chain_( action_chain )
            , score_( 0.0 )
          { }

        PassRoute( PassType type,
                   const AbstractPlayerObject * receiver,
                   const Vector2D & point,
                   const double & speed,
                   const bool one_step_kick,
                   std::vector< ActionStatePair > * action_chain
                   = static_cast< std::vector< ActionStatePair > * >( 0 ) )
            : index_( -1 )
            , type_( type )
            , receiver_( receiver )
            , receive_point_( point )
            , first_speed_( speed )
            , one_step_kick_( one_step_kick )
            , action_chain_( action_chain )
            , score_( 0.0 )
          { }
    };

    class PassRouteScoreComp
        : public std::binary_function< PassRoute, PassRoute, bool > {
    public:
        result_type operator()( const first_argument_type & lhs,
                                const second_argument_type & rhs ) const
          {
              return lhs.score_ < rhs.score_;
          }
    };

private:

    //! cache created time
    static GameTime S_cache_time;

    //! cached calculated pass data
    static std::vector< PassRoute > S_cached_pass_route;

    //! cached result
    static const PassRoute * S_best_pass;

    //! cached recursive final state
    static boost::shared_ptr< const PassRoute > S_cached_recursive_final_state;

    //! cached calculated recursive action chain graph
    static boost::shared_ptr< const rcsc::RecursiveActionChainGraph > S_rag;


private:
    /*!
      \brief generate passes and evaluate
      \param wm const rerefence to the WorldModel
    */
    static void create_cache( const WorldModel & wm );

public:
    /*!
      \brief accessible from global.
    */
    Bhv_PassTest2008()
      { }

    /*!
      \brief execute action
      \param agent pointer to the agent itself
      \return true if action is performed
    */
    bool execute( PlayerAgent * agent );

    /*!
      \brief calculate best pass route
      \param world const rerefence to the WorldModel
      \return pointer to the pass route variable. if no route, NULL is returned.
    */
    static
    const
    PassRoute * get_best_pass_first( const WorldModel & world );

    /*!
      \brief calculate final state of best pass route
      \param world const rerefence to the WorldModel
      \return pointer to the pass route variable. if no route, NULL is returned.
    */
    static
    const
    PassRoute * get_best_pass_final( const WorldModel & world );


    static
    const
    std::vector< PassRoute > & get_pass_route_cont( const WorldModel & world )
      {
          get_best_pass_first( world );
          return S_cached_pass_route;
      }

private:

    bool doPassKickFindReceiver( PlayerAgent * agent,
                                 const PassType pass_type,
                                 const AbstractPlayerObject * receiver,
                                 const Vector2D & receive_pos );

    void doSayPass( PlayerAgent * agent,
                    const PassType pass_type,
                    const int receiver_unum,
                    const Vector2D & receive_pos );

    /*!
      \brief execute pass using pass course table
      \param agent pointer to the agent itself
      \return true if action is performed
    */
    bool doRecursiveSearchPass( PlayerAgent * agent );

    bool doRecursiveActionChainPass( PlayerAgent * agent );

private:
    static
    void create_routes( const WorldModel & world );

    static
    void create_direct_pass( const WorldModel & world,
                             const PlayerObject * teammates );
    static
    void create_lead_pass( const WorldModel & world,
                           const PlayerObject * teammates );
    static
    void create_through_pass( const WorldModel & world,
                              const PlayerObject * teammates );
    static
    void create_recursive_pass( const WorldModel & world );

    static
    void create_recursive_action_chain_pass( const WorldModel & world );

    static
    bool verify_direct_pass( const WorldModel & world,
                             const PlayerObject * receiver,
                             const Vector2D & target_point,
                             const double & target_dist,
                             const AngleDeg & target_angle,
                             const double & first_speed );
    static
    bool verify_through_pass( const WorldModel & world,
                              const PlayerObject * receiver,
                              const Vector2D & receiver_pos,
                              const Vector2D & target_point,
                              const double & target_dist,
                              const AngleDeg & target_angle,
                              const double & first_speed,
                              const double & reach_step,
                              const bool is_through_pass );

    static
    void evaluate_routes( const WorldModel & world );

    static
    bool can_kick_by_one_step( const WorldModel & world,
                               const double & first_speed,
                               const AngleDeg & target_angle );
};

}

#endif
