// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "default_field_evaluator.h"
#include "fixed_pass_course_table.h"
#include "world_model_ext.h"
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/math_util.h>

#include <cmath>
#include <cfloat>

// #define OLD_CHAIN_EVALUATION

// #define DEBUG_PRINT


static const int VALID_PLAYER_THRESHOLD = 8;
static const double SHOOT_DIST_THRESHOLD = 16.0;
static const double SHOOT_ANGLE_THRESHOLD = 20.0;

static double s_evaluate_state( const rcsc::PredictState & state,
                                const rcsc::WorldModel & wm,
                                const std::vector< rcsc::ActionStatePair > & path );


static
double
get_opponent_dist( const rcsc::WorldModel & wm,
                   const rcsc::Vector2D & point,
                   int opponent_additional_chase_time = 0,
                   int valid_opponent_thr = -1 )
{
    static const double MAX_DIST = 65535.0;

    double min_dist = MAX_DIST;

    const rcsc::AbstractPlayerCont::const_iterator end = wm.allOpponents().end();
    for ( rcsc::AbstractPlayerCont::const_iterator it = wm.allOpponents().begin();
          it != end;
          ++it )
    {
        if ( (*it)->ghostCount() >= 3 )
        {
            continue;
        }

        if ( valid_opponent_thr != -1 )
        {
            if ( (*it)->posCount() > valid_opponent_thr )
            {
                continue;
            }
        }

        double dist = (*it)->pos().dist( point );
        dist -= ( rcsc::bound( 0, (*it)->posCount() - 2, 2 )
                  + opponent_additional_chase_time )
            * (*it)->playerTypePtr()->realSpeedMax();

        if ( (*it)->goalie()
             && dist > 5.0 )
        {
            continue;
        }

        if ( dist < min_dist )
        {
            min_dist = dist;
        }
    }

    return min_dist;
}


/*-------------------------------------------------------------------*/
/*!

 */
DefaultFieldEvaluator::DefaultFieldEvaluator()
{
}

/*-------------------------------------------------------------------*/
/*!

 */
DefaultFieldEvaluator::~DefaultFieldEvaluator()
{
}

/*-------------------------------------------------------------------*/
/*!

 */
#ifdef OLD_CHAIN_EVALUATION
double
DefaultFieldEvaluator::operator()
                   ( const rcsc::PredictState & state,
                     const rcsc::WorldModel & wm,
                     const std::vector< rcsc::ActionStatePair > & path ) const
{
    return s_evaluate_state( state, wm , path )
           + get_chain_length_penalty( path, state );
}
#elif 0
double
DefaultFieldEvaluator::operator()
                   ( const rcsc::PredictState & state,
                     const rcsc::WorldModel & wm,
                     const std::vector< rcsc::ActionStatePair > & path ) const
{
    const double final = s_evaluate_state( state, wm, path );

    double total = final;

    for ( int i = 0; i < static_cast< int >( path.size() ) - 1 ; ++i )
    {
        const double eval = s_evaluate_state( path[i].state(), wm, path );
        const double diff = final - eval;

        total += diff * 0.1;
    }

    return total + get_chain_length_penalty( path, state );
}
#elif 0
double
DefaultFieldEvaluator::operator()
                   ( const rcsc::PredictState & state,
                     const rcsc::WorldModel & wm,
                     const std::vector< rcsc::ActionStatePair > & path ) const
{
    if ( path.empty() )
    {
        return s_evaluate_state( state, wm, path )
               + get_chain_length_penalty( path, state, path );
    }


    const double first = s_evaluate_state( path[0].state(), wm, path );

    double total = first;

    const double rate = 0.98;

    for ( int i = 1; i < static_cast< int >( path.size() ) ; ++i )
    {
        const double eval = s_evaluate_state( path[i].state(), wm, path );
        const double diff = eval - first;

        total += diff * std::pow( rate, i );
    }

    return total + get_chain_length_penalty( path, state );
}
#elif 1
double
DefaultFieldEvaluator::operator()
                   ( const rcsc::PredictState & state,
                     const rcsc::WorldModel & wm,
                     const std::vector< rcsc::ActionStatePair > & path ) const
{
    const double final = s_evaluate_state( state, wm, path );
    double total = final;

#if 0
    for ( int i = 0; i < static_cast< int >( path.size() ) - 1 ; ++i )
#else
    for ( int i = 0; i < static_cast< int >( path.size() ); ++i )
#endif
    {
        //
        // free radius
        //
        int opponent_additional_chase_time = 0;

#if 1
        switch( path[i].action().type() )
        {
        case rcsc::CooperativeAction::ActionType::Hold:
            opponent_additional_chase_time = path[i].action().ballReachStep() * 1.0;
            break;
        case rcsc::CooperativeAction::ActionType::Dribble:
            opponent_additional_chase_time = path[i].action().ballReachStep() * 0.9;
            break;
        case rcsc::CooperativeAction::ActionType::Pass:
            opponent_additional_chase_time = path[i].action().ballReachStep() * 0.2;
            break;
        case rcsc::CooperativeAction::ActionType::Shoot:
            opponent_additional_chase_time = 0;
            break;
        }

        const double r = get_opponent_dist( wm, path[i].state().ballHolder()->pos(), opponent_additional_chase_time );
#endif

        const double oppenent_dist_penalty_rate = 0.3;

        double penalty;

        if ( r < 10.0 )
        {
            penalty = -std::max( 10.0 - r, 0.0 ) * oppenent_dist_penalty_rate;
        }
        else
        {
            penalty = 0.0;
        }


        const double goal_dist = state.ballHolder()->pos().dist( rcsc::ServerParam::i().theirTeamGoalPos() );

        if ( goal_dist < 20.0 )
        {
#if 0
            penalty += ( 20.0 - goal_dist ) * 0.5;
#else
            penalty += ( std::min( 0.0, 10.0 - goal_dist ) + 10.0 ) * 1.0;
#endif
        }

        total += std::min( penalty, 0.0 );

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) path=%d free radius penalty %f (%f)",
                            i,
                            penalty,
                            total );
#endif
    }

#if 0
    double chain_length_penalty = get_chain_length_penalty( path, state ) * 0.1;
#elif 0
    double chain_length_penalty = get_chain_length_penalty( path, state ) * 0.01;
#elif 0
    double chain_length_penalty = get_chain_length_penalty( path, state ) * 0.03;
#elif 0
    double chain_length_penalty = get_chain_length_penalty( path, state ) * 0.05;
#elif 0
    double chain_length_penalty = get_chain_length_penalty( path, state ) * 0.06;
#elif 0
    double chain_length_penalty = get_chain_length_penalty( path, state ) * 0.065;
#elif 0
    double chain_length_penalty = get_chain_length_penalty( path, state ) * 0.07;
#elif 0
    double chain_length_penalty = get_chain_length_penalty( path, state ) * 0.08;
#else
    double chain_length_penalty = get_chain_length_penalty( path, state ) * 0.005;
#endif

#ifdef DEBUG_PRINT
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "(eval) chain length penalty %f (%f)",
                        chain_length_penalty, total + chain_length_penalty );
#endif

    return total + chain_length_penalty;
}
#endif



static double s_evaluate_state( const rcsc::PredictState & state,
                                const rcsc::WorldModel & wm,
                                const std::vector< rcsc::ActionStatePair > & path )
{
    const rcsc::ServerParam & param = rcsc::ServerParam::i();

    const rcsc::AbstractPlayerObject * holder = state.ballHolder();
    const int holder_unum = holder->unum();

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "========= (evaluate_state) ==========" );
#endif

    //
    // if holder is invalid, return bad evaluation
    //
    if ( ! holder )
    {
#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) XXX null holder" );
#endif
        return - DBL_MAX / 2.0;
    }


    //
    // ball is in opponent goal
    //
    if ( state.ballPos().x > + ( param.pitchHalfLength() - 0.1 )
         && state.ballPos().absY() < param.goalHalfWidth() )
    {
#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) *** in opponent goal" );
#endif
#if 0
        return +1.0e+7 + get_chain_length_penalty( path, state );
#else
        return +1.0e+7;
#endif
    }

    //
    // ball is in our goal
    //
    if ( state.ballPos().x < - ( param.pitchHalfLength() - 0.1 )
         && state.ballPos().absY() < param.goalHalfWidth() )
    {
#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) XXX in our goal" );
#endif

        return -1.0e+7;
    }


    //
    // out of pitch
    //
    if ( state.ballPos().absX() > param.pitchHalfLength()
         || state.ballPos().absY() > param.pitchHalfWidth() )
    {
#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) XXX out of pitch" );
#endif

        return - DBL_MAX / 2.0;
    }


    //
    // set basic point
    //
    double point = state.ballPos().x;

#ifdef DEBUG_PRINT
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "(eval) ball pos (%f, %f)",
                        state.ballPos().x, state.ballPos().y );

    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "(eval) initial value (%f)", point );
#endif

    //
    // add bonus for goal, free situation near offside line
    //
    if ( DefaultFieldEvaluator::can_shoot_from
         ( state, state.ballHolder()->pos(), VALID_PLAYER_THRESHOLD ) )
    {
        point += 1.0e+6;
#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) bonus for goal %f (%f)", 1.0e+6, point );
#endif

        if ( holder_unum == state.self().unum() )
        {
            point += 5.0e+5;
#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                "(eval) bonus for goel self %f (%f)", 5.0e+5, point );
#endif
        }
    }
    else if ( DefaultFieldEvaluator::is_free_situation_near_offside_line( state, state.ballHolder() ) )
    {
        const double near_offside_line_bonus = 1.0e+5 + ( state.ballPos().x - param.pitchHalfLength() ) * 2.0;

        point += near_offside_line_bonus;

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) bonus for free near offside line %f (%f)",
                            near_offside_line_bonus,
                            point );
#endif
    }

    if ( state.ballPos().x > wm.offsideLineX() )
    {
        double over_offside_line_bonus = 5.0e+5;

        over_offside_line_bonus += ( state.ballPos().x - state.offsideLineX() ) * 5.0;

        const double their_goal_dist = ( holder->pos() - param.theirTeamGoalPos() ).r();

        if ( their_goal_dist < 50.0 )
        {
            over_offside_line_bonus += ( 50.0 - their_goal_dist ) * 100.0;
        }

        // center of field is higher, side is lower
        over_offside_line_bonus += ( rcsc::ServerParam::i().pitchHalfWidth()
                                     - state.ballPos().absY() ) * 100.0;

        point += over_offside_line_bonus;


#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) bonus for over offside line %f (%f)",
                            over_offside_line_bonus, point );
#endif
    }



    //
    // add distance bonus when near opponent goal
    //
    double their_goal_dist = ( holder->pos() - param.theirTeamGoalPos() ).r();
    point -= their_goal_dist;

#ifdef DEBUG_PRINT
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "(eval) bonus for opponent goal dist %f (%f)",
                        - their_goal_dist,
                        point );
#endif

    if ( their_goal_dist <= 30.0 )
    {
#if 0
        double goal_dist_bonus = ( 30.0 - their_goal_dist ) * 10.0;
#else
        double goal_dist_bonus = ( 30.0 - their_goal_dist ) * 30.0;
#endif

        point += goal_dist_bonus;

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) bonus for opponent goal dist %f (%f)",
                            goal_dist_bonus,
                            point );
#endif
    }


    //
    // add distance penalty when near our goal
    //
    double our_goal_dist = ( holder->pos() - param.ourTeamGoalPos() ).r();
    point += our_goal_dist;

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) penalty for our goal dist %f (%f)",
                            our_goal_dist,
                            point );
#endif
    if ( our_goal_dist <= 25.0 )
    {
        point -= ( 25.0 - our_goal_dist ) * 5.0;

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) penalty for our goal near %f (%f)",
                            - ( 25.0 - our_goal_dist ) * 5.0,
                            point );
#endif
    }



    //
    // add bonus near offside line
    //
    if ( state.ballPos().x > state.offsideLineX() - 5.0 )
    {
        point += ( state.ballPos().x - (state.offsideLineX() -5.0) ) * 10.0;


#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) bonus for near offside line %f (%f)",
                            ( state.ballPos().x - ( state.offsideLineX() - 5.0 ) ) * 10.0,
                            point );
#endif
    }


    //
    // add penalty near defense line
    //
    if ( state.ballPos().x < state.defenseLineX() + 5.0 )
    {
        point += ( state.ballPos().x - (state.defenseLineX() + 5.0) ) * 10.0;

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) penalty for defense line %f (%f)",
                            ( state.ballPos().x - (state.defenseLineX() + 5.0) ) * 10.0,
                            point );
#endif
    }


    //
    // add penalty center of field when our goalie not in penalty area
    //
    const rcsc::AbstractPlayerObject * our_goalie = state.getTeammateGoalie();
    if ( our_goalie != static_cast<rcsc::AbstractPlayerObject *>( 0 ) )
    {
        const bool our_goalie_is_in_side_area
            = ( our_goalie->pos().absY() > param.penaltyAreaHalfWidth() - 5.0
                && our_goalie->posCount() < VALID_PLAYER_THRESHOLD
                && our_goalie->pos().x < -20.0
                && state.ball().pos().x < -20.0 );

        if ( our_goalie_is_in_side_area )
        {

#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                "(eval) our goalie is in side area" );
#endif
            point -= 30.0;

            double side_line_dist;

            if ( our_goalie->pos().y > 0.0 )
            {
                side_line_dist = ( param.penaltyAreaHalfWidth()
                                   - state.ball().pos().y );
            }
            else
            {
                side_line_dist = ( state.ball().pos().y
                                   - (- param.penaltyAreaHalfWidth() ) );
            }

            point -= side_line_dist * 10.0;
        }
    }


    //
    // add absolute x pos bonus
    //
    if ( holder->pos().x >= 20.0 )
    {
        point += ( 20.0 - holder->pos().x ) * 5.0;

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) bonus for absolute_x %f (%f)",
                            ( 20.0 - holder->pos().x ) * 5.0,
                            point );
#endif
    }



    //
    // add free radius bonus
    //
    {
        //const double r = rcsc_ext::getDistOpponentNearestToPoint( wm, holder->pos() );
        const double r = get_opponent_dist( wm, holder->pos() );

#if 0
        double factor = 10.0;
#elif 0
        double factor = 9.5;
#elif 1
        double factor = 9.25;
#elif 0
        double factor = 9.0;
#else
        double factor = 8.0;
#endif

#if 1
        if ( holder_unum == state.self().unum() )
        {
# if 0
            factor = 1000.0;
# elif 0
            factor = 100.0;
# elif 0
            factor = 200.0;
# else
            //factor = 500.0;
            factor = 50.0;
# endif
        }
#endif

        const double opp_dist_threthold = 10.0;

        if ( r < opp_dist_threthold )
        {
            point -= factor * ( opp_dist_threthold - r );

#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "(eval) bonus for free radius %f (%f)",
                            - factor * ( opp_dist_threthold - r ),
                            point );
#endif
        }
    }


    //
    // add pass count bonus (fixed state)
    //
    const FixedPassCourseTable & evaluate_table = FixedPassCourseTable::instance( wm );
    const std::vector< FixedPassCourseTable::ConnectionTable > &
        table = evaluate_table.get_table( wm );

    int pass_count = 0;
    for ( int i = 1; i <= param.maxPlayer(); ++i )
    {
        if ( table[holder_unum].player[i]
             && i != holder->unum() )
        {
            ++pass_count;
        }
    }
   point += std::min( pass_count, 3 ) * 20.0;

#ifdef DEBUG_PRINT
   rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                       "(eval) bonus for pass count(%d) %f (%f)",
                       pass_count,
                       std::min( pass_count, 3 ) * 20.0,
                       point );
#endif


#if 0
    //
    // bonus if single kick pass
    //
    if ( ! path.empty() )
    {
        const rcsc::CooperativeAction & act = path[0].action();

        if ( act.type() == rcsc::CooperativeAction::ActionType::Pass
             && rcsc_ext::WorldModel_canKickByOneStep
                ( wm,
                  act.ballSpeed(),
                  ( act.targetPoint() - state.ballPos() ).th() ) )
        {
            point += 10.0;
        }
    }
#else
    static_cast<void>(path);
#endif

    return point;
}

double
DefaultFieldEvaluator::get_chain_length_penalty( const std::vector< rcsc::ActionStatePair > & path,
                                                 const rcsc::PredictState & state )
{
    double point = 0.0;

#if 0
    //
    // penalty by path length
    //
#if 0
    point -= path.size() * 1.0;
#else
    if ( ! path.empty() )
    {
        //point -= ( path.size() - 1 ) * 0.1;
        point -= ( path.size() - 1 );
    }
#endif
    //point -= std::pow( 2.0, path.size() - 1 );
#else
    static_cast<void>( path );
#endif

#if 1
    //
    // penalty by spend time
    //
    point -= state.spendTime() * 0.1;

#ifdef DEBUG_PRINT
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "(eval) penalty for spent time=(%d) %f (%f)",
                        state.spendTime(),
                        state.spendTime() * 0.1,
                        point );
#endif
#endif

    return point;
}



/*-------------------------------------------------------------------*/
/*!

 */
bool
DefaultFieldEvaluator::can_shoot_from( const rcsc::PredictState & wm,
                                       const rcsc::Vector2D & pos,
                                       long valid_opponent_threshold )
{
    if ( ( pos - rcsc::ServerParam::i().theirTeamGoalPos() ).r()
         > SHOOT_DIST_THRESHOLD )
    {
        return false;
    }

    return get_shoot_free_angle_from_pos( wm, valid_opponent_threshold, pos )
            >= SHOOT_ANGLE_THRESHOLD;
}

/*-------------------------------------------------------------------*/
/*!

 */
double
DefaultFieldEvaluator::get_shoot_free_angle_from_pos( const rcsc::PredictState & wm,
                                                      const long valid_opponent_threshold,
                                                      const rcsc::Vector2D & pos )
{
    const double goal_half_width = rcsc::ServerParam::i().goalHalfWidth();
    const double field_half_length = rcsc::ServerParam::i().pitchHalfLength();

    const rcsc::Vector2D goal_center = rcsc::ServerParam::i().theirTeamGoalPos();
    const rcsc::Vector2D goal_left( goal_center.x, +goal_half_width );
    const rcsc::Vector2D goal_right( goal_center.x, -goal_half_width );

    const rcsc::Vector2D goal_center_left( field_half_length,
                                           (+ goal_half_width - 1.5) / 2.0 );

    const rcsc::Vector2D goal_center_right( field_half_length,
                                            (- goal_half_width + 1.5) / 2.0 );


    double center_angle = get_shoot_minimum_free_angle( wm,
                                                        goal_center,
                                                        valid_opponent_threshold,
                                                        pos );

    double left_angle   = get_shoot_minimum_free_angle( wm,
                                                    goal_left,
                                                    valid_opponent_threshold,
                                                    pos );

    double right_angle  = get_shoot_minimum_free_angle( wm,
                                                    goal_right,
                                                    valid_opponent_threshold,
                                                    pos );

    double center_left_angle  = get_shoot_minimum_free_angle
                                           ( wm,
                                             goal_center_left,
                                             valid_opponent_threshold,
                                             pos );

    double center_right_angle = get_shoot_minimum_free_angle
                                           ( wm,
                                             goal_center_right,
                                             valid_opponent_threshold,
                                             pos );
#if 0
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "get_shoot_free_angle_from_pos pos=(%.3f %.3f)",
                        pos.x, pos.y );
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "__ center_angle=%f", center_angle );
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "__ left_angle=%f", left_angle );
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "__ right_angle=%f", right_angle );
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "__ center_left_angle=%f", center_left_angle );
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "__ center_right_angle=%f", center_right_angle );
#endif
    return std::max( center_angle,
                     std::max( left_angle,
                               std::max( right_angle,
                                         std::max( center_left_angle,
                                                   center_right_angle ) ) ) );
}


/*-------------------------------------------------------------------*/
/*!

 */
double
DefaultFieldEvaluator::get_shoot_minimum_free_angle( const rcsc::PredictState & wm,
                                                     const rcsc::Vector2D & goal,
                                                     const long valid_opponent_threshold,
                                                     const rcsc::Vector2D & pos )
{
    rcsc::AngleDeg test_dir = (goal - pos).th();

    double shoot_course_cone = +360.0;

    const rcsc::PlayerPtrCont::const_iterator o_end = wm.opponentsFromSelf().end();
    for ( rcsc::PlayerPtrCont::const_iterator opp = wm.opponentsFromSelf().begin();
          opp != o_end;
          ++opp )
    {
        if ( (*opp)->posCount() > valid_opponent_threshold )
        {
            continue;
        }

        double controllable_dist;

        if ( (*opp)->goalie() )
        {
            controllable_dist = rcsc::ServerParam::i().catchAreaLength();
        }
        else
        {
            controllable_dist = (*opp)->playerTypePtr()->kickableArea();
        }

        rcsc::Vector2D relative_player = (*opp)->pos() - pos;

        double hide_angle_radian = ( std::asin
                                     ( std::min( controllable_dist
                                                 / relative_player.r(),
                                                 1.0 ) ) );
        double angle_diff = std::max( (relative_player.th() - test_dir).abs()
                                      - hide_angle_radian / M_PI * 180,
                                      0.0 );

        if ( shoot_course_cone > angle_diff )
        {
            shoot_course_cone = angle_diff;
        }
    }

    return shoot_course_cone;
}


bool
DefaultFieldEvaluator::is_free_situation_near_offside_line
                       ( const rcsc::PredictState & wm,
                         const rcsc::AbstractPlayerObject * from )
{
    const long VALID_PLAYER_ACCURACY = 20;
    const long VALID_OPPONENT_ACCURACY = VALID_PLAYER_ACCURACY;
    const double FREE_WIDTH = 1.75;

    if ( ! from
         || from->posCount() > VALID_PLAYER_ACCURACY
         || from->isGhost() )
    {
        return false;
    }

    if ( from->pos().x > wm.offsideLineX() )
    {
#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "checking push %d: offside position",
                            from->unum() );
#endif
        return false;
    }
    else if ( from->pos().x < wm.offsideLineX() - 10.0
              && from->pos().x < rcsc_ext::WorldModel_forwardLineX( wm ) - 10.0 )
    {
#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "checking push %d: far from offside line",
                            from->unum() );
#endif
        return false;
    }


    static const double WIDE_FREE_WIDTH = 10.0;

    if ( from->isSelf()
         || from->pos().x >= wm.offsideLineX() - 20.0 )
    {
        rcsc::AbstractPlayerCont opp_set;
        opp_set = wm.getPlayerCont( new rcsc::AndPlayerPredicate
                                    ( new rcsc::OpponentOrUnknownPlayerPredicate( wm.ourSide() ),
                                      new rcsc::CoordinateAccuratePlayerPredicate
                                      ( VALID_PLAYER_ACCURACY ),
                                      new rcsc::XCoordinateForwardPlayerPredicate
                                      ( from->pos().x ),
                                      new rcsc::XCoordinateBackwardPlayerPredicate
                                      ( wm.offsideLineX() + 1.0 ),
                                      new rcsc::YCoordinatePlusPlayerPredicate
                                      ( from->pos().y - WIDE_FREE_WIDTH ),
                                      new rcsc::YCoordinateMinusPlayerPredicate
                                      ( from->pos().y + WIDE_FREE_WIDTH ) ) );

        double free_length = rcsc_ext::AbstractPlayerCont_getMinimumEvaluation
                             ( opp_set,
                               new rcsc::XPosPlayerEvaluator )
                             - from->pos().x;

        if ( free_length >= 12.5 )
        {
#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                "checking push %d: first check ok",
                                from->unum() );
#endif
            return true;
        }
    }


    if ( from->pos().x <= wm.defenseLineX() + 10.0 )
    {
# ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "checking push %d: near our defense line",
                            from->unum() );
# endif
        return false;
    }

    if ( from->pos().x < rcsc_ext::WorldModel_forwardLineX( wm ) - 5.0 )
    {
# ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "checking push %d: too far from our forward line: x = %.2f, forward line = %.2f",
                            from->unum(),
                            from->pos().x, rcsc_ext::WorldModel_forwardLineX( wm ) );
# endif
        return false;
    }

    rcsc::PlayerCont::const_iterator o_end = wm.opponents().end();
    for ( rcsc::PlayerCont::const_iterator opp = wm.opponents().begin();
          opp != o_end;
          ++opp )
    {
        if ( (*opp).posCount() > VALID_OPPONENT_ACCURACY )
        {
            continue;
        }

        if ( (from->pos().x < (*opp).pos().x
              && (*opp).pos().x < from->pos().x + 5.0)
             && std::fabs( (*opp).pos().y - from->pos().y ) < FREE_WIDTH )
        {
# ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                "checking push %d: too narrow free width",
                                from->unum() );
# endif
            return false;
        }
    }

    return true;
}
