// -*-c++-*-

/*!
  \file pass_table.cpp
  \brief strict checking pass course table Source File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA, Hiroki Shimora

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "pass_table2009.h"

#include <rcsc/player/field_analyzer.h>

#include <rcsc/action/kick_table.h>
#include <rcsc/player/world_model.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/timer.h>

#include <algorithm>
#include <limits>
// #include <sstream>

#include <cmath>

#define DEBUG_PROFILE

// #define DEBUG_PRINT_SUCCESS_PASS
// #define DEBUG_PRINT_FAILED_PASS

// #define DEBUG_PRINT_VORONOI_DIAGRAM

// #define DEBUG_SELF_PASS
// #define DEBUG_DIRECT_PASS
// #define DEBUG_DIRECT_LEAD_PASS
// #define DEBUG_VORONOI_PASS
// #define DEBUG_VORONOI_LEAD_PASS

// #define DEBUG_RECEIVER
// #define DEBUG_CHECK_PASS
// #define DEBUG_CHECK_PASS_LEVEL_2

namespace rcsc {

namespace {

/*-------------------------------------------------------------------*/
/*!

 */
struct AbstractPlayerBallDistCmp {
    bool operator()( const AbstractPlayerObject * lhs,
                     const AbstractPlayerObject * rhs ) const
      {
          return lhs->distFromBall() < rhs->distFromBall();
      }
};

//
//const double pos_index_step_dist = 1.75;
const double pos_index_step_dist = 1.0;

/*-------------------------------------------------------------------*/
/*!

 */
inline
int
get_point_index( const Vector2D & pos )
{
    return static_cast< int >( rint( ( pos.x + 100.0 ) / pos_index_step_dist ) ) * 1000
        + static_cast< int >( rint( ( pos.y + 100.0 ) / pos_index_step_dist ) );

}

/*-------------------------------------------------------------------*/
/*!

 */
inline
const
AbstractPlayerObject *
get_nearest_player( const Vector2D & pos,
                    const AbstractPlayerCont & players )
{
    const AbstractPlayerObject * rval = static_cast< const AbstractPlayerObject * >( 0 );

    double min_dist2 = std::numeric_limits< double >::max();

    const AbstractPlayerCont::const_iterator p_end = players.end();
    for ( AbstractPlayerCont::const_iterator p = players.begin();
          p != p_end;
          ++p )
    {
        double d2 = (*p)->pos().dist2( pos );
        if ( d2 < min_dist2 )
        {
            rval = *p;
            min_dist2 = d2;
        }
    }

    return rval;
}

/*-------------------------------------------------------------------*/
/*!

 */
inline
void
get_nearest_players( const Vector2D & pos,
                     const AbstractPlayerCont & players,
                     AbstractPlayerCont * result )
{
    const AbstractPlayerObject * first = static_cast< const AbstractPlayerObject * >( 0 );
    const AbstractPlayerObject * second = static_cast< const AbstractPlayerObject * >( 0 );
    double first_dist2 = std::numeric_limits< double >::max();
    double second_dist2 = std::numeric_limits< double >::max();

    const AbstractPlayerCont::const_iterator p_end = players.end();
    for ( AbstractPlayerCont::const_iterator p = players.begin();
          p != p_end;
          ++p )
    {
        double d2 = (*p)->pos().dist2( pos );
        if ( d2 < second_dist2 )
        {
            second = *p;
            second_dist2 = d2;

            if ( d2 < first_dist2 )
            {
                std::swap( first, second );
                std::swap( first_dist2, second_dist2 );
            }

        }
    }

    result->clear();

    if ( first )
    {
        result->push_back( first );
    }

    if ( second )
    {
        result->push_back( second );
    }
}

}

/*-------------------------------------------------------------------*/
/*!

 */
PassTable2009::PassTable2009()
    : M_total_count( 0 )
{
    M_courses.reserve( 1024 );
}


/*-------------------------------------------------------------------*/
/*!

 */
PassTable2009 &
PassTable2009::instance()
{
    static PassTable2009 s_instance;
    return s_instance;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
PassTable2009::update( const WorldModel & wm )
{
    static GameTime s_update_time( 0, 0 );
    if ( s_update_time == wm.time() )
    {
        return;
    }
    s_update_time = wm.time();

    M_total_count = 0;
    M_courses.clear();
    M_receive_point_set.clear();

    if ( ! wm.self().isKickable() )
    {
        return;
    }

#ifdef DEBUG_PRINT_VORONOI_DIAGRAM
    {
        const VoronoiDiagram::Segment2DCont::const_iterator s_end = FieldAnalyzer::i().passVoronoiDiagram().resultSegments().end();
        for ( VoronoiDiagram::Segment2DCont::const_iterator s = FieldAnalyzer::i().passVoronoiDiagram().resultSegments().begin();
              s != s_end;
              ++s )
        {
            dlog.addLine( Logger::PASS,
                          s->origin(), s->terminal(),
                          "#0000ff" );
        }
    }
#endif

#ifdef DEBUG_PROFILE
    Timer timer;
#endif

    const AbstractPlayerObject * ball_holder = &(wm.self());

    AbstractPlayerCont teammates;
    teammates.reserve( wm.allTeammates().size() );

    const AbstractPlayerCont::const_iterator t_end = wm.allTeammates().end();
    for ( AbstractPlayerCont::const_iterator t = wm.allTeammates().begin();
          t != t_end;
          ++t )
    {
        if ( *t == ball_holder ) continue;
        if ( (*t)->isGhost() ) continue;
        if ( (*t)->unumCount() > 10 ) continue;
        if ( (*t)->unum() == Unum_Unknown ) continue;
        if ( (*t)->posCount() > 10 ) continue;
        if ( (*t)->isTackling() ) continue;
        if ( (*t)->goalie() && (*t)->pos().x < -22.0 ) continue;

        if ( (*t)->pos().x > wm.offsideLineX() )
        {
            // offside players are rejected.
            dlog.addText( Logger::PASS,
                          "PassTable2009 receiver %d (%.1f %.1f) is offside",
                          (*t)->unum(),
                          (*t)->pos().x, (*t)->pos().y );
            continue;
        }

        teammates.push_back( *t );
    }

    std::sort( teammates.begin(), teammates.end(), AbstractPlayerBallDistCmp() );

    //createSelfPass( wm, teammates );
    createDirectPass( wm, wm.ball().pos(), teammates );
    createVoronoiPass( wm,
                       ball_holder,
                       wm.ball().pos(),
                       teammates );

#ifdef DEBUG_PROFILE
    dlog.addText( Logger::PASS,
                  "PassTable2009::update() success_course_size=%d total=%d elapsed %f [ms]",
                  (int)M_courses.size(),
                  M_total_count,
                  timer.elapsedReal() );
//     std::cerr << "PassTable: total_count=" << M_total_count
//               << " success_count=" << M_courses.size()
//               << std::endl;
#endif
}

/*-------------------------------------------------------------------*/
/*!

 */
const
std::vector< CooperativeAction::Ptr > &
PassTable2009::getPassCandidates( const WorldModel & wm )
{
    update( wm );
    return M_courses;
}


/*-------------------------------------------------------------------*/
/*!

 */
void
PassTable2009::createDirectPass( const WorldModel & wm,
                                 const Vector2D & first_ball_pos,
                                 const AbstractPlayerCont & teammates )
{
    static const double MIN_DIRECT_PASS_DIST = ServerParam::i().defaultKickableArea() * 2.2;
    static const double MAX_DIRECT_PASS_DIST
        = 0.8 * inertia_final_distance( ServerParam::i().ballSpeedMax(),
                                        ServerParam::i().ballDecay() );
    static const double OUR_GOAL_DIST_THR2 = std::pow( 15.0, 2 );

#ifdef DEBUG_DIRECT_PASS
    dlog.addText( Logger::PASS,
                  "(createDirectPass) max distance=%.3f",
                  MAX_DIRECT_PASS_DIST );
#endif

    const ServerParam & SP = ServerParam::i();
    const double MAX_RECEIVE_BALL_SPEED = SP.ballSpeedMax() * std::pow( SP.ballDecay(), 5 );
    const Vector2D our_goal( -SP.pitchHalfLength(), 0.0 );
    const double max_ball_speed = ( wm.gameMode().type() == GameMode::PlayOn
                                    ? SP.ballSpeedMax()
                                    : wm.self().kickRate() * ServerParam::i().maxPower() );

    const AbstractPlayerCont::const_iterator t_end = teammates.end();
    for ( AbstractPlayerCont::const_iterator t = teammates.begin();
          t != t_end;
          ++t )
    {
        ++M_total_count;

#ifdef DEBUG_DIRECT_PASS
        dlog.addText( Logger::PASS,
                      "***** DirectPass: %d receiver %d (%.1f %.1f) *****",
                      M_total_count,
                      (*t)->unum(),
                      (*t)->pos().x, (*t)->pos().y );
#endif

        //
        // filtering
        //

        if ( wm.gameMode().type() == GameMode::GoalKick_
             && (*t)->pos().x < SP.ourPenaltyAreaLineX()
             && (*t)->pos().absY() < SP.penaltyAreaHalfWidth() )
        {
#ifdef DEBUG_DIRECT_PASS
            dlog.addText( Logger::PASS,
                          "%d xxx GoalKick. (%.1f %.1f) is in our penalty area.",
                          M_total_count,
                          (*t)->pos().x, (*t)->pos().y );
#endif
            continue;
        }

        if ( (*t)->pos().x < - SP.pitchHalfLength() + 5.0
             || SP.pitchHalfLength() - 1.0 < (*t)->pos().x
             || (*t)->pos().absY() > SP.pitchHalfWidth() - 3.0
             || (*t)->pos().dist2( our_goal ) < OUR_GOAL_DIST_THR2 )
        {
#ifdef DEBUG_DIRECT_PASS
            dlog.addText( Logger::PASS,
                          "%d: xxx out of pitch, or close to our goal",
                          M_total_count );
#endif
            continue;
        }

#if 0
        if ( (*t)->posCount() >= 4 )
        {
#ifdef DEBUG_DIRECT_PASS
            dlog.addText( Logger::PASS,
                          "%d: xxx low pos_count %d",
                          M_total_count, (*t)->posCount() );
#endif
            continue;
        }
#endif

        if ( (*t)->posCount() >= 2
             && ( (*t)->pos().x < -10.0
                  || (*t)->pos().x < first_ball_pos.x - 20.0 ) )
        {
#ifdef DEBUG_DIRECT_PASS
            dlog.addText( Logger::PASS,
                          "%d: xxx low pos_count %d and back",
                          M_total_count, (*t)->posCount() );
#endif
            continue;
        }

        //
        // calculate receive point
        //

        const PlayerType * ptype = (*t)->playerTypePtr();

        Vector2D receive_pos = (*t)->pos();
        if ( (*t)->velCount() <= 2 )
        {
            Vector2D vel = (*t)->vel();
            vel /= ptype->playerDecay();
            vel *= min_max( 1, (*t)->velCount(), 2 );
            receive_pos += vel;
        }

        const double ball_move_dist = first_ball_pos.dist( receive_pos );

        if ( ball_move_dist < MIN_DIRECT_PASS_DIST
             || MAX_DIRECT_PASS_DIST < ball_move_dist )
        {
#ifdef DEBUG_DIRECT_PASS
            dlog.addText( Logger::PASS,
                          "%d: xxx over distance range %.2f [%.2f, %.2f]",
                          M_total_count,
                          ball_move_dist,
                          MIN_DIRECT_PASS_DIST );
#endif
            continue;
        }

        //
        // calculate ball speed
        //

        const double max_receive_ball_speed
            = std::min( MAX_RECEIVE_BALL_SPEED,
                        (*t)->posCount() == 0
                        ? ptype->kickableArea() + ( SP.maxDashPower()
                                                    * ptype->dashPowerRate()
                                                    * ptype->effortMax() ) * 1.8
                        : ptype->kickableArea() + ( SP.maxDashPower()
                                                    * ptype->dashPowerRate()
                                                    * ptype->effortMax() ) * 1.5
                        );
        const double min_receive_ball_speed = std::min( ptype->realSpeedMax() * 0.9,
                                                        max_receive_ball_speed );

#ifdef DEBUG_DIRECT_PASS
        dlog.addText( Logger::PASS,
                      "%d: __ min_receive_ball_speed=%.3f max_receive_ball_speed=%.3f",
                      M_total_count,
                      min_receive_ball_speed, max_receive_ball_speed );
#endif

        double ball_first_speed = max_ball_speed;
        int ball_reach_step
            = static_cast< int >( std::ceil( calc_length_geom_series( ball_first_speed,
                                                                      ball_move_dist,
                                                                      SP.ballDecay() ) ) );
        double ball_end_speed = ball_first_speed * std::pow( SP.ballDecay(), ball_reach_step );

        if ( ball_end_speed < min_receive_ball_speed )
        {
#ifdef DEBUG_DIRECT_PASS
            dlog.addText( Logger::PASS,
                          "%d: xxx too small ball end_speed=%f",
                          M_total_count,
                          ball_end_speed );
#endif
            continue;
        }

        if ( ball_end_speed < max_receive_ball_speed )
        {
            ball_first_speed = calc_first_term_geom_series( ball_move_dist,
                                                            SP.ballDecay(),
                                                            ball_reach_step );
            ball_first_speed = std::min( SP.ballSpeedMax(), ball_first_speed );
#ifdef DEBUG_DIRECT_PASS
            dlog.addText( Logger::PASS,
                          "%d: __ (end_speed < max_recv_speed) ball_first_speed=%.3f end_speed=%.3f reach_step=%d",
                          M_total_count,
                          ball_first_speed, ball_end_speed, ball_reach_step );
#endif
        }
        else
        {
            ball_end_speed = max_receive_ball_speed;
            ball_first_speed = calc_first_term_geom_series_last( ball_end_speed,
                                                                 ball_move_dist,
                                                                 SP.ballDecay() );
            ball_reach_step
                = static_cast< int >( std::ceil( calc_length_geom_series( ball_first_speed,
                                                                          ball_move_dist,
                                                                          SP.ballDecay() ) ) );
#ifdef DEBUG_DIRECT_PASS
            dlog.addText( Logger::PASS,
                          "%d: __ (end_speed > max_recv_speed) ball_first_speed=%.3f ball_end_speed=%.3f reach_step=%d",
                          M_total_count,
                          ball_first_speed, ball_end_speed, ball_reach_step );
#endif
        }

        if ( ball_reach_step < 3 )
        {
            ball_reach_step = 3;
            ball_first_speed = calc_first_term_geom_series( ball_move_dist,
                                                            SP.ballDecay(),
                                                            ball_reach_step );
            ball_first_speed = std::min( max_ball_speed, ball_first_speed );
            ball_end_speed = ball_first_speed * std::pow( SP.ballDecay(), 3 );
#ifdef DEBUG_DIRECT_PASS
            dlog.addText( Logger::PASS,
                          "%d: __ (ball_reach_step < 3) ball_first_speed=%.3f ball_end_speed=%.3f reach_step=%d",
                          M_total_count,
                          ball_first_speed, ball_end_speed, ball_reach_step );
#endif
        }

        //
        // check opponent
        //

        if ( checkCourse( wm,
                          *t,
                          receive_pos,
                          ball_first_speed,
                          ball_reach_step ) )
        {
#ifdef DEBUG_PRINT_SUCCESS_PASS
            dlog.addText( Logger::PASS,
                          "--> %d/%d add candidate. (ball first_speed=%.3f end_speed=%.3f reach_step=%d)",
                          (int)M_courses.size() + 1, M_total_count,
                          ball_first_speed, ball_end_speed, ball_reach_step );
            dlog.addRect( Logger::PASS,
                          receive_pos.x - 0.1, receive_pos.y - 0.1,
                          0.2, 0.2,
                          "#ff0000" );
            char num[8]; snprintf( num, 8, "d%d", M_total_count );
            dlog.addMessage( Logger::PASS,
                             receive_pos, num );
#endif
            CooperativeAction::Ptr ptr( new CooperativeAction( CooperativeAction::ActionType::Pass,
                                                               receive_pos,
                                                               (*t)->unum(),
                                                               ball_first_speed,
                                                               0,
                                                               ball_reach_step,
                                                               false,
                                                               "direct0" ) );
            M_courses.push_back( ptr );
        }
#ifdef DEBUG_PRINT_FAILED_PASS
        else
        {
            dlog.addText( Logger::PASS,
                          "XXX %d failed. (%.1f %.1f) (ball first_speed=%.3f end_speed=%.3f reach_step=%d)",
                          M_total_count,
                          receive_pos.x, receive_pos.y,
                          ball_first_speed, ball_end_speed, ball_reach_step );
            dlog.addRect( Logger::PASS,
                          receive_pos.x - 0.1, receive_pos.y - 0.1,
                          0.2, 0.2,
                          "#00ff00" );
            char num[8]; snprintf( num, 8, "d%d", M_total_count );
            dlog.addMessage( Logger::PASS,
                             receive_pos, num );
        }
#endif

        M_receive_point_set.insert( get_point_index( receive_pos ) );

        //
        // create leading pass
        //

        createDirectLeadPass( wm, first_ball_pos, *t, receive_pos );

        //
        // TODO: create one step kick
        //

        //
        // TODO: direct pass to edges of player's kickable area, if receiver distance is small.
        //
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
void
PassTable2009::createDirectLeadPass( const WorldModel & wm,
                                     const Vector2D & first_ball_pos,
                                     const AbstractPlayerObject * receiver,
                                     const Vector2D & first_target_point )
{
    static const double min_length = 2.0;
    static const int max_division = 10;

    const ServerParam & SP = ServerParam::i();

    const double first_target_dist = first_target_point.dist( first_ball_pos );

    const PlayerType * ptype = receiver->playerTypePtr();

    const Vector2D mid_point = first_target_point * 0.6 + first_ball_pos * 0.4;
    const double mid_to_first_target_dist = mid_point.dist( first_target_point );

    if ( mid_to_first_target_dist < min_length * 2.0 )
    {
#ifdef DEBUG_DIRECT_LEAD_PASS
        dlog.addText( Logger::PASS,
                      "DirectLeadPass receiver %d (%.1f %.1f) too near",
                      receiver->unum(),
                      receiver->pos().x, receiver->pos().y,
                      mid_point.x, mid_point.y,
                      div );
#endif
        return;
    }

    const double max_ball_speed = ( wm.gameMode().type() == GameMode::PlayOn
                                    ? SP.ballSpeedMax()
                                    : wm.self().kickRate() * ServerParam::i().maxPower() );
    const double max_receive_ball_speed = std::min( SP.ballSpeedMax() * std::pow( SP.ballDecay(), 5 ),
                                                    receiver->posCount() == 0
                                                    ? ptype->kickableArea() + ( SP.maxDashPower()
                                                                                * ptype->dashPowerRate()
                                                                                * ptype->effortMax() ) * 1.8
                                                    : ptype->kickableArea() + ( SP.maxDashPower()
                                                                                * ptype->dashPowerRate()
                                                                                * ptype->effortMax() ) * 1.5
                                                    );

    const int div = std::min( max_division,
                              static_cast< int >( std::ceil( mid_to_first_target_dist / min_length ) ) );
    const double div_real = static_cast< double >( div );

#ifdef DEBUG_DIRECT_LEAD_PASS
    dlog.addText( Logger::PASS,
                  "DirectLeadPass receiver %d (%.1f %.1f)  mid=(%.1f %.1f) div=%d",
                  receiver->unum(),
                  receiver->pos().x, receiver->pos().y,
                  mid_point.x, mid_point.y,
                  div );
#endif

    for ( int d = 1; d <= div; ++d )
    {
        ++M_total_count;

        const Vector2D receive_pos
            = first_target_point * ( ( div - d ) / div_real )
            + mid_point * ( d / div_real );

        if ( wm.gameMode().type() == GameMode::GoalKick_
             && receive_pos.x < SP.ourPenaltyAreaLineX()
             && receive_pos.absY() < SP.penaltyAreaHalfWidth() )
        {
#ifdef DEBUG_DIRECT_LEAD_PASS
            dlog.addText( Logger::PASS,
                          "%d xxx DirectLeadPass continue. GoalKick. (%.1f %.1f) is in our penalty area.",
                          M_total_count,
                          receive_pos.x, receive_pos.y );
#endif
            continue;
        }

        const int pos_index = get_point_index( receive_pos );
        if ( M_receive_point_set.find( pos_index ) != M_receive_point_set.end() )
        {
#ifdef DEBUG_DIRECT_LEAD_PASS
            dlog.addText( Logger::PASS,
                          "%d xxx DirectLeadPass continue. (%.1f %.1f) index=%d already there.",
                          M_total_count,
                          receive_pos.x, receive_pos.y,
                          pos_index );
#endif
            continue;
        }

        const int receiver_reach_step
            = predictReceiverReachStep( wm,
                                        receiver,
                                        receive_pos,
                                        wm.ball().pos(),
                                        ( receive_pos - wm.ball().pos() ).th() );

        const double ball_move_dist = first_target_dist - mid_to_first_target_dist * ( d / div_real );

        int ball_reach_step = receiver_reach_step;
        double ball_first_speed = calc_first_term_geom_series( ball_move_dist,
                                                               SP.ballDecay(),
                                                               ball_reach_step );
        if ( ball_first_speed > max_ball_speed )
        {
            ball_first_speed = max_ball_speed;
            ball_reach_step
                = static_cast< int >( std::ceil( calc_length_geom_series( ball_first_speed,
                                                                          ball_move_dist,
                                                                          SP.ballDecay() ) ) );
#ifdef DEBUG_DIRECT_LEAD_PASS
            dlog.addText( Logger::PASS,
                          "%d __ update ball speed. ball_step=%d ball_first_speed=%.3f",
                          M_total_count,
                          ball_reach_step,
                          ball_first_speed );
#endif
        }

        double ball_end_speed = ball_first_speed * std::pow( SP.ballDecay(), ball_reach_step );

        if ( ball_end_speed > max_receive_ball_speed )
        {
            ball_end_speed = max_receive_ball_speed;
            ball_first_speed = calc_first_term_geom_series_last( ball_end_speed,
                                                                 ball_move_dist,
                                                                 SP.ballDecay() );
            ball_reach_step
                = static_cast< int >( std::ceil( calc_length_geom_series( ball_first_speed,
                                                                          ball_move_dist,
                                                                          SP.ballDecay() ) ) );
#ifdef DEBUG_DIRECT_LEAD_PASS
            dlog.addText( Logger::PASS,
                          "%d __ (end_speed > max_recv_speed) ball_first_speed=%.3f ball_end_speed=%.3f reach_step=%d",
                          M_total_count,
                          ball_first_speed, ball_end_speed, ball_reach_step );
#endif
        }

        if ( ball_reach_step < 4 )
        {
            ball_reach_step = 4;
            ball_first_speed = calc_first_term_geom_series( ball_move_dist,
                                                            SP.ballDecay(),
                                                            ball_reach_step );
            ball_first_speed = std::min( max_ball_speed, ball_first_speed );
            ball_end_speed = ball_first_speed * std::pow( SP.ballDecay(), 3 );
#ifdef DEBUG_DIRECT_PASS
            dlog.addText( Logger::PASS,
                          "%d: __ (ball_reach_step < 4) ball_first_speed=%.3f ball_end_speed=%.3f reach_step=%d",
                          M_total_count,
                          ball_first_speed, ball_end_speed, ball_reach_step );
#endif
        }

        if ( checkCourse( wm,
                          receiver,
                          receive_pos,
                          ball_first_speed,
                          ball_reach_step ) )
        {
#ifdef DEBUG_PRINT_SUCCESS_PASS
            dlog.addText( Logger::PASS,
                          "--> %d/%d add candidate. receiver=%d (%.1f %.1f)"
                          " receive_pos=(%.1f %.1f)"
                          " (ball first_speed=%.3f end_speed=%.3f reach_step=%d)"
                          " (receiver step=%d)",
                          (int)M_courses.size(), M_total_count,
                          receiver->unum(),
                          receiver->pos().x, receiver->pos().y,
                          receive_pos.x, receive_pos.y,
                          ball_first_speed,
                          ball_first_speed * std::pow( SP.ballDecay(), ball_reach_step ),
                          ball_reach_step,
                          receiver_reach_step );
            dlog.addRect( Logger::PASS,
                          receive_pos.x - 0.1, receive_pos.y - 0.1,
                          0.2, 0.2,
                          "#ff0000" );
            char id[8]; snprintf( id, 8, "dL%d", M_total_count );
            dlog.addMessage( Logger::PASS,
                             receive_pos, id );
#endif
            CooperativeAction::Ptr ptr( new CooperativeAction( CooperativeAction::ActionType::Pass,
                                                               receive_pos,
                                                               receiver->unum(),
                                                               ball_first_speed,
                                                               receiver_reach_step,
                                                               ball_reach_step,
                                                               false,
                                                               "directLead0" ) );
            M_courses.push_back( ptr );
            M_receive_point_set.insert( get_point_index( receive_pos ) );
        }
#ifdef DEBUG_PRINT_FAILED_PASS
        else
        {
            dlog.addText( Logger::PASS,
                          "XXX %d failed. (ball first_speed=%.3f end_speed=%.3f reach_step=%d)",
                          M_total_count,
                          ball_first_speed, ball_end_speed, ball_reach_step );
            dlog.addRect( Logger::PASS,
                          receive_pos.x - 0.1, receive_pos.y - 0.1,
                          0.2, 0.2,
                          "#00ff00" );
            char num[8]; snprintf( num, 8, "dL%d", M_total_count );
            dlog.addMessage( Logger::PASS,
                             receive_pos, num );
        }
#endif
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
void
PassTable2009::createVoronoiPass( const WorldModel & wm,
                                  const AbstractPlayerObject * ball_holder,
                                  const Vector2D & first_ball_pos,
                                  const AbstractPlayerCont & teammates )
{
    static const double MIN_VORONOI_PASS_DIST = ServerParam::i().defaultKickableArea() * 2.2;
    static const double MAX_VORONOI_PASS_DIST
        = 0.95 * inertia_final_distance( ServerParam::i().ballSpeedMax(),
                                         ServerParam::i().ballDecay() );

    static const double MIN_LENGTH = 2.0;
    static const unsigned int MAX_DIVISION = 10;
    static const unsigned int NUMBER_OF_CHECK_TEAMMATE = 2;
    //static const double ANGLE_THR = 60.0;

    static std::vector< Vector2D > points;
    points.clear();

    if ( teammates.empty() )
    {
        return;
    }

    const ServerParam & SP = ServerParam::i();
    const double max_ball_speed = ( wm.gameMode().type() == GameMode::PlayOn
                                    ? SP.ballSpeedMax()
                                    : wm.self().kickRate() * ServerParam::i().maxPower() );
    const double MAX_RECEIVE_BALL_SPEED = SP.ballSpeedMax() * std::pow( SP.ballDecay(), 5 );
    //const double MAX_RECEIVE_BALL_SPEED = 1.6; // 2009-06-27
    //const double MAX_RECEIVE_BALL_SPEED = 1.7; // 2009-06-29
    const double min_receive_ball_speed = 0.01;

    FieldAnalyzer::i().passVoronoiDiagram().getPointsOnSegments( MIN_LENGTH,
                                                                 MAX_DIVISION,
                                                                 &points );

#ifdef DEBUG_VORONOI_PASS
    dlog.addText( Logger::PASS,
                  "(createVoronoiPass) point size = %d",
                  (int)points.size() );
    dlog.addText( Logger::PASS,
                  "(createVoronoiPass) distance range [%.3f, %.3f]",
                  MIN_VORONOI_PASS_DIST, MAX_VORONOI_PASS_DIST );
#endif

    AbstractPlayerCont sorted_teammates;

    //
    // for all points on voronoi diagram
    //
    const std::vector< Vector2D >::const_iterator p_end = points.end();
    for ( std::vector< Vector2D >::const_iterator p = points.begin();
          p != p_end;
          ++p )
    {
        ++M_total_count;

        const double ball_move_dist = first_ball_pos.dist( *p );

#ifdef DEBUG_VORONOI_PASS
        dlog.addText( Logger::PASS,
                      "***** VoronoiPass: %d (%.1f %.1f) ball_move=%.1f *****",
                      M_total_count,
                      p->x, p->y, ball_move_dist );
#endif
        if ( ball_move_dist < MIN_VORONOI_PASS_DIST
             || MAX_VORONOI_PASS_DIST < ball_move_dist )
        {
#ifdef DEBUG_VORONOI_PASS
            dlog.addText( Logger::PASS,
                          "%d xxx VoronoiPass continue. over the distance range. ball_move_dist=%.3f",
                          M_total_count,
                          ball_move_dist );
#endif
            continue;
        }

        if ( wm.gameMode().type() == GameMode::GoalKick_
             && p->x < SP.ourPenaltyAreaLineX()
             && p->absY() < SP.penaltyAreaHalfWidth() )
        {
#ifdef DEBUG_VORONOI_PASS
            dlog.addText( Logger::PASS,
                          "%d xxx VoronoiPass. GoalKick. (%.1f %.1f) is in our penalty area.",
                          M_total_count,
                          p->x, p->y );
#endif
            continue;
        }

        const int pos_index = get_point_index( *p );
        if ( M_receive_point_set.find( pos_index ) != M_receive_point_set.end() )
        {
#ifdef DEBUG_VORONOI_PASS
            dlog.addText( Logger::PASS,
                          "%d xxx VoronoiPass continue. index=%d already there.",
                          M_total_count,
                          pos_index );
#endif
            continue;
        }

        //
        // filter candidate points.
        // TODO: more filtering.
        //
        if ( p->x < -20.0
             || p->x < first_ball_pos.x - 30.0
             //&& std::fabs( first_ball_pos.y - p->y ) / ( first_ball_pos.x - p->x ) < ANGLE_THR
             )
        {
#ifdef DEBUG_VORONOI_PASS
            dlog.addText( Logger::PASS,
                          "%d cccc VoronoiPass xxx over the position range",
                          M_total_count );
#endif
            continue;
        }

        {
            int max_count = 100, ave_count = 100;
            wm.dirRangeCount( ( *p - wm.self().pos() ).th(), 20.0,
                              &max_count, NULL, &ave_count );
            if ( max_count >= 10 || ave_count >= 4 )
            {
#ifdef DEBUG_VORONOI_PASS
                dlog.addText( Logger::PASS,
                              "%d cccc VoronoiPass low dir accuracy. max=%d ave=%d",
                              M_total_count, max_count, ave_count );
#endif
                continue;
            }
        }

        //
        // get receiver candidates.
        //
        get_nearest_players( *p, teammates, &sorted_teammates );

        if ( ball_holder->pos().dist2( *p ) < sorted_teammates.front()->pos().dist2( *p ) )
        {
#ifdef DEBUG_VORONOI_PASS
            dlog.addText( Logger::PASS,
                          "%d cccc VoronoiPass continue. ball holder is nearest to the target point.",
                          M_total_count );
#endif
            continue;
        }

        //
        // decide max ball reach step.
        //
        double slowest_ball_first_speed = calc_first_term_geom_series_last( min_receive_ball_speed,
                                                                            ball_move_dist,
                                                                            SP.ballDecay() );
        if ( slowest_ball_first_speed > max_ball_speed )
        {
#ifdef DEBUG_VORONOI_PASS
            dlog.addText( Logger::PASS,
                          "%d cccc VoronoiPass continue. slowest_ball_first_speed %.3f > ball_speed_max",
                          min_receive_ball_speed );
#endif
            continue;
        }

        int slowest_ball_reach_step
            = static_cast< int >( std::ceil( calc_length_geom_series( slowest_ball_first_speed,
                                                                      ball_move_dist,
                                                                      SP.ballDecay() ) ) );

        // **************************************************
        // TODO: check max ball velocity by one step kick.
        // if player cannot kick the ball in one step, the penalty should be added to ball_reach_step.
        // **************************************************


#ifdef DEBUG_VORONOI_PASS
        dlog.addText( Logger::PASS,
                      "__ %d ball_move=%.3f slowest_ball_reach_step=%d slowest_speed(%.3f -> %.3f)",
                      M_total_count,
                      ball_move_dist,
                      slowest_ball_reach_step,
                      slowest_ball_first_speed, min_receive_ball_speed );
#endif

        //
        // get the fastest receiver.
        //

        const AbstractPlayerObject * best_receiver = static_cast< const AbstractPlayerObject *>( 0 );
        int fastest_receive_step = -1;

        const AbstractPlayerCont::const_iterator t_end = sorted_teammates.end();
        unsigned int n_checked = 0;
        for ( AbstractPlayerCont::const_iterator t = sorted_teammates.begin();
              t != t_end && n_checked < NUMBER_OF_CHECK_TEAMMATE;
              ++t, ++n_checked )
        {
#if 0
            if ( (*t)->posCount() >= 3 )
            {
                continue;
            }
#endif

            if ( (*t)->pos().x > p->x + 15.0
                 || ( (*t)->pos().x > p->x
                      && (*t)->pos().dist2( *p ) > 15.0 * 15.0 ) )
            {
                continue;
            }

            // 1 step has to be added to teammate_reach_step,
            // because receiver can start to intercept after recognizing the ball velocity.
            int teammate_reach_step
                = predictReceiverReachStep( wm, *t, *p, wm.ball().pos(),
                                            ( *p - wm.ball().pos() ).th() );

            // if ( p->x < wm.offsideLineX() + 3.0 ) teammate_reach_step += 1;
            // 2009-06-29
            teammate_reach_step += std::min( 2,
                                             static_cast< int >( rint( teammate_reach_step * 0.3 ) ) );

            if ( (*t)->pos().x > p->x + 3.0 ) teammate_reach_step += 1;
            if ( (*t)->pos().x > p->x + 5.0 ) teammate_reach_step += 1;
            if ( (*t)->pos().x > p->x + 7.0 ) teammate_reach_step += 1;

            teammate_reach_step = std::max( 3, teammate_reach_step );

#ifdef DEBUG_VORONOI_PASS
            dlog.addText( Logger::PASS,
                          "____ receiver %d (%.1f %.1f) reach_step=%d",
                          (*t)->unum(),
                          (*t)->pos().x, (*t)->pos().y,
                          teammate_reach_step );
#endif

            if ( teammate_reach_step <= slowest_ball_reach_step )
            {
                if ( ! best_receiver
                     || teammate_reach_step < fastest_receive_step )
                {
                    best_receiver = (*t);
                    fastest_receive_step = teammate_reach_step;
                }
            }
        }

        M_receive_point_set.insert( pos_index );

        if ( best_receiver )
        {
            //
            // create leading pass
            //
            createVoronoiLeadPass( wm, ball_holder, first_ball_pos, best_receiver, *p );

            if ( p->x < 0.0
                 && p->x < best_receiver->pos().x - 5.0 )
            {
#ifdef DEBUG_VORONOI_PASS
                dlog.addText( Logger::PASS,
                              "%d cccc VoronoiPass (%.1f %.1f) < receiver_pos(%.1f %.1f)",
                              M_total_count,
                              p->x, p->y,
                              best_receiver->pos().x, best_receiver->pos().y );
#endif
                continue;
            }

            double ball_first_speed = calc_first_term_geom_series( ball_move_dist,
                                                                   SP.ballDecay(),
                                                                   fastest_receive_step );
            int ball_reach_step = fastest_receive_step;
            if ( ball_first_speed > max_ball_speed )
            {
                ball_first_speed = max_ball_speed;
                ball_reach_step
                    = static_cast< int >( std::ceil( calc_length_geom_series( ball_first_speed,
                                                                              ball_move_dist,
                                                                              SP.ballDecay() ) ) );
#ifdef DEBUG_VORONOI_PASS
                dlog.addText( Logger::PASS,
                              "%d: __ (first_speed > max_speed) ball_first_speed=%.3f reach_step=%d",
                              M_total_count,
                              ball_first_speed, ball_reach_step );
#endif
            }

            const PlayerType * ptype = best_receiver->playerTypePtr();
            const double max_receive_ball_speed
                = std::min( MAX_RECEIVE_BALL_SPEED,
                            best_receiver->posCount() == 0
                            ? ptype->kickableArea() + ( SP.maxDashPower()
                                                        * ptype->dashPowerRate()
                                                        //* ptype->effortMax() ) * 1.5 // 2009-06-27
                                                        * ptype->effortMax() ) * 1.8 // 2009-06-29
                            : ptype->kickableArea() + ( SP.maxDashPower()
                                                        * ptype->dashPowerRate()
                                                        * ptype->effortMax() ) * 1.5
                            );

            double ball_end_speed = ball_first_speed * std::pow( SP.ballDecay(), ball_reach_step );
            if ( ball_end_speed > max_receive_ball_speed )
            {
                 ball_end_speed = max_receive_ball_speed;
                 ball_first_speed = calc_first_term_geom_series_last( ball_end_speed,
                                                                      ball_move_dist,
                                                                      SP.ballDecay() );
                 ball_reach_step
                     = static_cast< int >( std::ceil( calc_length_geom_series( ball_first_speed,
                                                                               ball_move_dist,
                                                                               SP.ballDecay() ) ) );
#ifdef DEBUG_VORONOI_PASS
                 dlog.addText( Logger::PASS,
                               "%d: __ (end_speed > max_recv_speed) ball_first_speed=%.3f ball_end_speed=%.3f reach_step=%d",
                               M_total_count,
                               ball_first_speed, ball_end_speed, ball_reach_step );
#endif
            }

            if ( checkCourse( wm,
                              best_receiver,
                              *p,
                              ball_first_speed,
                              ball_reach_step ) )
            {
#ifdef DEBUG_PRINT_SUCCESS_PASS
                dlog.addText( Logger::PASS,
                              "--> %d/%d add candidate. receiver %d (%.1f %.1f) ball_first_speed=%.3f step=%d",
                              (int)M_courses.size() + 1, M_total_count,
                              best_receiver->unum(),
                              best_receiver->pos().x, best_receiver->pos().y,
                              ball_first_speed,
                              ball_reach_step );
                dlog.addRect( Logger::PASS,
                              p->x - 0.1, p->y - 0.1, 0.2, 0.2,
                              "#ff0000" );
                char id[8]; snprintf( id, 8, "v%d", M_total_count );
                dlog.addMessage( Logger::PASS,
                                 *p, id );
#endif
                CooperativeAction::Ptr ptr( new CooperativeAction( CooperativeAction::ActionType::Pass,
                                                                   *p,
                                                                   best_receiver->unum(),
                                                                   ball_first_speed,
                                                                   fastest_receive_step,
                                                                   ball_reach_step,
                                                                   false,
                                                                   "voronoi0" ) );


                M_courses.push_back( ptr );
            }
#ifdef DEBUG_PRINT_FAILED_PASS
            else
            {
                dlog.addText( Logger::PASS,
                              "XXX %d failed. (%.1f %.1f) ball first_speed=%.3f reach_step=%d",
                              M_total_count,
                              p->x, p->y,
                              ball_first_speed, fastest_receive_step );
                dlog.addRect( Logger::PASS,
                              p->x - 0.1, p->y - 0.1,
                              0.2, 0.2,
                              "#00ff00" );
                char num[8]; snprintf( num, 8, "v%d", M_total_count );
                dlog.addMessage( Logger::PASS,
                                 *p, num );
            }
#endif

        }
// #ifdef DEBUG_PRINT_FAILED_PASS
//         else
//         {
//             dlog.addText( Logger::PASS,
//                           "XXX %d failed. (%.1f %.1f) no receiver",
//                           M_total_count,
//                           p->x, p->y );
//             dlog.addRect( Logger::PASS,
//                           p->x - 0.1, p->y - 0.1,
//                           0.2, 0.2,
//                           "#00ff00" );
//             char num[8]; snprintf( num, 8, "v%d", M_total_count );
//             dlog.addMessage( Logger::PASS,
//                              *p, num );
//         }
// #endif
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
void
PassTable2009::createVoronoiLeadPass( const WorldModel & wm,
                                      const AbstractPlayerObject * ball_holder,
                                      const Vector2D & first_ball_pos,
                                      const AbstractPlayerObject * receiver,
                                      const Vector2D & first_target_point )
{
    static const double MAX_VORONOI_PASS_DIST
        = 0.95 * inertia_final_distance( ServerParam::i().ballSpeedMax(),
                                         ServerParam::i().ballDecay() );

    static const double min_length = 2.0;
    static const int max_division = 10;

    const Vector2D receiver_inertia_pos = receiver->inertiaFinalPoint();
    const double receiver_move_dist = receiver_inertia_pos.dist( first_target_point );

    const ServerParam & SP = ServerParam::i();
    const double max_ball_speed = ( wm.gameMode().type() == GameMode::PlayOn
                                    ? SP.ballSpeedMax()
                                    : wm.self().kickRate() * ServerParam::i().maxPower() );

    const PlayerType * ptype = receiver->playerTypePtr();
    const double max_receive_ball_speed
        = std::min( //1.7, // 2009-06-29
                    //1.6, // 2009-06-27
                    SP.ballSpeedMax() * std::pow( SP.ballDecay(), 5 ),
                    receiver->posCount() == 0
                    ? ptype->kickableArea() + ( SP.maxDashPower()
                                                * ptype->dashPowerRate()
                                                //* ptype->effortMax() ) * 1.5 // 2009-06-27
                                                * ptype->effortMax() ) * 1.8 // 2009-06-29
                    : ptype->kickableArea() + ( SP.maxDashPower()
                                                * ptype->dashPowerRate()
                                                * ptype->effortMax() ) * 1.5
                    );

    const int div = std::min( max_division,
                              static_cast< int >( std::ceil( receiver_move_dist / min_length ) ) );
    const double div_real = static_cast< double >( div );

#ifdef DEBUG_VORONOI_LEAD_PASS
    dlog.addText( Logger::PASS,
                  "(createVoronoiLeadPass) receiver %d (%.1f %.1f)->(%.1f %.1f)"
                  " first_target=(%.1f %.1f)"
                  " receiver_move_dist=%.1f"
                  " div=%d",
                  M_total_count,
                  receiver->unum(),
                  receiver->pos().x, receiver->pos().y,
                  receiver_inertia_pos.x, receiver_inertia_pos.y,
                  first_target_point.x, first_target_point.y,
                  receiver_move_dist,
                  div );
#endif


    if ( div >= 2 )
    {
        for ( int d = 1; d < div; ++d )
        {
            ++M_total_count;

            const Vector2D receive_pos
                = receiver_inertia_pos * ( d / div_real )
                + first_target_point * ( ( div - d ) / div_real );
#ifdef DEBUG_VORONOI_LEAD_PASS
            dlog.addText( Logger::PASS,
                          "***** VoronoiLeadPass %d recv_pos=(%.1f %.1f)"
                          " d=%d, div=%d, max_recv_speed=%.3f",
                          M_total_count,
                          receive_pos.x, receive_pos.y,
                          d, div,
                          max_receive_ball_speed );
#endif

            if ( wm.gameMode().type() == GameMode::GoalKick_
                 && receive_pos.x < SP.ourPenaltyAreaLineX()
                 && receive_pos.absY() < SP.penaltyAreaHalfWidth() )
            {
#ifdef DEBUG_VORONOI_PASS
                dlog.addText( Logger::PASS,
                              "%d xxx VoronoiLeadPass. GoalKick. (%.1f %.1f) is in our penalty area.",
                              M_total_count,
                              receive_pos.x, receive_pos.y );
#endif
                continue;
            }

            if ( receive_pos.x < 0.0
                 && receive_pos.x < receiver->pos().x - 5.0 )
            {
#ifdef DEBUG_VORONOI_PASS
                dlog.addText( Logger::PASS,
                              "%d cccc VoronoiLeadPass continue. (%.1f %.1f) < receiver_pos(%.1f %.1f)",
                              M_total_count,
                              receive_pos.x, receive_pos.y,
                              receiver->pos().x, receiver->pos().y );
#endif
                continue;
            }

            if ( ball_holder->pos().dist2( receive_pos ) < receiver->pos().dist2( receive_pos ) )
            {
#ifdef DEBUG_VORONOI_PASS
                dlog.addText( Logger::PASS,
                              "%d cccc VoronoiLeadPass continue. (%.1f %.1f) ball holder is nearest to the target point.",
                              M_total_count,
                              receive_pos.x, receive_pos.y );
#endif
                continue;
            }

            const double ball_move_dist = first_ball_pos.dist( receive_pos );
            if ( ball_move_dist > MAX_VORONOI_PASS_DIST )
            {
#ifdef DEBUG_VORONOI_LEAD_PASS
                dlog.addText( Logger::PASS,
                              "%d cccc VoronoiLeadPass continue. (%.1f %.1f) over dist %.1f > max_dist=%.1f",
                              M_total_count,
                              receive_pos.x, receive_pos.y,
                              ball_move_dist, MAX_VORONOI_PASS_DIST );
#endif
                continue;
            }

            const int pos_index = get_point_index( receive_pos );
            if ( M_receive_point_set.find( pos_index ) != M_receive_point_set.end() )
            {
#ifdef DEBUG_VORONOI_LEAD_PASS
                dlog.addText( Logger::PASS,
                              "%d VoronoiLeadPass continue. (%.1f %.1f) index=%d already there.",
                              M_total_count,
                              receive_pos.x, receive_pos.y,
                              pos_index );
#endif
                continue;
            }

            {
                int max_count = 100, ave_count = 100;
                wm.dirRangeCount( ( receive_pos - wm.self().pos() ).th(), 20.0,
                                  &max_count, NULL, &ave_count );
                if ( max_count >= 10 || ave_count >= 4 )
                {
#ifdef DEBUG_VORONOI_LEAD_PASS
                    dlog.addText( Logger::PASS,
                                  "%d cccc VoronoiLeadPass low dir accuracy. max=%d ave=%d",
                                  M_total_count, max_count, ave_count );
#endif
                    continue;
                }
            }

            int receiver_reach_step
                = predictReceiverReachStep( wm,
                                            receiver,
                                            receive_pos, wm.ball().pos(),
                                            ( receive_pos - wm.ball().pos() ).th() );

//             if ( receive_pos.x < wm.offsideLineX() - 3.0
//                  && receive_pos.x < 36.0 )
//             {
//                 receiver_reach_step += 1;
//             }
            // 2009-06-29
            receiver_reach_step += std::min( 2,
                                             static_cast< int >( rint( receiver_reach_step * 0.3 ) ) );

            if ( receiver->pos().x > receive_pos.x + 3.0 ) receiver_reach_step += 1;
            if ( receiver->pos().x > receive_pos.x + 5.0 ) receiver_reach_step += 1;
            if ( receiver->pos().x > receive_pos.x + 7.0 ) receiver_reach_step += 1;

            receiver_reach_step = std::max( 3, receiver_reach_step );

            int ball_reach_step = receiver_reach_step;
            double ball_first_speed = calc_first_term_geom_series( ball_move_dist,
                                                                   SP.ballDecay(),
                                                                   receiver_reach_step );
            if ( ball_first_speed > max_ball_speed )
            {
                ball_first_speed = max_ball_speed;
                ball_reach_step
                    = static_cast< int >( std::ceil( calc_length_geom_series( ball_first_speed,
                                                                              ball_move_dist,
                                                                              SP.ballDecay() ) ) );
#ifdef DEBUG_VORONOI_LEAD_PASS
                dlog.addText( Logger::PASS,
                              "%d: __ (first_speed > max_speed) ball_first_speed=%.3f reach_step=%d",
                              M_total_count,
                              ball_first_speed, ball_reach_step );
#endif
            }

            double ball_end_speed = ball_first_speed * std::pow( SP.ballDecay(), ball_reach_step );
            if ( ball_end_speed > max_receive_ball_speed )
            {
                 ball_end_speed = max_receive_ball_speed;
                 ball_first_speed = calc_first_term_geom_series_last( ball_end_speed,
                                                                      ball_move_dist,
                                                                      SP.ballDecay() );
                 ball_reach_step
                     = static_cast< int >( std::ceil( calc_length_geom_series( ball_first_speed,
                                                                               ball_move_dist,
                                                                               SP.ballDecay() ) ) );
#ifdef DEBUG_VORONOI_LEAD_PASS
                 dlog.addText( Logger::PASS,
                               "%d: __ (end_speed > max_recv_speed) ball_first_speed=%.3f ball_end_speed=%.3f reach_step=%d",
                               M_total_count,
                               ball_first_speed, ball_end_speed, ball_reach_step );
#endif
            }


#ifdef DEBUG_PRINT_SUCCESS_PASS
            dlog.addText( Logger::PASS,
                          "%d (%.1f %.1f) receiver_reach_step=%d"
                          " (ball first_speed=%.3f reach_step=%d)",
                          M_total_count,
                          receive_pos.x, receive_pos.y,
                          receiver_reach_step,
                          ball_first_speed, ball_reach_step );
#endif

            if ( checkCourse( wm,
                              receiver,
                              receive_pos,
                              ball_first_speed,
                              ball_reach_step ) )
            {
#ifdef DEBUG_PRINT_SUCCESS_PASS
                dlog.addText( Logger::PASS,
                              "--> %d/%d add candidate. receiver=%d (%.1f %.1f)"
                              " receive_pos=(%.1f %.1f)"
                              " (ball first_speed=%.3f end_speed=%.3f reach_step=%d)",
                              (int)M_courses.size() + 1, M_total_count,
                              receiver->unum(),
                              receiver->pos().x, receiver->pos().y,
                              receive_pos.x, receive_pos.y,
                              ball_first_speed,
                              ball_first_speed * std::pow( SP.ballDecay(), receiver_reach_step ),
                              receiver_reach_step );
                dlog.addRect( Logger::PASS,
                              receive_pos.x - 0.1, receive_pos.y - 0.1,
                              0.2, 0.2,
                              "#ff0000" );
                char id[8]; snprintf( id, 8, "vL%d", M_total_count );
                dlog.addMessage( Logger::PASS,
                                 receive_pos, id );
#endif
                CooperativeAction::Ptr ptr( new CooperativeAction( CooperativeAction::ActionType::Pass,
                                                                   receive_pos,
                                                                   receiver->unum(),
                                                                   ball_first_speed,
                                                                   receiver_reach_step,
                                                                   ball_reach_step,
                                                                   false,
                                                                   "voronoiLead0" ) );
                M_courses.push_back( ptr );
            }
#ifdef DEBUG_PRINT_FAILED_PASS
            else
            {
                dlog.addText( Logger::PASS,
                              "XXX %d failed. receiver_reach_step=%d (ball first_speed=%.3f reach_step=%d)",
                              M_total_count,
                              receiver_reach_step,
                              ball_first_speed, ball_reach_step );
                dlog.addRect( Logger::PASS,
                              receive_pos.x - 0.1, receive_pos.y - 0.1,
                              0.2, 0.2,
                              "#00ff00" );
                char num[8]; snprintf( num, 8, "vL%d", M_total_count );
                dlog.addMessage( Logger::PASS,
                                 receive_pos, num );
            }
#endif
            M_receive_point_set.insert( pos_index );
        }
    }

    //     if ( receiver_inertia_pos.x < wm.defenseLineX()
    //          || receiver_inertia_pos.x < wm.offsideLineX() - 10.0 )
    //     {
    //         return;
    //     }

    //     if ( ( first_target_point - receiver_inertia_pos ).th().abs() > 80.0 )
    //     {
    //         return;
    //     }

}

/*-------------------------------------------------------------------*/
/*!

 */
void
PassTable2009::createSelfPass( const WorldModel & wm,
                               const AbstractPlayerCont & teammates )
{
    static const int angle_div = 24;
    static const double angle_step = 360.0 / angle_div;
    static const double dist_step = 2.0;

    static const double max_x = ServerParam::i().pitchHalfLength() - 3.0;
    static const double max_y = ServerParam::i().pitchHalfWidth() - 3.0;

    static const double max_back_dist = 6.1;
    static const double max_side_dist = 8.1;
    static const double max_forward_dist = 24.1;

    static const double min_stamina = ServerParam::i().recoverDecThrValue() + ServerParam::i().maxDashPower();

    const ServerParam & SP = ServerParam::i();

    if ( wm.gameMode().type() != GameMode::PlayOn )
    {
#ifdef DEBUG_SELF_PASS
        dlog.addText( Logger::PASS,
                      "(createSelfPass) no playon " );
#endif
        return;
    }

    if ( ! wm.self().isKickable() )
    {
#ifdef DEBUG_SELF_PASS
        dlog.addText( Logger::PASS,
                      "(createSelfPass) no kickable " );
#endif
        return;
    }

    if ( wm.self().stamina() < min_stamina + SP.maxDashPower() * 3.0 )
    {
#ifdef DEBUG_SELF_PASS
        dlog.addText( Logger::PASS,
                      "(createSelfPass) insufficient stamina." );
#endif
        return;
    }

    const Vector2D my_next = wm.self().pos() + wm.self().vel();
    const Vector2D my_pos = wm.self().inertiaFinalPoint();
    const double control_area = wm.self().playerType().kickableArea() * 0.7;
    const double collide_dist2 = std::pow( wm.self().playerType().playerSize() + SP.ballSize(), 2 );

// #ifdef DEBUG_SELF_PASS
//     char id[8];
//     snprintf( id, 8, "%d", wm.self().unum() );
// #endif

#ifdef DEBUG_PROFILE
    rcsc::Timer timer;
#endif

    int count = 0;
    AngleDeg angle = wm.self().body() - 180.0;
    for ( int a = 0; a < angle_div; ++a, angle += angle_step )
    {
        const Vector2D unit_vec = Vector2D::polar2vector( 1.0, angle );
        const double max_dist = ( angle.abs() > 90.0
                                  ? max_back_dist - ( ( max_back_dist - max_side_dist )
                                                      * ( 1.0 - ( angle.abs() - 90.0 ) / 90.0 ) )
                                  : max_side_dist + ( ( max_forward_dist - max_side_dist )
                                                      * ( 1.0 - angle.abs() / 90.0 ) ) );

        for ( double dist = 3.0; dist < max_dist; dist += dist_step )
        {
            ++count;
            ++M_total_count;

            const Vector2D receive_pos = my_pos + ( unit_vec * dist );

            if ( receive_pos.absX() > max_x
                 || receive_pos.absY() > max_y )
            {
                break;
            }

            //
            // check my reach step & stamina
            //

            double my_stamina = wm.self().stamina();
            const int my_reach_step
                = FieldAnalyzer::predict_self_reach_cycle( wm, receive_pos, control_area, 1, true, &my_stamina );

            if ( my_stamina < min_stamina )
            {
#ifdef DEBUG_SELF_PASS
                dlog.addText( Logger::PASS,
                              "%d: SelfPass xxx (%.1f %.1f) break. insufficient stamina. my_cycle=%d stamina=%.1f",
                              M_total_count,
                              receive_pos.x, receive_pos.y,
                              my_reach_step, my_stamina );
#endif
                break;
            }

            //
            // check first ball speed
            //

            const Vector2D max_ball_vel = KickTable::calc_max_velocity( ( receive_pos - wm.ball().pos() ).th(),
                                                                        wm.self().kickRate(),
                                                                        wm.ball().vel() );
            double ball_first_speed = calc_first_term_geom_series( ( receive_pos - wm.ball().pos() ).r(),
                                                                   SP.ballDecay(),
                                                                   my_reach_step );
            if ( std::pow( ball_first_speed, 2 ) > max_ball_vel.r2() )
            {
#ifdef DEBUG_SELF_PASS
                dlog.addText( Logger::PASS,
                              "%d: SelfPass continue (%.1f %.1f) could not kick by 1 step. requied_speed=%.3f",
                              M_total_count,
                              receive_pos.x, receive_pos.y,
                              ball_first_speed );
#endif
                // TODO: predict 2 step kick case.
                continue;
            }

            const Vector2D ball_first_vel = ( receive_pos - wm.ball().pos() ).setLengthVector( ball_first_speed );
            if ( my_next.dist2( wm.ball().pos() + ball_first_vel ) < collide_dist2 )
            {
#ifdef DEBUG_SELF_PASS
                dlog.addText( Logger::PASS,
                              "%d: SelfPass continue (%.1f %.1f) collision will occur. ball_first_vel=(%.3f %.3f)",
                              M_total_count,
                              receive_pos.x, receive_pos.y,
                              ball_first_vel.x, ball_first_vel.y );
#endif
                // TODO: predict 2 step kick case.
                continue;
            }

            //
            // check other receiver
            //

            const AbstractPlayerObject * nearest_teammate = get_nearest_player( receive_pos, teammates );
            if ( nearest_teammate )
            {
                int teammate_reach_step
                    = predictReceiverReachStep( wm,
                                                nearest_teammate,
                                                receive_pos,
                                                wm.ball().pos(),
                                                ( receive_pos - wm.ball().pos() ).th() );

                if ( teammate_reach_step < my_reach_step * 0.9  )
                {
#ifdef DEBUG_SELF_PASS
                    dlog.addText( Logger::PASS,
                                  "%d: SelfPass continue (%.1f %.1f) exist faster teammate. my_cycle=%d teammate_cycle=%d",
                                  M_total_count,
                                  receive_pos.x, receive_pos.y,
                                  my_reach_step, teammate_reach_step );
#endif
                    continue;
                }
            }

            if ( checkCourse( wm,
                              &(wm.self()),
                              receive_pos,
                              ball_first_speed,
                              my_reach_step ) )
            {
#ifdef DEBUG_PRINT_SUCCESS_PASS
                dlog.addText( Logger::PASS,
                              "%d/%d add candidate (%.1f %.1f) my_cycle=%d stamina=%.1f ball_speed=%.2f",
                              (int)M_courses.size() + 1, M_total_count,
                              receive_pos.x, receive_pos.y,
                              my_reach_step, my_stamina,
                              ball_first_speed );
                dlog.addRect( Logger::PASS,
                              receive_pos.x - 0.1, receive_pos.y - 0.1,
                              0.2, 0.2,
                              "#ff0000" );
                char num[8]; snprintf( num, 8, "s%d", M_total_count );
                dlog.addMessage( Logger::PASS,
                                 receive_pos, num );
#endif
                CooperativeAction::Ptr ptr( new CooperativeAction( CooperativeAction::ActionType::Pass,
                                                                   receive_pos,
                                                                   wm.self().unum(),
                                                                   ball_first_speed,
                                                                   my_reach_step,
                                                                   my_reach_step, // ball_reach_step
                                                                   false,
                                                                   "SelfPass" ) );
                M_courses.push_back( ptr );
                //M_receive_point_set.insert( get_point_index( receive_pos ) );
            }
#ifdef DEBUG_PRINT_FAILED_PASS
            else
            {
                dlog.addText( Logger::PASS,
                              "XXX %d failed. ball first_speed=%.3f reach_step=%d",
                              M_total_count,
                              ball_first_speed, my_reach_step );
                dlog.addRect( Logger::PASS,
                              receive_pos.x - 0.1, receive_pos.y - 0.1,
                              0.2, 0.2,
                              "#00ff00" );
                char num[8]; snprintf( num, 8, "s%d", M_total_count );
                dlog.addMessage( Logger::PASS,
                                 receive_pos, num );
            }
#endif
        }
    }

#ifdef DEBUG_PROFILE
    dlog.addText( Logger::PASS,
                  "(createSelfPass) count=%d elapsed %.3f [ms]",
                  count, timer.elapsedReal() );
#endif
}

/*-------------------------------------------------------------------*/
/*!

 */
int
PassTable2009::predictReceiverReachStep( const WorldModel & /*wm*/,
                                         const AbstractPlayerObject * receiver,
                                         const Vector2D & receive_pos,
                                         const Vector2D & first_ball_pos,
                                         const AngleDeg & first_ball_vel_angle )
{
    const ServerParam & SP = ServerParam::i();
    const PlayerType * ptype = receiver->playerTypePtr();

    //const int pos_count = std::min( 3, receiver->posCount() );
    const Vector2D & ppos = ( receiver->seenPosCount() <= receiver->posCount()
                              ? receiver->seenPos()
                              : receiver->pos() );
    const Vector2D & pvel = ( receiver->seenVelCount() <= receiver->velCount()
                              ? receiver->seenVel()
                              : receiver->vel() );
    const double pspeed = pvel.r();

    int min_cycle = 1;
    {
        Vector2D ball_to_player = ppos - first_ball_pos;
        ball_to_player.rotate( - first_ball_vel_angle );
        min_cycle = static_cast< int >( std::floor( ball_to_player.absY() / ptype->realSpeedMax() ) );
        min_cycle = std::max( 1, min_cycle );
    }

#ifdef DEBUG_RECEIVER
    dlog.addText( Logger::PASS,
                  "(predictReceiverReachStep) min_cycle=%d receiver=%d (%.2f %.2f)",
                  min_cycle, receiver->unum(), receiver->pos().x, receiver->pos().y );
//     std::ostringstream os;
//     for ( int i = 0; i < 5; ++i )
//         os << ptype->dashDistanceTable().at( i ) << ", ";
//     dlog.addText( Logger::PASS,
//                   "dash table = %s", os.str().c_str() );
#endif

    for ( int cycle = min_cycle; cycle <= 6; ++cycle )
    {
        Vector2D inertia_pos = inertia_n_step_point( ppos,
                                                     pvel,
                                                     cycle,
                                                     ptype->playerDecay() );
        double target_dist = inertia_pos.dist( receive_pos );

        if ( target_dist < ptype->kickableArea() * 0.7 )
        {
#ifdef DEBUG_RECEIVER
            dlog.addText( Logger::PASS,
                          "__ cycle=%d, can kick by inertia move. target_dist=%.3f < kickable*0.7=%.df",
                          cycle, target_dist, ptype->kickableArea()*0.7 );
#endif
            return cycle;
        }

        double dash_dist = target_dist - ptype->kickableArea() * 0.5;

        if ( dash_dist > ptype->realSpeedMax() * cycle )
        {
#ifdef DEBUG_RECEIVER
            dlog.addText( Logger::PASS,
                          "__ cycle=%d, target_dist=%.3f dash_dist=%.3f never reach",
                          cycle, target_dist, dash_dist );
#endif
            continue;
        }

        const int n_dash = ptype->cyclesToReachDistance( dash_dist );
        if ( n_dash > cycle )
        {
#ifdef DEBUG_RECEIVER
            dlog.addText( Logger::PASS,
                          "__ cycle=%d, target_dist=%.3f dash_dist=%.3f n_dash=%d never reach.",
                          cycle, target_dist, dash_dist, n_dash );
#endif
            continue;
        }

#ifdef DEBUG_RECEIVER
        dlog.addText( Logger::PASS,
                      "__ cycle=%d, target_dist=%.3f dash_dist=%.3f n_dash=%d",
                      cycle, target_dist, dash_dist, n_dash );
#endif

        //
        // turn
        //
        int n_turn = 0;

        if ( receiver->bodyCount() <= 1 )
        {
            AngleDeg target_angle = ( receive_pos - inertia_pos ).th();

            double angle_diff = ( target_angle - receiver->body() ).abs();
            double turn_margin = 180.0;
            if ( ptype->kickableArea() * 0.9 < target_dist )
            {
                turn_margin = std::max( 15.0,
                                        AngleDeg::asin_deg( ptype->kickableArea() * 0.9 / target_dist ) );
            }

            double speed = pspeed;
            while ( angle_diff > turn_margin )
            {
                angle_diff -= ptype->effectiveTurn( SP.maxMoment(), speed );
                speed *= ptype->playerDecay();
                ++n_turn;
            }

#ifdef DEBUG_RECEIVER
            dlog.addText( Logger::PASS,
                          "__ cycle=%d, n_turn=%d, angle_diff=%.1f",
                          cycle, n_turn, ( target_angle - receiver->body() ).abs() );
#endif
        }
        else
        {
            n_turn = 1;
#ifdef DEBUG_RECEIVER
            dlog.addText( Logger::PASS,
                          "__ cycle=%d, n_turn=%d, no body info",
                          cycle, n_turn );
#endif
        }

        if ( n_turn + n_dash < cycle )
        {
#ifdef DEBUG_RECEIVER
            dlog.addText( Logger::PASS,
                          "__ cycle=%d, n_turn=%d, n_dash=%d",
                          cycle, n_turn, n_dash );
#endif
            if ( n_turn > 0 )
            {
                return n_turn + n_dash + 2; // 2 step penalty
            }
            else
            {
                return n_turn + n_dash + 1; // 1 step penalty
            }
        }
    }

    {
        Vector2D inertia_pos = receiver->inertiaFinalPoint();
        double target_dist = ( receive_pos - inertia_pos ).r();

        int n_turn = 0;
        if ( receiver->bodyCount() <= 1 )
        {
            AngleDeg target_angle = ( receive_pos - inertia_pos ).th();

            double angle_diff = ( target_angle - receiver->body() ).abs();
            double turn_margin = 180.0;
            if ( ptype->kickableArea() * 0.9 < target_dist )
            {
                turn_margin = std::max( 15.0,
                                        AngleDeg::asin_deg( ptype->kickableArea() * 0.9 / target_dist ) );
            }

            double speed = pspeed;
            while ( angle_diff > turn_margin )
            {
                angle_diff -= ptype->effectiveTurn( SP.maxMoment(), speed );
                speed *= ptype->playerDecay();
                ++n_turn;
            }
        }
        else
        {
            n_turn = 1;
        }

        int n_dash = ptype->cyclesToReachDistance( target_dist - ptype->kickableArea() * 0.7 );
#ifdef DEBUG_RECEIVER
        dlog.addText( Logger::PASS,
                      "__ final. inertia_pos=(%.2f %.2f) n_turn=%d n_dash=%d",
                      inertia_pos.x, inertia_pos.y,
                      n_turn, n_dash );
#endif
        if ( n_turn > 0 )
        {
            return n_turn + n_dash + 2; // 2 step penalty
        }
        else
        {
            return n_turn + n_dash + 1; // 1 step penalty
        }
    }

    return 1000;
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
PassTable2009::checkCourse( const WorldModel & wm,
                            const AbstractPlayerObject * receiver,
                            const Vector2D & receive_point,
                            const double & ball_first_speed,
                            const int ball_reach_step )
{
    const Vector2D ball_first_pos = wm.ball().pos();
    const AngleDeg target_angle = ( receive_point - wm.ball().pos() ).th();
    const Vector2D ball_first_vel = Vector2D::polar2vector( ball_first_speed, target_angle );
    //const Vector2D ball_next_pos = ball_first_pos + ball_first_vel;
    //const double ball_next_speed = ball_first_speed * SP.ballDecay();

#ifdef DEBUG_CHECK_PASS
    dlog.addText( Logger::PASS,
                  "(checkCourse:%d) receiver=%d receive_point=(%.1f %.1f) ball_speed=%.3f ball_reach_step=%d",
                  M_total_count,
                  receiver->unum(),
                  receive_point.x, receive_point.y,
                  ball_first_speed, ball_reach_step );
#else
    (void)receiver;
#endif

    double opponent_min_dist = 1000000.0;
    const PlayerPtrCont::const_iterator o_end = wm.opponentsFromBall().end();
    for ( PlayerPtrCont::const_iterator o = wm.opponentsFromBall().begin();
          o != o_end;
          ++o )
    {
        if ( existInterceptor( **o,
                               ball_first_pos, ball_first_vel, target_angle,
                               ball_reach_step,
                               &opponent_min_dist ) )
        {
            return false;
        }
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
PassTable2009::existInterceptor( const PlayerObject & opponent,
                                 const Vector2D & ball_first_pos,
                                 const Vector2D & ball_first_vel,
                                 const AngleDeg & target_angle,
                                 const int ball_reach_step,
                                 double * /*min_dist*/ )
{
    static const Rect2D penalty_area( Vector2D( ServerParam::i().theirPenaltyAreaLineX(),
                                                -ServerParam::i().penaltyAreaHalfWidth() ),
                                      Size2D( ServerParam::i().penaltyAreaLength(),
                                              ServerParam::i().penaltyAreaWidth() ) );

    const ServerParam & SP = ServerParam::i();

    const int pos_count = std::min( opponent.seenPosCount(), opponent.posCount() );

#ifdef DEBUG_CHECK_PASS
    dlog.addText( Logger::PASS,
                  "====== opponent %d (%.1f %.1f) pos_count=%d ======",
                  opponent.unum(), opponent.pos().x, opponent.pos().y,
                  pos_count );
#endif

    if ( pos_count > 50
         || ( opponent.isGhost() && pos_count >= 4 ) )
    {
        return false;
    }

    const Vector2D ball_next_pos = ball_first_pos + ball_first_vel;

    const Vector2D & player_pos = ( opponent.seenPosCount() <= opponent.posCount()
                                    ? opponent.seenPos()
                                    : opponent.pos() );
    const Vector2D & player_vel = ( opponent.seenVelCount() <= opponent.velCount()
                                    ? opponent.seenVel()
                                    : opponent.vel() );

    const PlayerType * ptype = opponent.playerTypePtr();
    double control_area = ( opponent.goalie() && penalty_area.contains( ball_next_pos )
                            ? SP.catchableArea()
                            : ptype->kickableArea() );

#if 1
    if ( pos_count == 0
         //ball_next_pos.dist2( player_pos ) < std::pow( control_area + 0.1, 2 )
         && ball_next_pos.dist2( player_pos + player_vel ) < std::pow( control_area + 0.1, 2 ) )
    {
#ifdef DEBUG_CHECK_PASS
        dlog.addText( Logger::PASS,
                      "xxxx ball kickable at next cycle. ball_next_pos=(%.1f %.1f).",
                      ball_next_pos.x, ball_next_pos.y );
#endif
        return true;
    }
#endif

    Vector2D ball_to_opponent = player_pos - ball_first_pos;
    ball_to_opponent.rotate( - target_angle );

    if ( ball_to_opponent.x < -1.0 )
    {
        // ignore behind opponent
#ifdef DEBUG_CHECK_PASS
        dlog.addText( Logger::PASS,
                      "____ behind. rel_x=%.3f",
                      ball_to_opponent.x );
#endif
        return false;
    }

    //
    // estimate minimum reach step
    //

    int min_cycle = std::max( 2,
                              static_cast< int >( std::floor( ( ball_to_opponent.absY() - ptype->kickableArea() )
                                                              / ptype->realSpeedMax() ) ) );
    if ( opponent.isTackling() )
    {
        min_cycle += std::max( 0, SP.tackleCycles() - opponent.tackleCount() - 5 );
    }

    //
    // set start ball state
    //

    double opponent_min_dist = 1000000.0;

    Vector2D ball_vel = ball_first_vel;
    ball_vel *= SP.ballDecay(); // advance 1 step

    Vector2D ball_pos = inertia_n_step_point( ball_next_pos,
                                              ball_vel,
                                              min_cycle - 2,
                                              SP.ballDecay() );
    ball_vel *= std::pow( SP.ballDecay(), min_cycle - 2 );

#ifdef DEBUG_CHECK_PASS
    dlog.addText( Logger::PASS,
                  "____ min_cycle=%d, start ball pos=(%.2f %.2f) vel=(%.2f %.2f)",
                  min_cycle,
                  ball_pos.x, ball_pos.y,
                  ball_vel.x, ball_vel.y );
#endif

    //
    // simulation loop
    //

    int closest_cycle = 100000;
    int closest_opponent_cycle = 100000;

    for ( int cycle = min_cycle; cycle <= ball_reach_step; ++cycle )
    {
        ball_pos += ball_vel;
        ball_vel *= SP.ballDecay();

        Vector2D inertia_pos = ptype->inertiaPoint( player_pos, player_vel, cycle );
        Vector2D ball_rel = ball_pos - inertia_pos;
        double ball_dist = ball_rel.r();

        if ( opponent_min_dist > ball_dist )
        {
            opponent_min_dist = ball_dist;
        }

        control_area = ( opponent.goalie() && penalty_area.contains( ball_pos )
                         ? SP.catchableArea()
                         : ptype->kickableArea() );

        if ( ball_dist < control_area + 0.1 )
        {
#ifdef DEBUG_CHECK_PASS
            dlog.addText( Logger::PASS,
                          "%d xxxx cycle=%d bpos=(%.2f %.2f) ball_dist=%.3f opponent can reach the ball",
                          cycle, cycle,
                          ball_pos.x, ball_pos.y,
                          ball_dist );
#endif
            return true;
        }

        if ( ptype->realSpeedMax() * ( cycle + pos_count ) + control_area
             < ball_dist )
        {
            // never reach
            continue;
        }

        //
        // predict turn step
        //

        int n_turn = 0;
        if ( opponent.bodyCount() <= 3 )
        {
            double turn_margin = 180.0;
            if ( control_area < ball_dist )
            {
                turn_margin = AngleDeg::asin_deg( control_area / ball_dist );
            }
            turn_margin = std::max( turn_margin, 12.0 );

            double angle_diff = ( ball_rel.th() - opponent.body() ).abs();

#if 0
            if ( ball_dist < 3.0 // XXX magic number XXX
                 && angle_diff > 90.0 )
            {
                // assume back dash
                angle_diff = 180.0 - angle_diff;
            }
#endif

#ifdef DEBUG_CHECK_PASS_LEVEL_2
            dlog.addText( Logger::PASS,
                          "%d __ cycle=%d bpos=(%.1f %.1f) body_count=%d turn_margin=%.1f angle_diff=%.1f",
                          cycle, cycle, ball_pos.x, ball_pos.y,
                          opponent.bodyCount(), turn_margin, angle_diff );
#endif

            double speed = opponent.vel().r();
            if ( angle_diff > turn_margin )
            {
                double max_turn = ptype->effectiveTurn( SP.maxMoment(), speed );
                angle_diff -= max_turn;
                speed *= ptype->playerDecay();
                ++n_turn;
            }
        }

        const int max_dash = cycle - n_turn;
        if ( max_dash < 1 )
        {
            continue;
        }

        //
        // predict dash step
        //

        //double dash_dist = std::max( 0.0, ball_dist - control_area*0.7 - cycle*0.1 );
        double dash_dist = std::max( 0.0, ball_dist - control_area - cycle*0.1 );
        int n_real_dash = ptype->cyclesToReachDistance( dash_dist );

        int n_dash = n_real_dash;

        if ( ball_pos.x > 30.0 )
        {
            n_dash -= bound( 0, pos_count - n_turn - 1, 3 );
        }
        else
        {
            n_dash -= bound( 0, pos_count - n_turn, 3 );
        }

        if ( n_turn > 0 )
        {
            n_dash += 1; // 1 step penalty
        }

        if ( opponent.isTackling() )
        {
            n_dash += std::max( 0, SP.tackleCycles() - opponent.tackleCount() - 2 );
        }

#ifdef DEBUG_CHECK_PASS_LEVEL_2
        dlog.addText( Logger::PASS,
                      "%d __ cycle=%d bpos=(%.1f %.1f) turn=%d real_dash=%d dash=%d  ball_dist=%.3f",
                      cycle, cycle, ball_pos.x, ball_pos.y,
                      n_turn, n_real_dash, n_dash,
                      ball_dist );
#endif

        if ( n_dash <= max_dash )
        {
#ifdef DEBUG_CHECK_PASS
            double opponent_dash = ball_dist - control_area;
            if ( n_real_dash > 0 )
            {
                try
                {
                    opponent_dash =  ptype->dashDistanceTable().at( n_real_dash - 1 );
                }
                catch ( std::exception & e )
                {
                    std::cerr <<  e.what() << " n_real_dash=" << n_real_dash << std::endl;
                }
            }
            dlog.addText( Logger::PASS,
                          "%d xxxx cycle=%d bpos=(%.2f %.2f) ball_dist=%.3f opponent can reach.",
                          cycle, cycle,
                          ball_pos.x, ball_pos.y,
                          ball_dist );
            dlog.addText( Logger::PASS,
                          "%d ____ turn=%d dash=%d(real==%d) reachable_dist=%.2f required_dist=%.2f",
                          cycle,
                          n_turn, n_dash, n_real_dash,
                          opponent_dash, dash_dist );
#endif
            return true;
        }

        if ( n_turn + n_dash - cycle < closest_opponent_cycle )
        {
            closest_cycle = cycle;
            closest_opponent_cycle = n_turn + n_dash - cycle;
        }

        if ( cycle == ball_reach_step )
        {
            if ( ball_pos.x < 35.0
                 && n_dash <= max_dash + 1 )
            {
#ifdef DEBUG_CHECK_PASS_LEVEL_2
                dlog.addText( Logger::PASS,
                              "%d xxxx cycle=%d bpos=(%.1f %.1f) turn=%d dash=%d  ball_dist=%.3f dash_dist=%.3f opponent may reach",
                              cycle, cycle, ball_pos.x, ball_pos.y,
                              n_turn, n_dash,
                              ball_dist, dash_dist );
#endif
                return true;
            }
        }
    }

#ifdef DEBUG_CHECK_PASS
    dlog.addText( Logger::PASS,
                  "ok closest_cycle=%d  opponent_cycle=%d",
                  closest_cycle, closest_opponent_cycle );
#endif

    // no interceptor
    return false;
}

}
