// -*-c++-*-

/*!
  \file dribble_table2009.cpp
  \brief dribble course cache table Source File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "dribble_table2009.h"

#include <rcsc/player/world_model.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/timer.h>

#include <algorithm>
#include <limits>

#include <cmath>

// #define DEBUG_PROFILE
// #define DEBUG_PRINT

namespace rcsc {

/*-------------------------------------------------------------------*/
/*!

*/
DribbleTable2009::DribbleTable2009()
{
    M_courses.reserve( 128 );
    M_aggressive_courses.reserve( 128 );
}


/*-------------------------------------------------------------------*/
/*!

*/
DribbleTable2009 &
DribbleTable2009::instance()
{
    static DribbleTable2009 s_instance;
    return s_instance;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
DribbleTable2009::update( const WorldModel & wm )
{
    static GameTime s_update_time( 0, 0 );
    if ( s_update_time == wm.time() )
    {
        return;
    }
    s_update_time = wm.time();

    M_total_count = 0;
    M_courses.clear();
    M_aggressive_courses.clear();

    if ( wm.gameMode().type() != GameMode::PlayOn
         && ! wm.gameMode().isPenaltyKickMode() )
    {
        return;
    }

    if ( ! wm.self().isKickable() )
    {
        return;
    }

#ifdef DEBUG_PROFILE
    Timer timer;
#endif

    createTargetPoints( wm );

#ifdef DEBUG_PROFILE
    dlog.addText( Logger::DRIBBLE,
                  __FILE__": (update) elapsed %.3f [ms]",
                  timer.elapsedReal() );
#endif
}

/*-------------------------------------------------------------------*/
/*!

 */
const
std::vector< CooperativeAction::Ptr > &
DribbleTable2009::getDribbleCandidates( const WorldModel & wm )
{
    update( wm );
    return M_courses;
}

/*-------------------------------------------------------------------*/
/*!

 */
const
std::vector< CooperativeAction::Ptr > &
DribbleTable2009::getAggressiveDribbleCandidates( const WorldModel & wm )
{
    update( wm );
    return M_aggressive_courses;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
DribbleTable2009::createTargetPoints( const WorldModel & wm )
{
    static const int angle_div = 16;
    static const double angle_step = 360.0 / angle_div;

    const ServerParam & SP = ServerParam::i();

    const double my_first_speed = wm.self().vel().r();
    const PlayerType & ptype = wm.self().playerType();

    for ( int a = 0; a < angle_div; ++a )
    {
        AngleDeg dash_angle = wm.self().body() + ( angle_step * a );

        //
        // angle filter
        //

        if ( wm.self().pos().x < 16.0
             && dash_angle.abs() > 100.0 )
        {
#ifdef DEBUG_PRINT
            dlog.addText( Logger::DRIBBLE,
                          __FILE__": (createTargetPoints) canceled(1) dash_angle=%.1f",
                          dash_angle.degree() );
#endif
            continue;
        }

        if ( wm.self().pos().x < -36.0
             && wm.self().pos().absY() < 20.0
             && dash_angle.abs() > 45.0 )
        {
#ifdef DEBUG_PRINT
            dlog.addText( Logger::DRIBBLE,
                          __FILE__": (createTargetPoints) canceled(2) dash_angle=%.1f",
                          dash_angle.degree() );
#endif
            continue;
        }


        int n_turn = 0;

        double my_speed = my_first_speed * ptype.playerDecay(); // first action is kick
        double dir_diff = AngleDeg( angle_step * a ).abs();

        while ( dir_diff > 10.0 )
        {
            dir_diff -= ptype.effectiveTurn( SP.maxMoment(), my_speed );
            if ( dir_diff < 0.0 ) dir_diff = 0.0;
            my_speed *= ptype.playerDecay();
            ++n_turn;
        }

        if ( n_turn >= 2 )
        {
#ifdef DEBUG_PRINT
            dlog.addText( Logger::DRIBBLE,
                          __FILE__": (createTargetPoints) canceled dash_angle=%.1f n_turn=%d",
                          dash_angle.degree(), n_turn );
#endif
            continue;
        }

        if ( angle_step * a < 180.0 )
        {
            dash_angle -= dir_diff;
        }
        else
        {
            dash_angle += dir_diff;
        }

        simulateKickTurnsDashes( wm, dash_angle, n_turn );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
DribbleTable2009::simulateKickTurnsDashes( const WorldModel & wm,
                                           const AngleDeg & dash_angle,
                                           const int n_turn )
{
    //static const int max_dash = 5;
    static const int max_dash = 4; // 2009-06-29
    static const int min_dash = 2;

    static std::vector< Vector2D > self_cache;

    createSelfCache( wm, dash_angle, n_turn, max_dash, self_cache );

    const ServerParam & SP = ServerParam::i();
    const PlayerType & ptype = wm.self().playerType();

    const Vector2D trap_rel
        = Vector2D::polar2vector( ptype.playerSize() + ptype.kickableMargin() * 0.2 + SP.ballSize(),
                                  dash_angle );

    const double max_x = ServerParam::i().pitchHalfLength() - 1.0;
    const double max_y = ServerParam::i().pitchHalfWidth() - 1.0;

#ifdef DEBUG_PRINT
    dlog.addText( Logger::DRIBBLE,
                  __FILE__": (simulateKickTurnsDashes) dash_angle=%.1f n_turn=%d",
                  dash_angle.degree(), n_turn );
#endif

    for ( int n_dash = max_dash; n_dash >= min_dash; --n_dash )
    {
        const Vector2D ball_trap_pos = self_cache[n_turn + n_dash] + trap_rel;

        ++M_total_count;

#ifdef DEBUG_PRINT
        dlog.addText( Logger::DRIBBLE,
                      "%d __ n_turn=%d n_dash=%d ball_trap=(%.3f %.3f)",
                      M_total_count,
                      n_turn, n_dash,
                      ball_trap_pos.x, ball_trap_pos.y );
#endif
        if ( ball_trap_pos.absX() > max_x
             || ball_trap_pos.absY() > max_y )
        {
#ifdef DEBUG_PRINT
            dlog.addText( Logger::DRIBBLE,
                          "___ xxx out of pitch" );
#endif
            continue;
        }

        const double term
            = ( 1.0 - std::pow( SP.ballDecay(), 1 + n_turn + n_dash ) )
            / ( 1.0 - SP.ballDecay() );
        const Vector2D first_vel = ( ball_trap_pos - wm.ball().pos() ) / term;
        const Vector2D kick_accel = first_vel - wm.ball().vel();
        const double kick_power = kick_accel.r() / wm.self().kickRate();

        if ( kick_power > SP.maxPower()
             || kick_accel.r2() > std::pow( SP.ballAccelMax(), 2 )
             || first_vel.r2() > std::pow( SP.ballSpeedMax(), 2 ) )
        {
#ifdef DEBUG_PRINT
            dlog.addText( Logger::DRIBBLE,
                          "____ xxx cannot kick. first_vel=(%.1f %.1f)r=%.2f accel=(%.1f %.1f)r=%.2f power=%.1f",
                          first_vel.x, first_vel.y, first_vel.r(),
                          kick_accel.x, kick_accel.y, kick_accel.r(),
                          kick_power );
#endif
            continue;
        }

        if ( ( wm.ball().pos() + first_vel ).dist2( self_cache[0] )
             < std::pow( ptype.playerSize() + SP.ballSize() + 0.1, 2 ) )
        {
#ifdef DEBUG_PRINT
            dlog.addText( Logger::DRIBBLE,
                          "____ xxx collision. first_vel=(%.1f %.1f)r=%.2f accel=(%.1f %.1f)r=%.2f power=%.1f",
                          first_vel.x, first_vel.y, first_vel.r(),
                          kick_accel.x, kick_accel.y, kick_accel.r(),
                          kick_power );
#endif
            continue;
        }

        if ( checkOpponent( wm, ball_trap_pos, 1 + n_turn + n_dash, false ) )
        {
            CooperativeAction::Ptr ptr( new CooperativeAction( CooperativeAction::ActionType::Dribble,
                                                               ball_trap_pos,
                                                               wm.self().unum(),
                                                               first_vel.r(),
                                                               1 + n_turn + n_dash,
                                                               1 + n_turn + n_dash,
                                                               false,
                                                               "dribble0" ) );
            M_courses.push_back( ptr );

#ifdef DEBUG_PRINT
            dlog.addText( Logger::DRIBBLE,
                          ">>>> ok trap_pos=(%.2f %.2f) first_vel=(%.1f %.1f) n_turn=%d n_dash=%d",
                          ball_trap_pos.x, ball_trap_pos.y,
                          first_vel.x, first_vel.y,
                          n_turn, n_dash );
            dlog.addCircle( Logger::DRIBBLE,
                            ball_trap_pos.x, ball_trap_pos.y, 0.2,
                            "#ff0000" );
            char num[8]; snprintf( num, 8, "%d", M_total_count );
            dlog.addMessage( Logger::DRIBBLE,
                             ball_trap_pos, num );
#endif
            //break; // 2009-07-01 commented out
        }
        else if ( checkOpponent( wm, ball_trap_pos, 1 + n_turn + n_dash, true ) )
        {
            CooperativeAction::Ptr ptr( new CooperativeAction( CooperativeAction::ActionType::Dribble,
                                                               ball_trap_pos,
                                                               wm.self().unum(),
                                                               first_vel.r(),
                                                               1 + n_turn + n_dash,
                                                               1 + n_turn + n_dash,
                                                               false,
                                                               "aDribble0" ) );
            M_aggressive_courses.push_back( ptr );

#ifdef DEBUG_PRINT
            dlog.addText( Logger::DRIBBLE,
                          ">>>> ok aggressive trap_pos=(%.2f %.2f) first_vel=(%.1f %.1f) n_turn=%d n_dash=%d",
                          ball_trap_pos.x, ball_trap_pos.y,
                          first_vel.x, first_vel.y,
                          n_turn, n_dash );
            dlog.addCircle( Logger::DRIBBLE,
                            ball_trap_pos.x, ball_trap_pos.y, 0.2,
                            "#ff00ff" );
            char num[8]; snprintf( num, 8, "%d", M_total_count );
            dlog.addMessage( Logger::DRIBBLE,
                             ball_trap_pos, num );
#endif
        }
#ifdef DEBUG_PRINT
        else
        {
            dlog.addText( Logger::DRIBBLE,
                          "XXXX failed trap_pos=(%.2f %.2f) first_vel=(%.1f %.1f) n_turn=%d n_dash=%d",
                          ball_trap_pos.x, ball_trap_pos.y,
                          first_vel.x, first_vel.y,
                          n_turn, n_dash );
            dlog.addCircle( Logger::DRIBBLE,
                            ball_trap_pos.x, ball_trap_pos.y, 0.2,
                            "#00ff00" );
            char num[8]; snprintf( num, 8, "%d", M_total_count );
            dlog.addMessage( Logger::DRIBBLE,
                             ball_trap_pos, num );
        }
#endif
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
DribbleTable2009::createSelfCache( const WorldModel & wm,
                                   const AngleDeg & dash_angle,
                                   const int n_turn,
                                   const int n_dash,
                                   std::vector< Vector2D > & self_cache )
{
    const ServerParam & SP = ServerParam::i();
    const PlayerType & ptype = wm.self().playerType();

    self_cache.clear();

    double my_stamina = wm.self().stamina();
    double my_effort = wm.self().effort();
    double my_recovery = wm.self().recovery();

    Vector2D my_pos = wm.self().pos();
    Vector2D my_vel = wm.self().vel();

    my_pos += my_vel;
    my_vel *= ptype.playerDecay();

    self_cache.push_back( my_pos ); // first element is next cycle just after kick

    for ( int i = 0; i < n_turn; ++i )
    {
        my_pos += my_vel;
        my_vel *= ptype.playerDecay();
        self_cache.push_back( my_pos );
    }

    ptype.predictStaminaAfterWait( SP, 1 + n_turn,
                                   &my_stamina, &my_effort, my_recovery );

    const Vector2D unit_vec = Vector2D::polar2vector( 1.0, dash_angle );

    for ( int i = 0; i < n_dash; ++i )
    {
        double available_stamina = std::max( 0.0,
                                             my_stamina - SP.recoverDecThrValue() - 300.0 );
        double dash_power = std::min( available_stamina, SP.maxDashPower() );
        Vector2D dash_accel = unit_vec.setLengthVector( dash_power * ptype.dashPowerRate() * my_effort );

        my_vel += dash_accel;
        my_pos += my_vel;
        my_vel *= ptype.playerDecay();
        ptype.predictStaminaAfterOneDash( SP, dash_power, &my_stamina, &my_effort, &my_recovery );

        self_cache.push_back( my_pos );
    }

}

/*-------------------------------------------------------------------*/
/*!

*/
bool
DribbleTable2009::checkOpponent( const WorldModel & wm,
                                 const Vector2D & ball_trap_pos,
                                 const int dribble_step,
                                 const bool aggressive_mode )
{
    const ServerParam & SP = ServerParam::i();

    //const double control_area = SP.tackleDist() - 0.2;
    const rcsc::AngleDeg ball_move_angle = ( ball_trap_pos - wm.ball().pos() ).th();

    const int pos_count_penalty = ( aggressive_mode ? 1 : 0 );
    const int pos_count_max = ( aggressive_mode ? 1 : 5 );

    const PlayerPtrCont::const_iterator o_end = wm.opponentsFromSelf().end();
    for ( PlayerPtrCont::const_iterator o = wm.opponentsFromSelf().begin();
          o != o_end;
          ++o )
    {
        if ( (*o)->distFromSelf() > 20.0 ) break;

        const PlayerType * ptype = (*o)->playerTypePtr();

#if 1
        const double control_area
            = ( (*o)->goalie()
                && ball_trap_pos.x > SP.theirPenaltyAreaLineX()
                && ball_trap_pos.absY() < SP.penaltyAreaHalfWidth() )
            ? SP.catchableArea()
            : ptype->kickableArea();
#endif

        const Vector2D & opos = ( (*o)->seenPosCount() <= (*o)->posCount()
                                  ? (*o)->seenPos()
                                  : (*o)->pos() );
        const int vel_count = std::min( (*o)->seenVelCount(), (*o)->velCount() );
        const Vector2D & ovel = ( (*o)->seenVelCount() <= (*o)->velCount()
                                  ? (*o)->seenVel()
                                  : (*o)->vel() );

        Vector2D opp_pos = ( (*o)->velCount() <= 1
                             ? inertia_n_step_point( opos, ovel, dribble_step,
                                                     ptype->playerDecay() )
                             : opos + ovel );
        Vector2D opp_to_trap_pos = ball_trap_pos - opp_pos;
        double opp_dist = opp_to_trap_pos.r();

        //
        // predict turn
        //

        int opp_turn_step = 0;
        if ( (*o)->bodyCount() <= 5
             || vel_count <= 1 )
        {
            double angle_diff = ( (*o)->bodyCount() <= 1
                                  ? ( opp_to_trap_pos.th() - (*o)->body() ).abs()
                                  : ( opp_to_trap_pos.th() - ovel.th() ).abs() );

            double turn_margin = 180.0;
            if ( control_area < opp_dist )
            {
                turn_margin = AngleDeg::asin_deg( control_area / opp_dist );
            }
            turn_margin = std::max( turn_margin, 15.0 );

            double speed = ovel.r();
            while ( angle_diff > turn_margin )
            {
                angle_diff -= ptype->effectiveTurn( SP.maxMoment(), speed );
                speed *= ptype->playerDecay();
                ++opp_turn_step;
            }
        }

        opp_dist -= control_area;
        opp_dist -= 0.2;

        if ( opp_dist < 0.0 )
        {
#ifdef DEBUG_PRINT
            dlog.addText( Logger::DRIBBLE,
                          "____ xxx step=%d opponent %d(%.1f %.1f) is already at receive point",
                          dribble_step,
                          (*o)->unum(),
                          (*o)->pos().x, (*o)->pos().y );
#endif
            return false;
        }

        //
        // predict dash
        //

        int opp_reach_step = ptype->cyclesToReachDistance( opp_dist );
        opp_reach_step += opp_turn_step;

        if ( ( ( ball_trap_pos - (*o)->pos() ).th() - ball_move_angle ).abs() < 90.0 )
        {
            opp_reach_step -= bound( 0,
                                     (*o)->posCount() - pos_count_penalty,
                                     pos_count_max );
        }
        else
        {
            opp_reach_step -= bound( 0,
                                     (*o)->posCount() - pos_count_penalty,
                                     1 );
        }

        if ( aggressive_mode
             && opp_turn_step > 0 )
        {
            opp_reach_step += 1;
        }

        if ( opp_reach_step <= dribble_step
             || ( ! aggressive_mode
                  && ball_trap_pos.x < 33.0
                  && opp_reach_step <= dribble_step + 2 )
             )
        {
#ifdef DEBUG_PRINT
            dlog.addText( Logger::DRIBBLE,
                          "____ xxx step=%d opponent %d (%.1f %.1f) can reach faster then self."
                          " opp_step=%d(turn=%d)",
                          dribble_step,
                          (*o)->unum(),
                          (*o)->pos().x, (*o)->pos().y,
                          opp_reach_step,
                          opp_turn_step );
#endif
            return false;
        }

#ifdef DEBUG_PRINT
        dlog.addText( Logger::DRIBBLE,
                      "____ ok step=%d opponent %d (%.1f %.1f) opp_step=%d(turn=%d)",
                      dribble_step,
                      (*o)->unum(),
                      (*o)->pos().x, (*o)->pos().y,
                      opp_reach_step,
                      opp_turn_step );
#endif
    }

    return true;
}

}
