// -*-c++-*-

/*!
  \file cooperative_action.h
  \brief cooperative action object type Header File.
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA, Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 3 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef RCSC_PLAYER_COOPERATIVE_ACTION_H
#define RCSC_PLAYER_COOPERATIVE_ACTION_H

#include <rcsc/geom/vector_2d.h>

#include <boost/shared_ptr.hpp>

#include <vector>

namespace rcsc {

class AbstractPlayerObject;

/*!
  \class CooperativeAction
  \brief cooperative action object.
*/
class CooperativeAction {

public:
    static const int INVALID_TIME = -1;

public:
    /*!
      \struct ActionType
      \brief action type enumeration.
     */
    struct ActionType {
        enum Type {
            // ball holder action
            Hold,
            Dribble,
            Pass,
            Shoot
        };
    };

    class SupportPlayerMove {
    private:
        int M_unum;
        rcsc::Vector2D M_target_point;

    public:
        SupportPlayerMove( int unum, const rcsc::Vector2D & target_point )
            : M_unum( unum ), M_target_point( target_point )
          { }

        int unum() const
          {
              return M_unum;
          }

        const rcsc::Vector2D & targetPoint() const
          {
              return M_target_point;
          }
    };

    typedef std::vector< SupportPlayerMove > SupportPlayerMovesCont;

private:
    const char * M_tag;
    ActionType::Type M_type;
    rcsc::Vector2D M_target_point;
    int M_target_player_unum;
    double M_ball_speed;
    int M_receiver_reach_step;
    int M_ball_reach_step;
    bool M_is_final_action;

    SupportPlayerMovesCont M_support_player_moves;

public:
    CooperativeAction( const ActionType::Type & type,
                       bool is_final_action,
                       const char * tag )
        : M_tag( tag )
        , M_type( type )
        , M_target_point()
        , M_target_player_unum( -1 )
        , M_ball_speed( 0.0 )
        , M_receiver_reach_step( INVALID_TIME )
        , M_ball_reach_step( INVALID_TIME )
        , M_is_final_action( is_final_action )
        , M_support_player_moves()
      { }

    CooperativeAction( const ActionType::Type & type,
                       const rcsc::Vector2D & target_point,
                       int target_player_unum,
                       int receiver_reach_step,
                       int ball_reach_step,
                       bool is_final_action,
                       const char * tag = static_cast<char *>(0) )
        : M_tag( tag )
        , M_type( type )
        , M_target_point( target_point )
        , M_target_player_unum( target_player_unum )
        , M_ball_speed( 0.0 )
        , M_receiver_reach_step( receiver_reach_step )
        , M_ball_reach_step( ball_reach_step )
        , M_is_final_action( is_final_action )
        , M_support_player_moves()
      { }

    CooperativeAction( const ActionType::Type & type,
                       const rcsc::Vector2D & target_point,
                       int target_player_unum,
                       double ball_speed,
                       int receiver_reach_step,
                       int ball_reach_step,
                       bool is_final_action,
                       const char * tag )
        : M_tag( tag )
        , M_type( type )
        , M_target_point( target_point )
        , M_target_player_unum( target_player_unum )
        , M_ball_speed( ball_speed )
        , M_receiver_reach_step( receiver_reach_step )
        , M_ball_reach_step( ball_reach_step )
        , M_is_final_action( is_final_action )
        , M_support_player_moves()
      { }

    void addSupportPlayerMove( const SupportPlayerMove & move )
      {
          M_support_player_moves.push_back( move );
      }

    const char * tag() const
      {
          return M_tag;
      }

    const ActionType::Type & type() const
      {
          return M_type;
      }

    bool isFinalAction() const
      {
          return M_is_final_action;
      }

    const rcsc::Vector2D & targetPoint() const
      {
          return M_target_point;
      }

    int targetPlayerUnum() const
      {
          return M_target_player_unum;
      }

    double ballSpeed() const
      {
          return M_ball_speed;
      }

    int receiverReachStep() const
      {
          return M_receiver_reach_step;
      }

    int ballReachStep() const
      {
          return M_ball_reach_step;
      }

    const SupportPlayerMovesCont & supportPlayerMoves() const
      {
          return M_support_player_moves;
      }

public:
    typedef boost::shared_ptr< CooperativeAction > Ptr;
};

}

#endif
