// -*-c++-*-

/*!
  \file field_analyzer.cpp
  \brief miscellaneous field analysis Source File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "field_analyzer.h"

#include <rcsc/player/world_model.h>
#include <rcsc/player/player_predicate.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/player_type.h>
#include <rcsc/common/logger.h>
#include <rcsc/timer.h>

//#define DEBUG_PRINT
//#define DEBUG_SELF_PREDICT

using namespace rcsc;

/*-------------------------------------------------------------------*/
/*!

 */
FieldAnalyzer::FieldAnalyzer()
{

}

/*-------------------------------------------------------------------*/
/*!

 */
FieldAnalyzer &
FieldAnalyzer::instance()
{
    static FieldAnalyzer s_instance;
    return s_instance;
}

/*-------------------------------------------------------------------*/
/*!

 */
int
FieldAnalyzer::predict_self_reach_cycle( const WorldModel & wm,
                                         const Vector2D & target_point,
                                         const double & dist_thr,
                                         const int wait_cycle,
                                         const bool save_recovery,
                                         double * stamina )
{
    if ( wm.self().inertiaPoint( wait_cycle ).dist2( target_point ) < std::pow( dist_thr, 2 ) )
    {
        return 0;
    }

    const ServerParam & param = ServerParam::i();
    const PlayerType & ptype = wm.self().playerType();
    const double max_moment = param.maxMoment();
    const double recover_dec_thr = param.recoverDecThrValue();

    const double first_my_speed = wm.self().vel().r() * std::pow( ptype.playerDecay(), wait_cycle );

    double first_my_stamina = wm.self().stamina();
    double first_my_effort = wm.self().effort();
    double first_my_recovery = wm.self().recovery();
    ptype.predictStaminaAfterWait( param,
                                   wait_cycle,
                                   &first_my_stamina,
                                   &first_my_effort,
                                   first_my_recovery );
#ifdef DEBUG_SELF_PREDICT
    dlog.addText( Logger::ACTION,
                  __FILE__" (predict_self_reach_cycle) target=(%.1f %.1f) dist_thr=%.3f wait=%d",
                  target_point.x, target_point.y,
                  dist_thr,
                  wait_cycle );
#endif

    for ( int cycle = wait_cycle + 1; cycle < 25; ++cycle )
    {
        const Vector2D inertia_my_pos = wm.self().inertiaPoint( cycle );
        const double inertia_target_dist2 = inertia_my_pos.dist2( target_point );
        const double inertia_target_dist = std::sqrt( inertia_target_dist2 );

#ifdef DEBUG_SELF_PREDICT
        dlog.addText( Logger::ACTION,
                      __FILE__">>> cycle=%d wait=%d. start. inertia_pos(%.1f %.1f) inertia_dist=%.3f",
                      cycle, wait_cycle,
                      inertia_my_pos.x, inertia_my_pos.y,
                      inertia_target_dist );
#endif

        if ( inertia_target_dist < dist_thr )
        {
#ifdef DEBUG_SELF_PREDICT
    dlog.addText( Logger::ACTION,
                  __FILE__"__ <<< reached only by waiting. inertia_dist=%.3f",
                  inertia_target_dist );
#endif
            return cycle;
        }

        if ( inertia_target_dist - dist_thr > ptype.realSpeedMax() * ( cycle - wait_cycle ) )
        {
#ifdef DEBUG_SELF_PREDICT
        dlog.addText( Logger::ACTION,
                      __FILE__"__ cycle=%d wait=%d. continue. never reach. required_dist=%.3f",
                      cycle, wait_cycle, inertia_target_dist - dist_thr );
#endif
            continue;
        }

        int n_turn = 0;
        int n_dash = 0;

        AngleDeg target_angle = ( target_point - inertia_my_pos ).th();
        double my_speed = first_my_speed;

        //
        // turn
        //

        double angle_diff = ( target_angle - wm.self().body() ).abs();
        double turn_margin = 180.0;
        if ( dist_thr < inertia_target_dist )
        {
            turn_margin = std::max( 15.0,
                                    AngleDeg::asin_deg( dist_thr / inertia_target_dist ) );
        }

        while ( angle_diff > turn_margin )
        {
            angle_diff -= ptype.effectiveTurn( max_moment, my_speed );
            my_speed *= ptype.playerDecay();
            ++n_turn;
        }

        if ( wait_cycle + n_turn >= cycle )
        {
#ifdef DEBUG_SELF_PREDICT
        dlog.addText( Logger::ACTION,
                      __FILE__"__ cycle=%d wait=%d turn=%d, insufficient dash cycle",
                      cycle, wait_cycle, n_turn );
#endif
            continue;
        }

        double my_stamina = first_my_stamina;
        double my_effort = first_my_effort;
        double my_recovery = first_my_recovery;

        AngleDeg dash_angle = wm.self().body();
        if ( n_turn > 0 )
        {
            angle_diff = std::max( 0.0, angle_diff );
            dash_angle = target_angle;
            if ( ( target_angle - wm.self().body() ).degree() > 0.0 )
            {
                dash_angle -= angle_diff;
            }
            else
            {
                dash_angle += angle_diff;
            }

            ptype.predictStaminaAfterWait( param,
                                           n_turn,
                                           &my_stamina,
                                           &my_effort,
                                           my_recovery );
        }

        //
        // dash
        //

        int n_dash_by_table = ptype.cyclesToReachDistance( inertia_target_dist - dist_thr );
        if ( n_dash_by_table > cycle - wait_cycle - n_turn )
        {

#ifdef DEBUG_SELF_PREDICT
            dlog.addText( Logger::ACTION,
                          __FILE__"__ cycle=%d wait=%d turn=%d continue, dash_dist=%.3f n_dash_by_table=%d",
                          cycle, wait_cycle, n_turn,
                          inertia_target_dist, n_dash_by_table );
#endif
            continue;
        }

        Vector2D my_pos = wm.self().inertiaPoint( wait_cycle + n_turn );
        Vector2D my_vel = wm.self().vel() * std::pow( ptype.playerDecay(), wait_cycle + n_turn );
        double target_dist = inertia_target_dist;

#ifdef DEBUG_SELF_PREDICT
        dlog.addText( Logger::ACTION,
                      __FILE__"__ cycle=%d wait=%d turn=%d  start_pos=(%.1f %.1f) vel=(%.1f %.1f)"
                      " required_dist=%.3f",
                      cycle, wait_cycle, n_turn,
                      my_pos.x, my_pos.y,
                      my_vel.x, my_vel.y,
                      target_dist );
#endif

        while ( wait_cycle + n_turn + n_dash < cycle )
        {
            double dash_power = std::min( param.maxDashPower(), my_stamina );
            if ( save_recovery
                 && my_stamina - dash_power < recover_dec_thr )
            {
                dash_power = std::max( 0.0, my_stamina - recover_dec_thr );
                if ( dash_power < 1.0 )
                {
                    break;
                }
            }

            Vector2D accel = Vector2D::polar2vector( dash_power
                                                     * ptype.dashPowerRate()
                                                     * my_effort,
                                                     dash_angle );
            my_vel += accel;

            double speed = my_vel.r();
            if ( speed > ptype.playerSpeedMax() )
            {
                my_vel *= ptype.playerSpeedMax() / speed;
            }

            my_pos += my_vel;
            my_vel *= ptype.playerDecay();

            ptype.predictStaminaAfterOneDash( param,
                                              param.maxDashPower(),
                                              &my_stamina,
                                              &my_effort,
                                              &my_recovery );
            target_dist = my_pos.dist( target_point );
            ++n_dash;

#ifdef DEBUG_SELF_PREDICT
            dlog.addText( Logger::ACTION,
                          __FILE__"____ dash=%d pos=(%.1f %.1f) vel=(%.1f %.1f)"
                          " target_dist=%.3f stamina=%.1f"
                          " first_dist=%.3f move=%.3f",
                          n_dash,
                          my_pos.x, my_pos.y,
                          my_vel.x, my_vel.y,
                          target_dist,
                          my_stamina,
                          inertia_target_dist, inertia_my_pos.dist( my_pos ) );
#endif

            if ( target_dist < dist_thr
                 || inertia_target_dist2 < inertia_my_pos.dist2( my_pos ) ) // run over
            {
                target_dist = -1.0;
                break;
            }
        }

        if ( target_dist <= dist_thr )
        {
            if ( stamina )
            {
                *stamina = my_stamina;
            }
            return wait_cycle + n_turn + n_dash;
        }
    }

    return 1000;
}



/*-------------------------------------------------------------------*/
/*!

 */
int
FieldAnalyzer::predit_player_turn_cycle( const WorldModel & wm,
                                         const AbstractPlayerObject * player,
                                         const Vector2D & target_point,
                                         const int total_cycle,
                                         const double & dist_thr )
{
    if ( player->bodyCount() > 1 )
    {
        if ( player->side() == wm.ourSide() )
        {
            return 2;
        }
        else
        {
            return 0;
        }
    }

    const Vector2D player_pos = player->inertiaPoint( total_cycle );
    const PlayerType * player_type = player->playerTypePtr();
    const double max_moment = ServerParam::i().maxMoment();

    const double target_dist = ( target_point - player_pos ).r();
    const AngleDeg target_angle = ( target_point - player_pos ).th();

    double angle_diff = ( target_angle - player->body() ).abs();
    double turn_margin = 180.0;
    if ( dist_thr < target_dist )
    {
        turn_margin = std::max( 15.0,
                                AngleDeg::asin_deg( dist_thr / target_dist ) );
    }

    int n_turn = 0;

    double player_speed = player->vel().r();
    while ( angle_diff > turn_margin )
    {
        angle_diff -= player_type->effectiveTurn( max_moment, player_speed );
        player_speed *= player_type->playerDecay();
        ++n_turn;
    }

    return n_turn;
}

/*-------------------------------------------------------------------*/
/*!

 */
int
FieldAnalyzer::predit_player_turn_cycle( const WorldModel & wm,
                                         const AbstractPlayerObject * player,
                                         const AngleDeg & target_angle,
                                         const double & target_dist,
                                         const double & dist_thr )
{
    if ( player->bodyCount() > 1 )
    {
        if ( player->side() == wm.ourSide() )
        {
            return 2;
        }
        else
        {
            return 0;
        }
    }

    const PlayerType * player_type = player->playerTypePtr();
    const double max_moment = ServerParam::i().maxMoment();

    double angle_diff = ( target_angle - player->body() ).abs();
    double turn_margin = 180.0;
    if ( dist_thr < target_dist )
    {
        turn_margin = std::max( 15.0,
                                AngleDeg::asin_deg( dist_thr / target_dist ) );
    }

    int n_turn = 0;

    double player_speed = player->vel().r();
    while ( angle_diff > turn_margin )
    {
        angle_diff -= player_type->effectiveTurn( max_moment, player_speed );
        player_speed *= player_type->playerDecay();
        ++n_turn;
    }

    return n_turn;
}


/*-------------------------------------------------------------------*/
/*!

 */
int
FieldAnalyzer::predict_player_reach_cycle( const WorldModel & wm,
                                           const AbstractPlayerObject * player,
                                           const Vector2D & target_point,
                                           const double & dist_thr,
                                           const int wait_cycle )
{
//     if ( player->pos().dist2( target_point ) < std::pow( dist_thr, 2 ) )
//     {
//         return 0;
//     }

    const ServerParam & SP = ServerParam::i();
    const PlayerType * ptype = player->playerTypePtr();
    const double max_moment = SP.maxMoment();

    const double first_player_speed = player->vel().r() * std::pow( ptype->playerDecay(), wait_cycle );

    const int pos_count = ( player->side() == wm.ourSide()
                            ? std::min( 3, player->posCount() )
                            : std::max( -3, - player->posCount() ) );

    const AngleDeg player_body = player->body();
    const int player_body_count = ( ( player == &(wm.self()) )
                                    || ( player->bodyCount() > 0
                                         && player->seenPosCount() == 0
                                         && wm.self().pos().dist2( player->pos() )
                                         < std::pow( SP.visibleDistance(), 2 )
                                         && player->velCount() <= 3
                                         && player->vel().r2() > 0.1*0.1 )
                                    ? 0
                                    : player->bodyCount() );

    const int start_cycle = std::max( 0, wait_cycle );
    const int max_cycle = 6;
    for ( int cycle = start_cycle; cycle <= max_cycle; ++cycle )
    {
        Vector2D player_pos = player->inertiaPoint( cycle );
        double target_dist = player_pos.dist( target_point );

        if ( target_dist < dist_thr )
        {
            return cycle;
        }

        if ( target_dist - dist_thr > ptype->realSpeedMax() * cycle )
        {
            continue;
        }

        const int n_dash = ptype->cyclesToReachDistance( target_dist - dist_thr );
        if ( n_dash > cycle - wait_cycle )
        {
            continue;
        }

        int n_turn = 0;

        AngleDeg target_angle = ( target_point - player_pos ).th();
        double player_speed = first_player_speed;

        //
        // turn
        //
        if ( player_body_count == 0 )
        {
            double angle_diff = ( target_angle - player_body ).abs();
            double turn_margin = 180.0;
            if ( dist_thr < target_dist )
            {
                turn_margin = std::max( 15.0,
                                        AngleDeg::asin_deg( dist_thr / target_dist ) );
            }

            while ( angle_diff > turn_margin )
            {
                angle_diff -= ptype->effectiveTurn( max_moment, player_speed );
                player_speed *= ptype->playerDecay();
                ++n_turn;
            }
        }
        else if ( player->side() == wm.ourSide() )
        {
            n_turn = 2;
        }
        else
        {
            n_turn = 0;
        }

        if ( wait_cycle + n_turn + n_dash + pos_count <= cycle )
        {
            return wait_cycle + n_turn + n_dash + pos_count;
        }
    }

    //
    // rough estimation
    //

    Vector2D player_pos = player->inertiaFinalPoint();
    double target_dist = ( target_point - player_pos ).r();

    int n_turn = 0;
    if ( player->bodyCount() == 0 )
    {
        AngleDeg target_angle = ( target_point - player_pos ).th();
        double player_speed = first_player_speed;

        double angle_diff = ( target_angle - player->body() ).abs();
        double turn_margin = 180.0;
        if ( dist_thr < target_dist )
        {
            turn_margin = std::max( 15.0,
                                    AngleDeg::asin_deg( dist_thr / target_dist ) );
        }

        while ( angle_diff > turn_margin )
        {
            angle_diff -= ptype->effectiveTurn( max_moment, player_speed );
            player_speed *= ptype->playerDecay();
            ++n_turn;
        }
    }
    else if ( player->side() == wm.ourSide() )
    {
        n_turn = 2;
    }
    else
    {
        n_turn = 0;
    }

    int n_dash = ptype->cyclesToReachDistance( target_dist - dist_thr );
    return wait_cycle + n_turn + n_dash + pos_count;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
FieldAnalyzer::update( const WorldModel & wm )
{
    static GameTime s_update_time( 0, 0 );

    if ( s_update_time == wm.time() )
    {
        return;
    }
    s_update_time = wm.time();

#ifdef DEBUG_PRINT
    Timer timer;
#endif

    updateVoronoiDiagram( wm );
    updatePlayerGraph( wm );

#ifdef DEBUG_PRINT
    dlog.addText( Logger::TEAM,
                  "FieldAnalyzer::update() elapsed %f [ms]",
                  timer.elapsedReal() );
    writeDebugLog();
#endif

}

/*-------------------------------------------------------------------*/
/*!

 */
void
FieldAnalyzer::updateVoronoiDiagram( const WorldModel & wm )
{
    const Rect2D rect = Rect2D::from_center( 0.0, 0.0,
                                             ServerParam::i().pitchLength() - 10.0,
                                             ServerParam::i().pitchWidth() - 10.0 );

    M_all_players_voronoi_diagram.clear();
    M_teammates_voronoi_diagram.clear();
    M_pass_voronoi_diagram.clear();

    const SideID our = wm.ourSide();

    const AbstractPlayerCont::const_iterator end = wm.allPlayers().end();
    for ( AbstractPlayerCont::const_iterator p = wm.allPlayers().begin();
          p != end;
          ++p )
    {
        M_all_players_voronoi_diagram.addPoint( (*p)->pos() );

        if ( (*p)->side() == our )
        {
            M_teammates_voronoi_diagram.addPoint( (*p)->pos() );
        }
        else
        {
            M_pass_voronoi_diagram.addPoint( (*p)->pos() );
        }
    }

    // our goal
    M_pass_voronoi_diagram.addPoint( Vector2D( - ServerParam::i().pitchHalfLength() + 5.5, 0.0 ) );
//     M_pass_voronoi_diagram.addPoint( Vector2D( - ServerParam::i().pitchHalfLength() + 5.5,
//                                                - ServerParam::i().goalHalfWidth() ) );
//     M_pass_voronoi_diagram.addPoint( Vector2D( - ServerParam::i().pitchHalfLength() + 5.5,
//                                                + ServerParam::i().goalHalfWidth() ) );

    // opponent side corners
    M_pass_voronoi_diagram.addPoint( Vector2D( + ServerParam::i().pitchHalfLength() + 10.0,
                                               - ServerParam::i().pitchHalfWidth() - 10.0 ) );
    M_pass_voronoi_diagram.addPoint( Vector2D( + ServerParam::i().pitchHalfLength() + 10.0,
                                               + ServerParam::i().pitchHalfWidth() + 10.0 ) );

    M_pass_voronoi_diagram.setBoundingRect( rect );

    M_all_players_voronoi_diagram.compute();
    M_teammates_voronoi_diagram.compute();
    M_pass_voronoi_diagram.compute();
}

/*-------------------------------------------------------------------*/
/*!

 */
void
FieldAnalyzer::updatePlayerGraph( const WorldModel & wm )
{
    //
    // update field player graph
    //
    M_field_player_graph.setPredicate( new AndPlayerPredicate
                                       ( new FieldPlayerPredicate(),
                                         new XCoordinateForwardPlayerPredicate( wm.ball().pos().x - 50.0 ),
                                         new XCoordinateBackwardPlayerPredicate( wm.ball().pos().x + 50.0 ) ) );
    M_field_player_graph.update( wm );
}

/*-------------------------------------------------------------------*/
/*!

 */
void
FieldAnalyzer::writeDebugLog()
{

    if ( dlog.isEnabled( Logger::PASS ) )
    {
        const VoronoiDiagram::Segment2DCont::const_iterator s_end = M_pass_voronoi_diagram.resultSegments().end();
        for ( VoronoiDiagram::Segment2DCont::const_iterator s = M_pass_voronoi_diagram.resultSegments().begin();
              s != s_end;
              ++s )
        {
            dlog.addLine( Logger::PASS,
                          s->origin(), s->terminal(),
                          "#0000ff" );
        }

        const VoronoiDiagram::Ray2DCont::const_iterator r_end = M_pass_voronoi_diagram.resultRays().end();
        for ( VoronoiDiagram::Ray2DCont::const_iterator r = M_pass_voronoi_diagram.resultRays().begin();
              r != r_end;
              ++r )
        {
            dlog.addLine( Logger::PASS,
                          r->origin(), r->origin() + Vector2D::polar2vector( 20.0, r->dir() ),
                          "#0000ff" );
        }
    }

    if ( dlog.isEnabled( Logger::MARK ) )
    {
        //
        // field player graph
        //
        const std::vector< PlayerGraph::Connection >::const_iterator e_end = M_field_player_graph.connections().end();
        for ( std::vector< PlayerGraph::Connection >::const_iterator e = M_field_player_graph.connections().begin();
              e != e_end;
              ++e )
        {
            dlog.addLine( Logger::MARK,
                          e->first_->pos(), e->second_->pos(),
                          "#ff00ff" );
        }
    }
}
