// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifndef RCSC_PLAYER_PREDICT_STATE_H
#define RCSC_PLAYER_PREDICT_STATE_H

/////////////////////////////////////////////////////////////////////

#include <rcsc/player/world_model.h>
#include <rcsc/player/abstract_player_object.h>
#include <rcsc/player/predict_player_object.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/vector_2d.h>
#include <rcsc/player/player_predicate.h>

#include <boost/shared_ptr.hpp>
#include <vector>
#include <algorithm>
#include <iostream>

namespace rcsc {

class PredictState /* : public AbstractWorldModel */ {

public:
    typedef std::vector< boost::shared_ptr< PredictPlayerObject > > PlayerPtrCont;

public:
    static const int VALID_PLAYER_THRESHOLD = 8;

private:
    class PredictBallObject {

    private:
        rcsc::Vector2D M_pos;
        rcsc::Vector2D M_vel;

    public:
        PredictBallObject( const rcsc::Vector2D & pos,
                           const rcsc::Vector2D & vel )
            : M_pos( pos )
            , M_vel( vel )
          { }

        const rcsc::Vector2D & pos() const
          {
              return M_pos;
          }

        const rcsc::Vector2D & vel() const
          {
              return M_vel;
          }
    };

private:
    const WorldModel * M_world;
    unsigned long M_spend_time;

    int M_ball_holder_unum;
    rcsc::Vector2D M_ball_pos;
    rcsc::Vector2D M_ball_vel;

    int M_self_unum;

    PlayerPtrCont M_all_teammates;

protected:
    void initializeWithWorldModel( const WorldModel & wm )
    {
        //
        // initialize world
        //
        M_world = &wm;


        //
        // initialize spend_time
        //
        M_spend_time = 0;


        //
        // initialize ball pos & vel
        //
        M_ball_pos = wm.ball().pos();
        M_ball_vel = wm.ball().vel();


        //
        // initialize self_unum
        //
        M_self_unum = wm.self().unum();


        //
        // initialize ball holder
        //
        const AbstractPlayerObject * h = wm.getTeammateNearestToBall
                                            ( VALID_PLAYER_THRESHOLD );

        if ( h && ( h->pos() - wm.ball().pos() ).r()
                  < ( wm.self().pos() - wm.ball().pos() ).r() )
        {
            M_ball_holder_unum = h->unum();
        }
        else
        {
            M_ball_holder_unum = wm.self().unum();
        }

        //
        // initialize all teammates
        //
        for ( int n = 1; n <= ServerParam::i().maxPlayer(); ++n )
        {
            if ( n == M_self_unum )
            {
                M_all_teammates.push_back
                    ( boost::shared_ptr< PredictPlayerObject >
                      ( new PredictPlayerObject( wm.self() ) ) );
            }
            else
            {
                const AbstractPlayerObject * t = wm.teammate( n );

                if ( t )
                {
                    M_all_teammates.push_back
                        ( boost::shared_ptr< PredictPlayerObject >
                          ( new PredictPlayerObject( *t ) ) );
                }
                else
                {
                    M_all_teammates.push_back
                        ( boost::shared_ptr< PredictPlayerObject>
                          ( new PredictPlayerObject() ) );
                }
            }
        }
    }


public:
    PredictState( const PredictState & rhs,
                  unsigned long append_spend_time = 0 )
        : M_world( rhs.M_world )
        , M_spend_time( rhs.spendTime() + append_spend_time )
        , M_ball_holder_unum( rhs.ballHolderUnum() )
        , M_ball_pos( rhs.ballPos() )
        , M_ball_vel( rhs.M_ball_vel )
        , M_self_unum( rhs.M_self_unum )
        , M_all_teammates( rhs.M_all_teammates )
      { }

    PredictState( const WorldModel & wm,
                  unsigned long append_spend_time = 0 )
        : M_world( &wm )
        , M_spend_time( append_spend_time )
        , M_ball_holder_unum( rcsc::Unum_Unknown )
        , M_ball_pos()
        , M_ball_vel()
        , M_self_unum( rcsc::Unum_Unknown )
        , M_all_teammates()
      {
          initializeWithWorldModel( wm );
      }

    PredictState( const PredictState & rhs,
                  unsigned long append_spend_time,
                  int ball_holder_unum,
                  const rcsc::Vector2D & ball_pos )
        : M_world( rhs.M_world )
        , M_spend_time( rhs.M_spend_time + append_spend_time )
        , M_ball_holder_unum( ball_holder_unum )
        , M_ball_pos( ball_pos )
        , M_ball_vel()
        , M_self_unum( rhs.M_self_unum )
        , M_all_teammates( rhs.M_all_teammates )
      {
          M_all_teammates[ ball_holder_unum - 1 ]
              = boost::shared_ptr< PredictPlayerObject >
                ( new PredictPlayerObject
                  ( *(M_all_teammates[ ball_holder_unum - 1 ] ),
                    ball_pos ) );
      }

    PredictState( const WorldModel & wm,
                  unsigned long append_spend_time,
                  int ball_holder_unum,
                  const rcsc::Vector2D & ball_pos )
        : M_world( &wm )
        , M_spend_time( 0 )
        , M_ball_holder_unum( rcsc::Unum_Unknown )
        , M_ball_pos()
        , M_ball_vel()
        , M_self_unum( rcsc::Unum_Unknown )
        , M_all_teammates()
      {
          initializeWithWorldModel( wm );

          M_spend_time = append_spend_time;
          M_ball_holder_unum = ball_holder_unum;
          M_ball_pos = ball_pos;
          M_all_teammates[ ball_holder_unum - 1 ]
              = boost::shared_ptr< PredictPlayerObject >
                ( new PredictPlayerObject
                  ( *(M_all_teammates[ ball_holder_unum - 1 ] ),
                    ball_pos ) );
      }

    PredictState( const PredictState & rhs,
                  unsigned long append_spend_time,
                  const rcsc::Vector2D & ball_pos )
        : M_world( rhs.M_world )
        , M_spend_time( rhs.M_spend_time + append_spend_time )
        , M_ball_holder_unum( rhs.M_ball_holder_unum )
        , M_ball_pos( ball_pos )
        , M_ball_vel()
        , M_self_unum( rhs.M_self_unum )
        , M_all_teammates( rhs.M_all_teammates )
      {
      }


    unsigned long spendTime() const
      {
          return M_spend_time;
      }

    int ballHolderUnum() const
      {
          return M_ball_holder_unum;
      }

    const AbstractPlayerObject * ballHolder() const
      {
          return teammate( ballHolderUnum() );
      }

    PredictBallObject ball() const
      {
          return PredictBallObject( M_ball_pos, M_ball_vel );
      }

    const Vector2D & ballPos() const
      {
          return M_ball_pos;
      }

    const
    AbstractPlayerObject &
    self() const
      {
          if ( ! ( 1 <= M_self_unum
                   && M_self_unum <= ServerParam::i().maxPlayer() ) )
          {
              std::cerr << "internal error: "
                        << __FILE__ << ":" << __LINE__
                        << "invalid self unum " << M_self_unum << std::endl;
          }

          return *(M_all_teammates[ M_self_unum - 1 ]);
      }

    const
    AbstractPlayerObject * teammate( const int unum ) const
      {
          if ( ! ( 1 <= unum && unum <= ServerParam::i().maxPlayer() ) )
          {
              std::cerr << "internal error: "
                        << __FILE__ << ":" << __LINE__ << ": "
                        << "invalid unum " << unum << std::endl;
          }

          return &( *(M_all_teammates[ unum - 1 ]) );
      }

    const
    AbstractPlayerObject * opponent( const int unum ) const
      {
          return M_world->opponent( unum );
      }

    const
    PlayerPtrCont &
    allTeammates() const
      {
          return M_all_teammates;
      }

    const rcsc::PlayerCont & opponents() const
      {
          return M_world->opponents();
      }

    const rcsc::PlayerPtrCont & opponentsFromSelf() const
      {
          return M_world->opponentsFromSelf();
      }

    const PlayerObject * getOpponentNearestTo( const Vector2D & point,
                                               const int count_thr,
                                               double * dist_to_point ) const
      {
          return M_world->getOpponentNearestTo( point, count_thr, dist_to_point );
      }

    const
    rcsc::PlayerType * teammatePlayerType( const int unum ) const
      {
          return M_world->teammatePlayerType( unum );
      }

    const 
    rcsc::AbstractPlayerObject * getTeammateGoalie() const
      {
          if ( M_world->teammateGoalieUnum() == rcsc::Unum_Unknown )
          {
              return static_cast< rcsc::AbstractPlayerObject * >( 0 );
          }

          return teammate( M_world->teammateGoalieUnum() );
      }

    const 
    rcsc::AbstractPlayerObject * getOpponentGoalie() const
      {
          return M_world->getOpponentGoalie();
      }

    const
    rcsc::GameMode & gameMode() const
      {
        return M_world->gameMode();
      }

    const
    rcsc::GameTime & time() const
      {
        return M_world->time();
      }


    rcsc::SideID ourSide() const
      {
          return M_world->ourSide();
      }

    double offsideLineX() const
      {
          return std::max( M_world->offsideLineX(), M_ball_pos.x );
      }

    double defenseLineX() const
      {
          // XXX: tentative implementation, should consider M_all_teammates
          return std::min( M_world->defenseLineX(), M_ball_pos.x );
      }

    int dirCount( const AngleDeg & angle )
      {
          return M_world->dirCount( angle );
      }

    int dirRangeCount( const AngleDeg & angle,
                       const double & width,
                       int * max_count,
                       int * sun_count,
                       int * ave_count ) const
      {
          return M_world->dirRangeCount( angle, width,
                                         max_count, sun_count, ave_count );
      }

    const
    AudioMemory & audioMemory() const
      {
          return M_world->audioMemory();
      }

    AbstractPlayerCont getPlayerCont( const PlayerPredicate * predicate ) const
      {
          AbstractPlayerCont ret;

          if ( ! predicate )
          {
              return ret;
          }

          const PlayerPtrCont::const_iterator t_end = M_all_teammates.end();

          for ( PlayerPtrCont::const_iterator it = M_all_teammates.begin();
                it != t_end;
                ++it )
          {
              if ( (*predicate)( **it ) )
              {
                  ret.push_back( &(**it) );
              }
          }

          const rcsc::PlayerPtrCont::const_iterator o_end = M_world->opponentsFromSelf().end();
          for ( rcsc::PlayerPtrCont::const_iterator it = M_world->opponentsFromSelf().begin();
                it != o_end;
                ++it )
          {
              if ( (*predicate)( **it ) )
              {
                  ret.push_back( *it );
              }
          }

          delete predicate;

          return ret;
      }
};

}

#endif
