// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "recursive_action_chain_graph.h"
#include <rcsc/player/player_agent.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/time/timer.h>
#include <cfloat>
#include <string>
#include <sstream>
#include <cstdio>

#if 0
#define DO_MONTECARLO_SEARCH
#endif

//#define RECURSIVE_ACTION_CHAIN_DEBUG
//#define RECURSIVE_ACTION_CHAIN_LOAD_DEBUG

#ifdef RECURSIVE_ACTION_CHAIN_LOAD_DEBUG
#include <iostream>
#endif

namespace rcsc {

const size_t RecursiveActionChainGraph::DEFAULT_MAX_RECURSIVE_LEVEL = 2;
//const size_t RecursiveActionChainGraph::DEFAULT_MAX_EVALUATE_LIMIT = 1000;
const size_t RecursiveActionChainGraph::DEFAULT_MAX_EVALUATE_LIMIT = 500;

void
RecursiveActionChainGraph::calculateResult( const WorldModel & wm )
{
#ifdef RECURSIVE_ACTION_CHAIN_DEBUG
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE, "=====" );

    for ( int n = 1; n <= ServerParam::i().maxPlayer(); ++n )
    {
        if ( n == wm.self().unum() )
        {
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                "n = %d, self", n );
        }
        else
        {
            const AbstractPlayerObject * t = wm.teammate( n );

            if ( t )
            {
                rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                    "n = %d, OK", n );
            }
            else
            {
                rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                    "n = %d, null", n );
            }
        }
    }
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE, "=====" );


    AbstractPlayerCont::const_iterator end = wm.allTeammates().end();
    for ( AbstractPlayerCont::const_iterator it = wm.allTeammates().begin();
          it != end;
          ++it )
    {
        if ( (*it)->isSelf() )
        {
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                "teammate %d, self", (*it)->unum() );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                "teammate %d", (*it)->unum() );
        }
    }

    rcsc::dlog.addText( rcsc::Logger::RECURSIVE, "=====" );
#endif


    M_result.clear();

    PredictState current_state( wm );

    std::vector< ActionStatePair > empty_path;
    double evaluation;
    unsigned long n_evaluated = 0;

#ifdef RECURSIVE_ACTION_CHAIN_LOAD_DEBUG
    rcsc::Timer stopwatch;
#endif
    doRecursiveSearch( wm, current_state, empty_path,
                       &M_result, &evaluation, &n_evaluated,
                       M_max_recursive_level,
                       M_max_evaluate_limit );

#ifdef RECURSIVE_ACTION_CHAIN_LOAD_DEBUG
    const double msec = stopwatch.elapsedReal();
    std::fprintf( stderr,
                  "# recursive search: %2d [%ld] % 10lf msec, n=%4lu, average=% 10lf\n",
                  wm.self().unum(),
                  wm.time().cycle(),
                  msec, n_evaluated,
                  ( n_evaluated == 0 ? 0.0 : msec / n_evaluated ) );
#endif

    if ( M_result.empty() )
    {
        M_result.push_back( ActionStatePair
                            ( new CooperativeAction
                              ( CooperativeAction::ActionType::Hold,
                                true,
                                "hold" ),
                              new PredictState( current_state, 1 ) ) );
    }

    writeChainInfoLog( "best chain: ", wm, M_result, evaluation );
}


bool
RecursiveActionChainGraph::doRecursiveSearch( const WorldModel & wm,
                                              const PredictState & state,
                                              const std::vector< ActionStatePair > & path,
                                              std::vector< ActionStatePair > * result,
                                              double * result_evaluation,
                                              unsigned long * n_evaluated,
                                              unsigned long max_recursive_level,
                                              long max_evaluate_limit )
{
    if ( path.size() > max_recursive_level )
    {
        return false;
    }


    //
    // check evaluation limit
    //
    if ( max_evaluate_limit != -1
         && *n_evaluated >= static_cast< unsigned int >( max_evaluate_limit ) )
    {
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "cut by max evaluate limit %d", *n_evaluated );
#if 0
#if defined( RECURSIVE_ACTION_CHAIN_DEBUG ) || defined( RECURSIVE_ACTION_CHAIN_LOAD_DEBUG )
        std::cerr << "Max evaluate limit " << max_evaluate_limit
                  << " exceeded" << std::endl;
#endif
#endif
        return false;
    }


    //
    // check current state
    //
    std::vector< ActionStatePair > best_path = path;
    double max_ev = (*M_evaluator)( state, wm, path );
#ifdef DO_MONTECARLO_SEARCH
    double eval_sum = 0;
    long n_eval = 0;
#endif

#ifdef RECURSIVE_ACTION_CHAIN_DEBUG
    writeChainInfoLog( wm, path, max_ev );
#endif
    ++ (*n_evaluated);


    //
    // generate candidates
    //
    std::vector< ActionStatePair > candidates;
    if ( path.empty()
         || ! ( *( path.rbegin() ) ).action().isFinalAction() )
    {
        M_action_generator->addCandidates( &candidates, state, wm, path );
    }


    //
    // test each candidate
    //
    for ( std::vector< ActionStatePair >::const_iterator it = candidates.begin();
          it != candidates.end(); ++it )
    {
        std::vector< ActionStatePair > new_path = path;
        std::vector< ActionStatePair > candidate_result;

        new_path.push_back( *it );

        double ev;
        const bool exist = doRecursiveSearch( wm,
                                              (*it).state(),
                                              new_path,
                                              &candidate_result,
                                              &ev,
                                              n_evaluated,
                                              max_recursive_level,
                                              max_evaluate_limit );
        if ( ! exist )
        {
            continue;
        }
#ifdef RECURSIVE_ACTION_CHAIN_DEBUG
        writeChainInfoLog( wm, candidate_result, ev );
#endif
        if ( ev > max_ev )
        {
            max_ev = ev;
            best_path = candidate_result;
        }

#ifdef DO_MONTECARLO_SEARCH
        eval_sum += ev;
        ++n_eval;
#endif
    }

    *result = best_path;
#ifdef DO_MONTECARLO_SEARCH
    if ( n_eval == 0 )
    {
        *result_evaluation = max_ev;
    }
    else
    {
        *result_evaluation = eval_sum / n_eval;
    }
#else
    *result_evaluation = max_ev;
#endif

    return true;
}


void
RecursiveActionChainGraph::setDebugPassRoute( PlayerAgent * agent,
                                              const std::vector< ActionStatePair > & path )
{
    const double DIRECT_PASS_DIST = 3.0;

    const PredictState current_state( agent->world() );

    for ( size_t i = 0; i < path.size(); ++i )
    {
        std::string action_string = "?";
        const CooperativeAction & action = path[i].action();
        const PredictState * s0;
        const PredictState * s1;

        if ( i == 0 )
        {
            s0 = &current_state;
            s1 = &(path[0].state());
        }
        else
        {
            s0 = &(path[i - 1].state());
            s1 = &(path[i].state());
        }


        switch( action.type() )
        {
        case rcsc::CooperativeAction::ActionType::Hold:
            {
                action_string = "hold";
                break;
            }

        case rcsc::CooperativeAction::ActionType::Dribble:
            {
                action_string = "dribble";

                agent->debugClient().addLine( s0->ballPos(), s1->ballPos() );
                break;
            }


        case rcsc::CooperativeAction::ActionType::Pass:
            {
                action_string = "pass";

                if ( i == 0
                     && s0->ballHolderUnum() == agent->world().self().unum() )
                {
                    agent->debugClient().setTarget( action.targetPlayerUnum() );

                    if ( (s1->ballPos() - s0->teammate( action.targetPlayerUnum() )->pos() ).r() > DIRECT_PASS_DIST )
                    {
                        agent->debugClient().addLine( s0->ballPos(), s1->ballPos() );
                    }
                }
                else
                {
                    agent->debugClient().addLine( s0->ballPos(), s1->ballPos() );
                }

                break;
            }

        case rcsc::CooperativeAction::ActionType::Shoot:
            {
                action_string = "shoot";

                agent->debugClient().addLine
                    ( s0->ballPos(),
                      rcsc::Vector2D( rcsc::ServerParam::i().pitchHalfLength(),
                                      0.0 ) );

                break;
            }

        default:
            {
                action_string = "?";

                break;
            }
        }

        const CooperativeAction::SupportPlayerMovesCont::const_iterator end = action.supportPlayerMoves().end();
        for ( CooperativeAction::SupportPlayerMovesCont::const_iterator it = action.supportPlayerMoves().begin();
              it != end;
              ++it)
        {
            agent->debugClient().addLine
                ( current_state.teammate( (*it).unum() )->pos(),
                  (*it).targetPoint() );

            agent->debugClient().addLine
                ( current_state.teammate( (*it).unum() )->pos() + rcsc::Vector2D( 0.0, + 2.0 ),
                  (*it).targetPoint() );
            agent->debugClient().addLine
                ( current_state.teammate( (*it).unum() )->pos() + rcsc::Vector2D( 0.0, - 2.0 ),
                  (*it).targetPoint() );
        }

        if ( action.tag() )
        {
            action_string += '|';
            action_string += action.tag();
        }

        agent->debugClient().addMessage( "%s", action_string.c_str() );

        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "chain %d %s, time = %lu",
                            i, action_string.c_str(), s1->spendTime() );
    }
}


void
RecursiveActionChainGraph::writeChainInfoLog( const rcsc::WorldModel & world,
                                              const std::vector< ActionStatePair > & path,
                                              double eval )
{
    writeChainInfoLog( "", world, path, eval );
}

void
RecursiveActionChainGraph::writeChainInfoLog( const std::string & pre_log_message,
                                              const WorldModel & world,
                                              const std::vector< ActionStatePair > & path,
                                              double eval )
{
    std::stringstream buf;

    buf << pre_log_message;
    const PredictState current_state( world );

    for ( size_t i = 0; i < path.size(); ++i )
    {
        buf << '(' << i << ": ";

        std::string action_string = "?";
        const CooperativeAction & a = path[i].action();
        const PredictState * s0;
        const PredictState * s1;

        if ( i == 0 )
        {
            s0 = &current_state;
            s1 = &(path[0].state());
        }
        else
        {
            s0 = &(path[i - 1].state());
            s1 = &(path[i].state());
        }


        switch( a.type() )
        {
        case rcsc::CooperativeAction::ActionType::Hold:
            {
                buf << "hold";
                break;
            }

        case rcsc::CooperativeAction::ActionType::Dribble:
            {
                buf << "dribble unum " << s0->ballHolderUnum()
                    << " target = [" << s1->ballPos() << "]";

                break;
            }

        case rcsc::CooperativeAction::ActionType::Pass:
            {
                buf << "pass from " << s0->ballHolderUnum()
                    << " to " << s1->ballHolderUnum()
                    << " " << a.targetPoint();
                break;
            }

        case rcsc::CooperativeAction::ActionType::Shoot:
            {
                buf << "shoot unum = " << s0->ballHolderUnum();

                break;
            }

        default:
            {
                buf << "?";

                break;
            }
        }

        if ( a.tag() )
        {
            buf << " " << a.tag();
        }

        buf << " t=" << s1->spendTime();

        buf << ")";

        if ( i + 1 == path.size() )
        {
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                __FILE__": %s: %f", buf.str().c_str(), eval );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                __FILE__": %s", buf.str().c_str());
        }
    }
}



const boost::shared_ptr< const RecursiveActionChainGraph >
RecursiveActionChainGraph::instance( const boost::shared_ptr< const FieldEvaluator > & evaluator,
                                     const boost::shared_ptr< const CooperativeActionCandidateGenerator > & generator,
                                     const WorldModel & wm,
                                     unsigned long max_recursive_level,
                                     long max_evaluate_limit )
{
    static rcsc::GameTime S_cache_time;
    static boost::shared_ptr< RecursiveActionChainGraph > S_cache;
    static boost::shared_ptr< const FieldEvaluator > S_cache_evaluator;
    static boost::shared_ptr< const CooperativeActionCandidateGenerator > S_cache_generator;
    static unsigned long S_cache_max_recursive_level = DEFAULT_MAX_RECURSIVE_LEVEL;
    static long S_cache_max_evaluate_limit = DEFAULT_MAX_EVALUATE_LIMIT;

    if ( wm.time() != S_cache_time
         || evaluator != S_cache_evaluator
         || generator != S_cache_generator
         || max_recursive_level != S_cache_max_recursive_level
         || max_evaluate_limit != S_cache_max_evaluate_limit )
    {
        S_cache_time = wm.time();
        S_cache_evaluator = evaluator;
        S_cache_generator = generator;
        S_cache_max_recursive_level = max_recursive_level;
        S_cache_max_evaluate_limit = max_evaluate_limit;

        S_cache = boost::shared_ptr< RecursiveActionChainGraph >
                  ( new RecursiveActionChainGraph( S_cache_evaluator,
                                                   S_cache_generator,
                                                   wm,
                                                   max_recursive_level,
                                                   max_evaluate_limit ) );
    }

    return S_cache;
}


}
