// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifndef RCSC_RECURSIVE_ACTION_CHAIN_GRAPH_H
#define RCSC_RECURSIVE_ACTION_CHAIN_GRAPH_H

/////////////////////////////////////////////////////////////////////

#include <rcsc/player/cooperative_action.h>
#include <rcsc/player/predict_state.h>
#include <rcsc/player/action_state_pair.h>
#include <rcsc/player/field_evaluator.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/abstract_player_object.h>
#include <rcsc/player/predict_player_object.h>
#include <rcsc/common/server_param.h>
#include <rcsc/game_time.h>

#include <boost/shared_ptr.hpp>
#include <vector>
#include <algorithm>
#include <iostream>

namespace rcsc {

/*!
  \class CooperativeActionCandidateGenerator
  \brief a class to abstruct algorithms which generating action state pairs
*/
class CooperativeActionCandidateGenerator {

protected:
    /*!
      \brief protected constructor to inhibit instantiation of this class
     */
    CooperativeActionCandidateGenerator()
      { }

public:
    /*!
      \brief virtual destructor
     */
    virtual
    ~CooperativeActionCandidateGenerator()
      { }

    /*!
      \brief genarate candidates
      \param state assumption of state to generate action and new state
      \param path sequence of action-state pairs from beginning to `state'
      \param current_wm world model of initial state
      \return generated action-state pairs
     */
    virtual
    void addCandidates( std::vector< ActionStatePair > * result,
                        const PredictState & state,
                        const rcsc::WorldModel & current_wm,
                        const std::vector< ActionStatePair > & path ) const = 0;
};


class CompositeCooperativeActionCandidateGenerator
    : public CooperativeActionCandidateGenerator {

private:
    std::vector< boost::shared_ptr< const CooperativeActionCandidateGenerator > > M_generators;

public:
    void addGenerator( const CooperativeActionCandidateGenerator * g )
      {
          if ( ! g )
          {
              return;
          }

          M_generators.push_back
              ( boost::shared_ptr<const CooperativeActionCandidateGenerator>
                ( g ) );
      }


public:
    /*!
      \brief constructor
      \param g1 generator 1
      \param g2 generator 2
      \param g3 generator 3
      \param g4 generator 4
      \param g5 generator 5
      \param g6 generator 6
      \param g7 generator 7
      \param g8 generator 8
     */
    CompositeCooperativeActionCandidateGenerator
      ( const CooperativeActionCandidateGenerator * g1 = 0,
        const CooperativeActionCandidateGenerator * g2 = 0,
        const CooperativeActionCandidateGenerator * g3 = 0,
        const CooperativeActionCandidateGenerator * g4 = 0,
        const CooperativeActionCandidateGenerator * g5 = 0,
        const CooperativeActionCandidateGenerator * g6 = 0,
        const CooperativeActionCandidateGenerator * g7 = 0,
        const CooperativeActionCandidateGenerator * g8 = 0 )
        : M_generators()
      {
          addGenerator( g1 );
          addGenerator( g2 );
          addGenerator( g3 );
          addGenerator( g4 );
          addGenerator( g5 );
          addGenerator( g6 );
          addGenerator( g7 );
          addGenerator( g8 );
      }

    /*!
      \brief destructor
     */
    ~CompositeCooperativeActionCandidateGenerator()
      { }

    /*!
      \brief genarate candidates
      \param state assumption of world
      \return candidates of action
     */
    void addCandidates( std::vector< ActionStatePair > * result,
                        const PredictState & state,
                        const rcsc::WorldModel & current_wm,
                        const std::vector< ActionStatePair > & path ) const
    {
        for ( size_t i = 0; i < M_generators.size(); ++i )
        {
            M_generators[i] -> addCandidates( result, state, current_wm, path );
        }
    }
};


class RecursiveActionChainGraph {

private:
    static const size_t DEFAULT_MAX_RECURSIVE_LEVEL;
    static const size_t DEFAULT_MAX_EVALUATE_LIMIT;

private:
    boost::shared_ptr< const FieldEvaluator > M_evaluator;
    boost::shared_ptr< const CooperativeActionCandidateGenerator > M_action_generator;

    unsigned long M_max_recursive_level;
    long M_max_evaluate_limit;

private:
    std::vector< ActionStatePair > M_result;

    void calculateResult( const WorldModel & wm );
    bool doRecursiveSearch( const WorldModel & wm,
                            const PredictState & state,
                            const std::vector< ActionStatePair > & path,
                            std::vector< ActionStatePair > * result,
                            double * result_evaluation,
                            unsigned long * n_evaluated,
                            unsigned long max_recursive_level,
                            long max_evaluate_limit );

private:
    /*!
      \brief constructor
      \param evaluator evaluator of each state
      \param generator action and state generator
      \param agent pointer to player agent
     */
    RecursiveActionChainGraph
      ( const boost::shared_ptr< const FieldEvaluator > & evaluator,
        const boost::shared_ptr< const CooperativeActionCandidateGenerator > & generator,
        const WorldModel & wm,
        unsigned long max_recursive_level = DEFAULT_MAX_RECURSIVE_LEVEL,
        long max_evaluate_limit = DEFAULT_MAX_EVALUATE_LIMIT )
        : M_evaluator( evaluator )
        , M_action_generator( generator )
        , M_max_recursive_level( max_recursive_level )
        , M_max_evaluate_limit( max_evaluate_limit )
      {
          calculateResult( wm );
      }

public:
    static
    const boost::shared_ptr< const RecursiveActionChainGraph >
    instance( const boost::shared_ptr< const FieldEvaluator > & evaluator,
              const boost::shared_ptr< const CooperativeActionCandidateGenerator > & generator,
              const WorldModel & wm,
              unsigned long max_recursive_level = DEFAULT_MAX_RECURSIVE_LEVEL,
              long max_evaluate_limit = DEFAULT_MAX_EVALUATE_LIMIT );

    const std::vector< ActionStatePair > & getAllChain() const
      {
          return M_result;
      };

    const CooperativeAction & getFirstAction() const
      {
          return (*(M_result.begin())).action();
      };

    const PredictState & getFirstState() const
      {
          return (*(M_result.begin())).state();
      };

    const PredictState & getFinalResult() const
      {
          return (*(M_result.rbegin())).state();
      };

public:
    static
    void setDebugPassRoute( PlayerAgent * agent,
                            const std::vector< ActionStatePair > & path );

    static
    void writeChainInfoLog( const std::string & pre_log_message,
                            const rcsc::WorldModel & world,
                            const std::vector< ActionStatePair > & path,
                            double eval );

    static
    void writeChainInfoLog( const rcsc::WorldModel & world,
                            const std::vector< ActionStatePair > & path,
                            double eval );
};

}

#endif
