// -*-c++-*-

/*!
  \file strategy.cpp
  \brief team strategy Source File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA, Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "strategy.h"

#include "soccer_role.h"
#include "mark_table.h"

#ifndef USE_GENERIC_FACTORY

#include "role_sample.h"
#include "role_goalie.h"
#include "role_savior.h"

#include "role_442_center_back.h"
#include "role_442_center_half.h"
#include "role_442_forward.h"
#include "role_442_side_back.h"
#include "role_442_side_half.h"

#include "role_center_back.h"
#include "role_center_forward.h"
#include "role_defensive_half.h"
#include "role_offensive_half.h"
#include "role_side_back.h"
#include "role_side_forward.h"

#include <rcsc/formation/formation_dt.h>
#include <rcsc/formation/formation_static.h>

#endif

#include <rcsc/formation/formation.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/world_model.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/param/cmd_line_parser.h>
#include <rcsc/param/param_map.h>
#include <rcsc/game_mode.h>

#include <fstream>
#include <iostream>
#include <cstring>

using namespace rcsc;

const std::string Strategy::BEFORE_KICK_OFF_CONF = "before-kick-off.conf";
const std::string Strategy::NORMAL_FORMATION_CONF = "normal-formation.conf";
const std::string Strategy::DEFENSE_FORMATION_CONF = "defense-formation.conf";
const std::string Strategy::OFFENSE_FORMATION_CONF = "offense-formation.conf";
const std::string Strategy::GOALIE_FORMATION_CONF = "goalie-formation.conf";
const std::string Strategy::GOAL_KICK_OPP_FORMATION_CONF = "goal-kick-opp.conf";
const std::string Strategy::GOAL_KICK_OUR_FORMATION_CONF = "goal-kick-our.conf";
const std::string Strategy::GOALIE_CATCH_OPP_FORMATION_CONF = "goalie-catch-opp.conf";
const std::string Strategy::GOALIE_CATCH_OUR_FORMATION_CONF = "goalie-catch-our.conf";
const std::string Strategy::KICKIN_OUR_FORMATION_CONF = "kickin-our-formation.conf";
const std::string Strategy::SETPLAY_OPP_FORMATION_CONF = "setplay-opp-formation.conf";
const std::string Strategy::SETPLAY_OUR_FORMATION_CONF = "setplay-our-formation.conf";
const std::string Strategy::INDIRECT_FREEKICK_OPP_FORMATION_CONF = "indirect-freekick-opp-formation.conf";
const std::string Strategy::INDIRECT_FREEKICK_OUR_FORMATION_CONF = "indirect-freekick-our-formation.conf";

/*-------------------------------------------------------------------*/
/*!

*/
Strategy::Strategy()
    : M_current_situation( Normal_Situation )
    , M_position_types( 11, Position_Center )
    , M_positions( 11 )
    , M_mark_table( new MarkTable() )
    , M_opponent_offense_strategy( No_Strategy )
    , M_opponent_defense_strategy( No_Strategy )
    , M_field_evaluator()
    , M_action_generator()
    , M_savior( true )
{
#ifndef USE_GENERIC_FACTORY

    //
    // roles
    //

    M_role_factory[RoleSample::name()] = &RoleSample::create;

    M_role_factory[RoleGoalie::name()] = &RoleGoalie::create;
    M_role_factory[RoleSavior::name()] = &RoleSavior::create;

    // 433
    M_role_factory[RoleCenterBack::name()] = &RoleCenterBack::create;
    M_role_factory[RoleSideBack::name()] = &RoleSideBack::create;
    M_role_factory[RoleDefensiveHalf::name()] = &RoleDefensiveHalf::create;
    M_role_factory[RoleOffensiveHalf::name()] = &RoleOffensiveHalf::create;
    M_role_factory[RoleSideForward::name()] = &RoleSideForward::create;
    M_role_factory[RoleCenterForward::name()] = &RoleCenterForward::create;

    // 442
    M_role_factory[Role442CenterBack::name()] = &Role442CenterBack::create;
    M_role_factory[Role442SideBack::name()] = &Role442SideBack::create;
    M_role_factory[Role442CenterHalf::name()] = &Role442CenterHalf::create;
    M_role_factory[Role442SideHalf::name()] = &Role442SideHalf::create;
    M_role_factory[Role442Forward::name()] = &Role442Forward::create;

    M_role_factory["442OffensiveHalf"] = &RoleOffensiveHalf::create;
    M_role_factory["442Sweeper"] = &Role442CenterBack::create;
    M_role_factory["442Stopper"] = &Role442CenterBack::create;

    //
    // formations
    //

    M_formation_factory[rcsc::FormationDT::name()] = &rcsc::FormationDT::create;
    M_formation_factory[rcsc::FormationStatic::name()] = &rcsc::FormationStatic::create;
#endif
}

/*-------------------------------------------------------------------*/
/*!

*/
Strategy &
Strategy::instance()
{
    static Strategy s_instance;
    return s_instance;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Strategy::init( rcsc::CmdLineParser & cmd_parser )
{
    rcsc::ParamMap param_map( "HELIOS options" );
    param_map.add()
        ( "savior", "", &M_savior, "Savior is used as the goalie player." );

    //
    //
    //

    if ( cmd_parser.count( "help" ) > 0 )
    {
        param_map.printHelp( std::cout );
        return false;
    }

    //
    //
    //

    cmd_parser.parse( param_map );

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Strategy::read( const std::string & config_dir )
{
    static bool s_initialized = false;

    if ( s_initialized )
    {
        std::cerr << __FILE__ << ' ' << __LINE__ << ": already initialized."
                  << std::endl;
        return false;
    }

#if 0
#ifdef USE_GENERIC_FACTORY
    std::cerr << "Registered formations:\n";
    std::list< std::string > formation_names = rcsc::Formation::creators().list();
    for ( std::list< std::string >::iterator it = formation_names.begin();
          it != formation_names.end();
          ++it )
    {
        std::cerr << "  " << *it << '\n';
    }
    std::cerr << "Registered roles:\n";
    std::list< std::string > role_names = SoccerRole::creators().list();
    for ( std::list< std::string >::iterator it = role_names.begin();
          it != role_names.end();
          ++it )
    {
        std::cerr << "  " << *it << '\n';
    }
    std::cerr << std::flush;
#else
    std::cerr << "Supported formations:\n";
    for ( FormationFactory::const_iterator it = M_formation_factory.begin();
          it != M_formation_factory.end();
          ++it )
    {
        std::cerr << "  " << it->first << '\n';
    }
    std::cerr << "Supported roles:\n";
    for ( RoleFactory::const_iterator it = M_role_factory.begin();
          it != M_role_factory.end();
          ++it )
    {
        std::cerr << "  " << it->first << '\n';
    }
    std::cerr << std::flush;
#endif
#endif

    std::string configpath = config_dir;
    if ( ! configpath.empty()
         && configpath[ configpath.length() - 1 ] != '/' )
    {
        configpath += '/';
    }

    // before kick off
    M_before_kick_off_formation = readFormation( configpath + BEFORE_KICK_OFF_CONF );
    if ( ! M_before_kick_off_formation )
    {
        std::cerr << "Failed to read before_kick_off formation" << std::endl;
        return false;
    }

    ///////////////////////////////////////////////////////////
    M_normal_formation = readFormation( configpath + NORMAL_FORMATION_CONF );
    if ( ! M_normal_formation )
    {
        std::cerr << "Failed to read normal formation" << std::endl;
        return false;
    }

    M_defense_formation = readFormation( configpath + DEFENSE_FORMATION_CONF );
    if ( ! M_defense_formation )
    {
        std::cerr << "Failed to read defense formation" << std::endl;
        return false;
    }

    M_offense_formation = readFormation( configpath + OFFENSE_FORMATION_CONF );
    if ( ! M_offense_formation )
    {
        std::cerr << "Failed to read offense formation" << std::endl;
        return false;
    }

    M_goalie_formation = readFormation( configpath + GOALIE_FORMATION_CONF );
    if ( ! M_goalie_formation )
    {
        std::cerr << "Failed to read goalie formation" << std::endl;
        //return false;
    }

    M_goal_kick_opp_formation = readFormation( configpath + GOAL_KICK_OPP_FORMATION_CONF );
    if ( ! M_goal_kick_opp_formation )
    {
        return false;
    }

    M_goal_kick_our_formation = readFormation( configpath + GOAL_KICK_OUR_FORMATION_CONF );
    if ( ! M_goal_kick_our_formation )
    {
        return false;
    }

    M_goalie_catch_opp_formation = readFormation( configpath + GOALIE_CATCH_OPP_FORMATION_CONF );
    if ( ! M_goalie_catch_opp_formation )
    {
        return false;
    }

    M_goalie_catch_our_formation = readFormation( configpath + GOALIE_CATCH_OUR_FORMATION_CONF );
    if ( ! M_goalie_catch_our_formation )
    {
        return false;
    }

    M_kickin_our_formation = readFormation( configpath + KICKIN_OUR_FORMATION_CONF );
    if ( ! M_kickin_our_formation )
    {
        std::cerr << "Failed to read kickin our formation" << std::endl;
        return false;
    }

    M_setplay_opp_formation = readFormation( configpath + SETPLAY_OPP_FORMATION_CONF );
    if ( ! M_setplay_opp_formation )
    {
        std::cerr << "Failed to read setplay opp formation" << std::endl;
        return false;
    }

    M_setplay_our_formation = readFormation( configpath + SETPLAY_OUR_FORMATION_CONF );
    if ( ! M_setplay_our_formation )
    {
        std::cerr << "Failed to read setplay our formation" << std::endl;
        return false;
    }

    M_indirect_freekick_opp_formation = readFormation( configpath + INDIRECT_FREEKICK_OPP_FORMATION_CONF );
    if ( ! M_indirect_freekick_opp_formation )
    {
        std::cerr << "Failed to read indirect freekick opp formation" << std::endl;
        return false;
    }

    M_indirect_freekick_our_formation = readFormation( configpath + INDIRECT_FREEKICK_OUR_FORMATION_CONF );
    if ( ! M_indirect_freekick_our_formation )
    {
        std::cerr << "Failed to read indirect freekick our formation" << std::endl;
        return false;
    }

    s_initialized = true;
    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
rcsc::Formation::Ptr
Strategy::readFormation( const std::string & filepath )
{
    rcsc::Formation::Ptr f;

    std::ifstream fin( filepath.c_str() );
    if ( ! fin.is_open() )
    {
        std::cerr << __FILE__ << ':' << __LINE__ << ':'
                  << " ***ERROR*** failed to open file [" << filepath << "]"
                  << std::endl;
        return f;
    }

    std::string temp, type;
    fin >> temp >> type; // read training method type name
    fin.seekg( 0 );

    f = createFormation( type );

    if ( ! f )
    {
        std::cerr << __FILE__ << ':' << __LINE__ << ':'
                  << " ***ERROR*** failed to create formation [" << filepath << "]"
                  << std::endl;
        return f;
    }

    //
    // read data from file
    //
    if ( ! f->read( fin ) )
    {
        std::cerr << __FILE__ << ':' << __LINE__ << ':'
                  << " ***ERROR*** failed to read formation [" << filepath << "]"
                  << std::endl;
        f.reset();
        return f;
    }

    //
    // check role names
    //
    for ( int unum = 1; unum <= 11; ++unum )
    {
#ifdef USE_GENERIC_FACTORY
        SoccerRole::Ptr role = SoccerRole::create( f->getRoleName( unum ) );
        if ( ! role )
        {
            std::cerr << __FILE__ << ':' << __LINE__ << ':'
                      << " ***ERROR*** Unsupported role name ["
                      << f->getRoleName( unum ) << "] is appered in ["
                      << filepath << "]" << std::endl;
            f.reset();
            return f;
        }
#else
        if ( M_role_factory.find( f->getRoleName( unum ) ) == M_role_factory.end() )
        {
            std::cerr << __FILE__ << ':' << __LINE__ << ':'
                      << " ***ERROR*** Unsupported role name ["
                      << f->getRoleName( unum ) << "] is appered in ["
                      << filepath << "]" << std::endl;
            f.reset();
            return f;
        }
#endif
    }

    return f;
}

/*-------------------------------------------------------------------*/
/*!

*/
rcsc::Formation::Ptr
Strategy::createFormation( const std::string & type_name ) const
{
    rcsc::Formation::Ptr f;

#ifdef USE_GENERIC_FACTORY
    f = rcsc::Formation::create( type_name );
#else
    FormationFactory::const_iterator creator = M_formation_factory.find( type_name );
    if ( creator == M_formation_factory.end() )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** unsupported formation type ["
                  << type_name << "]"
                  << std::endl;
        return f;
    }
    f = creator->second();
#endif

    if ( ! f )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** unsupported formation type ["
                  << type_name << "]"
                  << std::endl;
    }

    return f;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
Strategy::update( const rcsc::WorldModel & wm )
{
    static rcsc::GameTime s_update_time( -1, 0 );

    if ( s_update_time == wm.time() )
    {
        return;
    }
    s_update_time = wm.time();

    updateSituation( wm );
    updatePosition( wm );
    analyzeOpponentStrategy( wm );

    M_mark_table->update( wm );

    M_field_evaluator = createFieldEvaluatorImpl( wm );
    M_action_generator = createActionGeneratorImpl( wm );
}

/*-------------------------------------------------------------------*/
/*!

*/
SoccerRole::Ptr
Strategy::createRole( const int number,
                      const rcsc::WorldModel & wm ) const
{
    SoccerRole::Ptr role;

    if ( number < 1 || 11 < number )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** Invalid player number " << number
                  << std::endl;
        return role;
    }

    rcsc::Formation::Ptr f = getFormation( wm );
    if ( ! f )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** faled to create role. Null formation" << std::endl;
        return role;
    }

    const std::string role_name = f->getRoleName( number );

#ifdef USE_GENERIC_FACTORY
    role = SoccerRole::create( role_name );
#else
    RoleFactory::const_iterator factory = M_role_factory.find( role_name );
    if ( factory != M_role_factory.end() )
    {
        role = factory->second();
    }
#endif

    if ( ! role )
    {
        std::cerr << __FILE__ << ": " << __LINE__
                  << " ***ERROR*** unsupported role name ["
                  << role_name << "]"
                  << std::endl;
    }
    return role;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
Strategy::updateSituation( const rcsc::WorldModel & wm )
{
    M_current_situation = Normal_Situation;

    if ( wm.gameMode().type() != rcsc::GameMode::PlayOn )
    {
        if ( wm.gameMode().isPenaltyKickMode() )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": Situation PenaltyKick" );
            M_current_situation = PenaltyKick_Situation;
        }
        else if ( wm.gameMode().isPenaltyKickMode() )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": Situation OurSetPlay" );
            M_current_situation = OurSetPlay_Situation;
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": Situation OppSetPlay" );
            M_current_situation = OppSetPlay_Situation;
        }
        return;
    }

    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();
    int our_min = std::min( self_min, mate_min );

    if ( opp_min <= our_min - 2 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": Situation Defense" );
        M_current_situation = Defense_Situation;
        return;
    }

    if ( our_min <= opp_min - 2 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": Situation Offense" );
        M_current_situation = Offense_Situation;
        return;
    }

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        __FILE__": Situation Normal" );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
Strategy::updatePosition( const rcsc::WorldModel & wm )
{
    static rcsc::GameTime s_update_time( 0, 0 );
    if ( s_update_time == wm.time() )
    {
        return;
    }
    s_update_time = wm.time();

    rcsc::Formation::Ptr f = getFormation( wm );
    if ( ! f )
    {
        std::cerr << wm.time()
                  << " ***ERROR*** could not get the current formation" << std::endl;
        return;
    }

    int ball_step = 0;
    if ( wm.gameMode().type() == rcsc::GameMode::PlayOn
         || wm.gameMode().type() == rcsc::GameMode::GoalKick_ )
    {
        ball_step = std::min( 1000, wm.interceptTable()->teammateReachCycle() );
        ball_step = std::min( ball_step, wm.interceptTable()->opponentReachCycle() );
        ball_step = std::min( ball_step, wm.interceptTable()->selfReachCycle() );
    }

    rcsc::Vector2D ball_pos = wm.ball().inertiaPoint( ball_step );

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        __FILE__": HOME POSITION: ball pos=(%.1f %.1f) step=%d",
                        ball_pos.x, ball_pos.y,
                        ball_step );

    M_positions.clear();
    f->getPositions( ball_pos, M_positions );

    if ( M_positions.size() != 11 )
    {
        M_positions.resize( 11 );
    }

    for ( size_t i = 0; i < 11; ++i )
    {
        if ( ! M_positions[i].isValid() ) M_positions[i].assign( 0.0, 0.0 );
    }


    if ( rcsc::ServerParam::i().useOffside() )
    {
        double max_x = wm.offsideLineX();
        int mate_step = wm.interceptTable()->teammateReachCycle();
        if ( mate_step < 50 )
        {
            rcsc::Vector2D trap_pos = wm.ball().inertiaPoint( mate_step );
            if ( trap_pos.x > max_x ) max_x = trap_pos.x;
        }

        for ( int unum = 1; unum <= 11; ++unum )
        {
            if ( M_positions[unum-1].x > max_x - 1.0 )
            {
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "____ %d offside. home_pos_x %.2f -> %.2f",
                                    unum,
                                    M_positions[unum-1].x, max_x - 1.0 );
                M_positions[unum-1].x = max_x - 1.0;
            }
        }
    }

    M_position_types.clear();
    for ( int unum = 1; unum <= 11; ++unum )
    {
        PositionType type = Position_Center;
        if ( f->isSideType( unum ) )
        {
            type = Position_Left;
        }
        else if ( f->isSymmetryType( unum ) )
        {
            type = Position_Right;
        }

        M_position_types.push_back( type );

        if ( wm.time().stopped() < 1 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "__ %d home pos (%.2f %.2f) type=%d",
                                unum,
                                M_positions[unum-1].x, M_positions[unum-1].y,
                                type );
            rcsc::dlog.addCircle( rcsc::Logger::TEAM,
                                  M_positions[unum-1], 0.5,
                                  "#000000" );
        }
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
Strategy::analyzeOpponentStrategy( const rcsc::WorldModel & wm )
{
    if ( M_opponent_defense_strategy == No_Strategy
         && ! wm.opponentTeamName().empty() )
    {
        M_opponent_offense_strategy = Normal_Strategy;
        M_opponent_defense_strategy = Normal_Strategy;
#if 0
        if ( wm.opponentTeamName().compare( 0, std::strlen( "Brainstormers" ),
                                            "Brainstormers" ) == 0
             || wm.opponentTeamName().compare( 0, std::strlen( "BS0" ),
                                               "BS0" ) == 0
             )
        {
            M_opponent_offense_strategy = FastPassShoot_Strategy;
            M_opponent_defense_strategy = ManMark_Strategy;
        }
        else if ( wm.opponentTeamName() == "AmoyNQ"
                  || wm.opponentTeamName() == "AmoiensisNQ" )
        {
            M_opponent_defense_strategy = ManMark_Strategy;
        }
#endif
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
PositionType
Strategy::getPositionType( const int number ) const
{
    if ( number < 1 || 11 < number )
    {
        std::cerr << __FILE__ << ' ' << __LINE__
                  << ": Illegal number : " << number
                  << std::endl;
        return Position_Center;
    }
    return M_position_types[number - 1];
}

/*-------------------------------------------------------------------*/
/*!

*/
rcsc::Vector2D
Strategy::getPosition( const int number ) const
{
    if ( number < 1 || 11 < number )
    {
        std::cerr << __FILE__ << ' ' << __LINE__
                  << ": Illegal number : " << number
                  << std::endl;
        return rcsc::Vector2D::INVALIDATED;
    }
    return M_positions[number - 1];
}

/*-------------------------------------------------------------------*/
/*!

*/
rcsc::Formation::Ptr
Strategy::getFormation( const rcsc::WorldModel & wm ) const
{
    if ( wm.gameMode().type() == rcsc::GameMode::PlayOn )
    {
        if ( wm.self().goalie()
             && M_goalie_formation )
        {
            return M_goalie_formation;
        }

        switch ( M_current_situation ) {
        case Defense_Situation:
            return M_defense_formation;
        case Offense_Situation:
            return M_offense_formation;
        default:
            break;
        }
        return M_normal_formation;
    }

    if ( wm.gameMode().type() == rcsc::GameMode::KickIn_
         || wm.gameMode().type() == rcsc::GameMode::CornerKick_ )
    {
        if ( wm.ourSide() == wm.gameMode().side() )
        {
            // our kick-in or corner-kick
            return M_kickin_our_formation;
        }
        else
        {
            return M_setplay_opp_formation;
        }
    }

    if ( ( wm.gameMode().type() == rcsc::GameMode::BackPass_
           && wm.gameMode().side() == wm.theirSide() )
         || ( wm.gameMode().type() == rcsc::GameMode::IndFreeKick_
              && wm.gameMode().side() == wm.ourSide() ) )
    {
        return M_indirect_freekick_our_formation;
    }

    if ( ( wm.gameMode().type() == rcsc::GameMode::BackPass_
           && wm.gameMode().side() == wm.ourSide() )
         || ( wm.gameMode().type() == rcsc::GameMode::IndFreeKick_
              && wm.gameMode().side() == wm.theirSide() ) )
    {
        return M_indirect_freekick_opp_formation;
    }

    if ( wm.gameMode().type() == rcsc::GameMode::GoalKick_ )
    {
        if ( wm.gameMode().side() == wm.ourSide() )
        {
            return M_goal_kick_our_formation;
        }
        else
        {
            return M_goal_kick_opp_formation;
        }
    }

    if ( wm.gameMode().type() == rcsc::GameMode::GoalieCatch_ )
    {
        if ( wm.gameMode().side() == wm.ourSide() )
        {
            return M_goalie_catch_our_formation;
        }
        else
        {
            return M_goalie_catch_opp_formation;
        }
    }

    if ( wm.gameMode().type() == rcsc::GameMode::BeforeKickOff
         || wm.gameMode().type() == rcsc::GameMode::AfterGoal_ )
    {
        return M_before_kick_off_formation;
    }

    if ( wm.gameMode().isOurSetPlay( wm.ourSide() ) )
    {
        return M_setplay_our_formation;
    }

    if ( wm.gameMode().type() != rcsc::GameMode::PlayOn )
    {
        return M_setplay_opp_formation;
    }

    if ( wm.self().goalie()
         && M_goalie_formation )
    {
        return M_goalie_formation;
    }


    switch ( M_current_situation ) {
    case Defense_Situation:
        return M_defense_formation;
    case Offense_Situation:
        return M_offense_formation;
    default:
        break;
    }

    return M_normal_formation;
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Strategy::isMarkerType( const int unum ) const
{
    if ( unum == 2
         || unum == 3
         || unum == 4
         || unum == 5 )
    {
        return true;
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

*/
Strategy::BallArea
Strategy::get_ball_area( const rcsc::WorldModel & wm )
{
    int ball_step = 1000;
    ball_step = std::min( ball_step, wm.interceptTable()->teammateReachCycle() );
    ball_step = std::min( ball_step, wm.interceptTable()->opponentReachCycle() );
    ball_step = std::min( ball_step, wm.interceptTable()->selfReachCycle() );

    return get_ball_area( wm.ball().inertiaPoint( ball_step ) );
}

/*-------------------------------------------------------------------*/
/*!

*/
Strategy::BallArea
Strategy::get_ball_area( const rcsc::Vector2D & ball_pos )
{
    rcsc::dlog.addLine( rcsc::Logger::TEAM,
                        52.5, -17.0, -52.5, -17.0,
                        "#999999" );
    rcsc::dlog.addLine( rcsc::Logger::TEAM,
                        52.5, 17.0, -52.5, 17.0,
                        "#999999" );
    rcsc::dlog.addLine( rcsc::Logger::TEAM,
                        36.0, -34.0, 36.0, 34.0,
                        "#999999" );
    rcsc::dlog.addLine( rcsc::Logger::TEAM,
                        -1.0, -34.0, -1.0, 34.0,
                        "#999999" );
    rcsc::dlog.addLine( rcsc::Logger::TEAM,
                        -30.0, -17.0, -30.0, 17.0,
                        "#999999" );
    rcsc::dlog.addLine( rcsc::Logger::TEAM,
                        //-36.5, -34.0, -36.5, 34.0,
                        -35.5, -34.0, -35.5, 34.0,
                        "#999999" );

    if ( ball_pos.x > 36.0 )
    {
        if ( ball_pos.absY() > 17.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_ball_area: Cross" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                36.0, -34.0, 52.5 - 36.0, 34.0 - 17.0,
                                "#00ff00" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                36.0, 17.0, 52.5 - 36.0, 34.0 - 17.0,
                                "#00ff00" );
            return BA_Cross;
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_ball_area: ShootChance" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                36.0, -17.0, 52.5 - 36.0, 34.0,
                                "#00ff00" );
            return BA_ShootChance;
        }
    }
    else if ( ball_pos.x > -1.0 )
    {
        if ( ball_pos.absY() > 17.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_ball_area: DribbleAttack" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                -1.0, -34.0, 36.0 + 1.0, 34.0 - 17.0,
                                "#00ff00" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                -1.0, 17.0, 36.0 + 1.0, 34.0 - 17.0,
                                "#00ff00" );
            return BA_DribbleAttack;
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_ball_area: OffMidField" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                -1.0, -17.0, 36.0 + 1.0, 34.0,
                                "#00ff00" );
            return BA_OffMidField;
        }
    }
    else if ( ball_pos.x > -30.0 )
    {
        if ( ball_pos.absY() > 17.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_ball_area: DribbleBlock" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                -30.0, -34.0, -1.0 + 30.0, 34.0 - 17.0,
                                "#00ff00" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                -30.0, 17.0, -1.0 + 30.0, 34.0 - 17.0,
                                "#00ff00" );
            return BA_DribbleBlock;
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_ball_area: DefMidField" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                -30.0, -17.0, -1.0 + 30.0, 34.0,
                                "#00ff00" );
            return BA_DefMidField;
        }
    }
    // 2009-06-17 akiyama: -36.5 -> -35.5
    //else if ( ball_pos.x > -36.5 )
    else if ( ball_pos.x > -35.5 )
    {
        if ( ball_pos.absY() > 17.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_ball_area: CrossBlock" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                //-36.5, -34.0, 36.5 - 30.0, 34.0 - 17.0,
                                -35.5, -34.0, 35.5 - 30.0, 34.0 - 17.0,
                                "#00ff00" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                -35.5, 17.0, 35.5 - 30.0, 34.0 - 17.0,
                                "#00ff00" );
            return BA_CrossBlock;
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_ball_area: Stopper" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                //-36.5, -17.0, 36.5 - 30.0, 34.0,
                                -35.5, -17.0, 35.5 - 30.0, 34.0,
                                "#00ff00" );
            // 2009-06-17 akiyama: Stopper -> DefMidField
            //return BA_Stopper;
            return BA_DefMidField;
        }
    }
    else
    {
        if ( ball_pos.absY() > 17.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_ball_area: CrossBlock" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                -52.5, -34.0, 52.5 - 36.5, 34.0 - 17.0,
                                "#00ff00" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                -52.5, 17.0, 52.5 - 36.5, 34.0 - 17.0,
                                "#00ff00" );
            return BA_CrossBlock;
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_ball_area: Danger" );
            rcsc::dlog.addRect( rcsc::Logger::TEAM,
                                -52.5, -17.0, 52.5 - 36.5, 34.0,
                                "#00ff00" );
            return BA_Danger;
        }
    }

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        __FILE__": get_ball_area: unknown area" );
    return BA_None;
}

/*-------------------------------------------------------------------*/
/*!

*/
double
Strategy::get_normal_dash_power( const rcsc::WorldModel & wm )
{
    static bool s_recover_mode = false;

    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();

    // check recover
    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        s_recover_mode = false;
    }

    /*--------------------------------------------------------*/
    double dash_power = rcsc::ServerParam::i().maxDashPower();
    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();

    if ( wm.defenseLineX() > wm.self().pos().x
         && wm.ball().pos().x < wm.defenseLineX() + 20.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": (get_normal_dash_power) correct DF line. keep max power" );
        // keep max power
        dash_power = rcsc::ServerParam::i().maxDashPower();
    }
    else if ( s_recover_mode )
    {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": (get_normal_dash_power) recovering" );
    }
    // exist kickable teammate
    else if ( wm.existKickableTeammate()
              && wm.ball().distFromSelf() < 20.0 )
    {
        dash_power = std::min( my_inc * 1.1,
                               rcsc::ServerParam::i().maxDashPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": (get_normal_dash_power) exist kickable teammate. dash_power=%.1f",
                            dash_power );
    }
    // in offside area
    else if ( wm.self().pos().x > wm.offsideLineX() )
    {
        dash_power = rcsc::ServerParam::i().maxDashPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": in offside area. dash_power=%.1f",
                            dash_power );
    }
    else if ( wm.ball().pos().x > 25.0
              && wm.ball().pos().x > wm.self().pos().x + 10.0
              && self_min < opp_min - 6
              && mate_min < opp_min - 6 )
    {
        dash_power = rcsc::bound( rcsc::ServerParam::i().maxDashPower() * 0.1,
                                  my_inc * 0.5,
                                  rcsc::ServerParam::i().maxDashPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": (get_normal_dash_power) opponent ball dash_power=%.1f",
                            dash_power );
    }
    // normal
    else
    {
        dash_power = std::min( my_inc * 1.7,
                               rcsc::ServerParam::i().maxDashPower() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": (get_normal_dash_power) normal mode dash_power=%.1f",
                            dash_power );
    }

    return dash_power;
}

/*-------------------------------------------------------------------*/
/*!

*/
double
Strategy::get_defender_dash_power( const rcsc::WorldModel & wm,
                                   const rcsc::Vector2D & home_pos )
{
    static bool S_recover_mode = false;

    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        S_recover_mode = true;
    }
    else if ( wm.self().stamina()
              > rcsc::ServerParam::i().staminaMax() * 0.85 )
    {
        S_recover_mode = false;
    }

    const double ball_xdiff
        //= wm.ball().pos().x - home_pos.x;
        = wm.ball().pos().x - wm.self().pos().x;

    const rcsc::PlayerType & mytype = wm.self().playerType();
    const double my_inc = mytype.staminaIncMax() * wm.self().recovery();

    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();

    double dash_power;
    if ( S_recover_mode )
    {
#if 1
        // 2009-06-29
        if ( opp_min <= mate_min + 1
             && wm.self().pos().x < home_pos.x - 10.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": (get_defender)dash_power) recover mode, but opponent ball. max power" );
            dash_power = rcsc::ServerParam::i().maxDashPower();
        }
        else
#endif
        if ( wm.defenseLineX() > wm.self().pos().x )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_dash_power. correct DF line & recover" );
            dash_power = my_inc;
        }
        else if ( ball_xdiff < 5.0 )
        {
            dash_power = rcsc::ServerParam::i().maxDashPower();
        }
        else if ( ball_xdiff < 10.0 )
        {
            dash_power = rcsc::ServerParam::i().maxDashPower();
            dash_power *= 0.7;
            //dash_power
            //    = mytype.getDashPowerToKeepSpeed( 0.7, wm.self().effort() );
        }
        else if ( ball_xdiff < 20.0 )
        {
            dash_power = std::max( 0.0, my_inc - 10.0 );
        }
        else // >= 20.0
        {
            dash_power = std::max( 0.0, my_inc - 20.0 );
        }

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": get_dash_power. recover mode dash_power= %.1f",
                            dash_power );

        return dash_power;
    }

    // normal case

#if 1
    // added 2006/06/11 03:34
    if ( wm.ball().pos().x > 0.0
         && wm.self().pos().x < home_pos.x
         && wm.self().pos().x > wm.defenseLineX() - 0.05 ) // 20080712
    {
        double power_for_max_speed = mytype.getDashPowerToKeepMaxSpeed( rcsc::ServerParam::i(),
                                                                        wm.self().effort() );
        double defense_dash_dist = wm.self().pos().dist( rcsc::Vector2D( -48.0, 0.0 ) );
        int cycles_to_reach = mytype.cyclesToReachDistance( defense_dash_dist );
        int available_dash_cycles = mytype.getMaxDashCyclesSavingStamina( rcsc::ServerParam::i(),
                                                                          power_for_max_speed,
                                                                          wm.self().stamina(),
                                                                          wm.self().recovery() );
        if ( available_dash_cycles < cycles_to_reach )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_dash_power. keep stamina for defense back dash,"
                                " power_for_max=%.1f"
                                " dash_dist=%.1f, reach_cycle=%d, dashable_cycle=%d",
                                power_for_max_speed,
                                defense_dash_dist, cycles_to_reach, available_dash_cycles );
            dash_power = std::max( 0.0, my_inc - 20.0 );
            return dash_power;
        }
    }
#endif

#if 1
    if ( wm.defenseLineX() < wm.self().pos().x - 2.0
         && wm.self().pos().x < home_pos.x )
    {
        dash_power = my_inc * 0.5;
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": get_dash_power. over defense line. power=%.1f",
                            dash_power );
    }
    else
#endif
    if ( wm.self().pos().x < -30.0
         && wm.defenseLineX() > wm.self().pos().x )
    {
        dash_power = rcsc::ServerParam::i().maxDashPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": get_dash_power. correct dash power for the defense line. power=%.1f",
                            dash_power );
    }
    else if ( home_pos.x < wm.self().pos().x )
    {
        dash_power = rcsc::ServerParam::i().maxDashPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": get_dash_power. max power to go to the behind home position. power=%.1f",
                            dash_power );
    }
    else if ( ball_xdiff > 20.0 )
    {
        dash_power = my_inc;
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": get_dash_power. correct dash power to save stamina(1). power=%.1f",
                            dash_power );
    }
    else if ( ball_xdiff > 10.0 )
    {
        dash_power = rcsc::ServerParam::i().maxDashPower();
        dash_power *= 0.6;
        //dash_power = mytype.getDashPowerToKeepSpeed( 0.6, wm.self().effort() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": get_dash_power. correct dash power to save stamina(2). power=%.1f",
                            dash_power );
    }
    else if ( ball_xdiff > 5.0 )
    {
        dash_power = rcsc::ServerParam::i().maxDashPower();
        dash_power *= 0.85;
        //dash_power = mytype.getDashPowerToKeepSpeed( 0.85, wm.self().effort() );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": get_dash_power. correct dash power to save stamina(3). power=%.1f",
                            dash_power );
    }
    else
    {
        dash_power = rcsc::ServerParam::i().maxDashPower();
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": get_dash_power. max power. power=%.1f",
                            dash_power );
    }

    return dash_power;
}


/*-------------------------------------------------------------------*/
/*!

*/
double
Strategy::get_defense_line_x( const rcsc::WorldModel & wm,
                              const rcsc::Vector2D & home_pos )
{
#ifdef USE_DEFENSE_LINE
    double x = home_pos.x;

    if ( wm.ball().pos().x > -30.0
         && wm.self().pos().x > home_pos.x
         && wm.ball().pos().x > wm.self().pos().x + 3.0 )
    {
        if ( wm.ball().pos().x > 0.0 ) x = -3.5;
        else if ( wm.ball().pos().x > -8.0 ) x = -11.5;
        else if ( wm.ball().pos().x > -16.0 ) x = -19.5;
        else if ( wm.ball().pos().x > -24.0 ) x = -27.5;
        else if ( wm.ball().pos().x > -30.0 ) x = -33.5;
        else x = home_pos.x;

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__": get_defense_line_x. x= %.1f",
                            x );

        if ( x > 0.0 ) x = 0.0;
        if ( x > home_pos.x + 5.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_defense_line_x. adjust to the home_pos(+) %.1f",
                                home_pos.x );
            x = home_pos.x + 5.0;
        }

        if ( x < home_pos.x - 2.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_defense_line_x. adjust to the home_pos(-) %.1f",
                                home_pos.x );
            x = home_pos.x - 2.0;
        }

        if ( wm.defenseLineX() - 1.0 > x )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                __FILE__": get_defense_line_x. adjust to the defense line %.1f",
                                wm.defenseLineX() );
            x = wm.defenseLineX() - 1.0;
        }
    }

    return x;
#else
    (void)wm;
    return home_pos.x;
#endif
}

/*-------------------------------------------------------------------*/
/*!

*/
double
Strategy::get_set_play_dash_power( const rcsc::WorldModel & wm )
{
    if ( wm.gameMode().type() == rcsc::GameMode::PenaltySetup_ )
    {
        return wm.self().getSafetyDashPower( rcsc::ServerParam::i().maxDashPower() );
    }

    double rate = 1.0;
    if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.8 )
    {
        rate = 1.5
            * wm.self().stamina()
            / rcsc::ServerParam::i().staminaMax();
    }
    else
    {
        rate = 0.9
            * ( wm.self().stamina()
                - rcsc::ServerParam::i().recoverDecThrValue() )
            / rcsc::ServerParam::i().staminaMax();
        rate = std::max( 0.0, rate );
    }

    return ( wm.self().playerType().staminaIncMax()
             * wm.self().recovery()
             * rate );
}

/*-------------------------------------------------------------------*/
/*!

*/
Vector2D
Strategy::get_block_opponent_trap_point( const WorldModel & wm )
{
    int opp_min = wm.interceptTable()->opponentReachCycle();
    Vector2D opp_trap_pos = wm.ball().inertiaPoint( opp_min );

#ifdef DEBUG_PRINT_BLODK
    dlog.addText( Logger::BLOCK,
                  __FILE__": opponent trap pos (%.3f %.3f)",
                  opp_trap_pos.x, opp_trap_pos.y );
#endif

    const PlayerObject * opp = wm.interceptTable()->fastestOpponent();
    if ( opp )
    {
        Vector2D tmp_trap_pos = opp_trap_pos;
        if ( opp_min == 0 )
        {
            if ( opp->bodyCount() <= 1 )
            {
                tmp_trap_pos = opp->inertiaPoint( 2 )
                    + Vector2D::polar2vector( 0.4, opp->body() );
#ifdef DEBUG_PRINT_BLODK
                dlog.addText( Logger::BLOCK,
                              __FILE__": adjust opponent trap point to the body direction(1) (%.3f %.3f)",
                              tmp_trap_pos.x, tmp_trap_pos.y );
#endif
            }
            else
            {
                tmp_trap_pos = opp->inertiaPoint( 10 );
#ifdef DEBUG_PRINT_BLODK
                dlog.addText( Logger::BLOCK,
                              __FILE__": adjust opponent trap point to the inertia point (%.3f %.3f)",
                              tmp_trap_pos.x, tmp_trap_pos.y );
#endif
            }
        }
        else if ( opp_min <= 1
                  && opp->distFromSelf() < ServerParam::i().visibleDistance()
                  && opp->bodyCount() <= 1 )
        {
            tmp_trap_pos += Vector2D::polar2vector( 0.4, opp->body() );
#ifdef DEBUG_PRINT_BLOCK
            dlog.addText( Logger::BLOCK,
                          __FILE__": adjust opponent trap point to the body direction(2) (%.3f %.3f)",
                          tmp_trap_pos.x, tmp_trap_pos.y );
#endif
        }

        Vector2D center( -44.0, 0.0 );
        if ( tmp_trap_pos.dist2( center ) < opp_trap_pos.dist2( center ) )
        {
#ifdef DEBUG_PRINT_BLOCK
            dlog.addText( Logger::BLOCK,
                          __FILE__": change opponent trap pos (%.3f %.3f) -> (%.3f %.3f)",
                          opp_trap_pos.x, opp_trap_pos.y,
                          tmp_trap_pos.x, tmp_trap_pos.y );
#endif
            opp_trap_pos = tmp_trap_pos;
        }
    }

    return opp_trap_pos;
}

/*-------------------------------------------------------------------*/
/*!

*/
Vector2D
Strategy::get_block_center_point( const WorldModel & wm )
{
    Vector2D opp_trap_pos = get_block_opponent_trap_point( wm );

    //
    // searth the best point
    //
    Vector2D center_pos( -44.0, 0.0 );
    if ( opp_trap_pos.x < -38.0
         && opp_trap_pos.absY() < 7.0 )
    {
        center_pos.x = -52.5;
#if 0
        center_pos.y = opp_trap_pos.y * 0.6;
#else
        // 2009-06-17
        center_pos.y = -2.0 * sign( wm.self().pos().y );
#endif
        dlog.addText( Logger::BLOCK,
                      __FILE__": center (1) (%.1f %.1f)",
                      center_pos.x, center_pos.y );
    }
    else if ( opp_trap_pos.x < -38.0
              && opp_trap_pos.absY() < 9.0 )
    {
        center_pos.x = std::min( -49.0, opp_trap_pos.x - 0.2 );
        //center_pos.y = opp_trap_pos.y * 0.6;
        Vector2D goal_pos( - ServerParam::i().pitchHalfLength(),
                           ServerParam::i().goalHalfWidth() * 0.5 );
        if ( opp_trap_pos.y > 0.0 )
        {
            goal_pos.y *= -1.0;
        }

        Line2D opp_line( opp_trap_pos, goal_pos );
        center_pos.y = opp_line.getY( center_pos.x );
        if ( center_pos.y == Line2D::ERROR_VALUE )
        {
            center_pos.y = opp_trap_pos.y * 0.6;
        }

        dlog.addText( Logger::BLOCK,
                      __FILE__": center (2) (%.1f %.1f)",
                      center_pos.x, center_pos.y );
    }
    else if ( opp_trap_pos.x < -38.0
              && opp_trap_pos.absY() < 12.0 )
    {
        //center_pos.x = -50.0;
        center_pos.x = std::min( -49.0, opp_trap_pos.x - 0.2 );
        //center_pos.y = 2.5 * sign( opp_trap_pos.y );
        //center_pos.y = 6.5 * sign( opp_trap_pos.y );
        center_pos.y = opp_trap_pos.y * 0.8;
        dlog.addText( Logger::BLOCK,
                      __FILE__": center (3) (%.1f %.1f)",
                      center_pos.x, center_pos.y );
    }
    else if ( opp_trap_pos.x < -30.0
              && 2.0 < opp_trap_pos.absY()
              && opp_trap_pos.absY() < 8.0 )
    {
        center_pos.x = -50.0;
        center_pos.y = opp_trap_pos.y * 0.9;
        dlog.addText( Logger::BLOCK,
                      __FILE__": center (4) (%.1f %.1f)",
                      center_pos.x, center_pos.y );
    }
    else if ( opp_trap_pos.absY() > 25.0 )
    {
        center_pos.x = -44.0;
#if 0
        center_pos.y = 20.0 * sign( opp_trap_pos.y );
#else
        // 2009-06-17
        center_pos.y = 16.0 * sign( opp_trap_pos.y );
#endif
        dlog.addText( Logger::BLOCK,
                      __FILE__": center (5) (%.1f %.1f)",
                      center_pos.x, center_pos.y );
    }
    else if ( opp_trap_pos.absY() > 20.0 )
    {
        center_pos.x = -44.0;
        //center_pos.y = 15.0 * sign( opp_trap_pos.y );
        center_pos.y = 5.0 * sign( opp_trap_pos.y );
        dlog.addText( Logger::BLOCK,
                      __FILE__": center (6) (%.1f %.1f)",
                      center_pos.x, center_pos.y );
    }
    else if ( opp_trap_pos.absY() > 15.0 )
    {
        center_pos.x = -44.0;
        //center_pos.y = 10.0 * sign( opp_trap_pos.y );
        center_pos.y = 5.0 * sign( opp_trap_pos.y );
        dlog.addText( Logger::BLOCK,
                      __FILE__": center (7) (%.1f %.1f)",
                      center_pos.x, center_pos.y );
    }
    else
    {
        dlog.addText( Logger::BLOCK,
                      __FILE__": center(default) (%.1f %.1f)",
                      center_pos.x, center_pos.y );
    }

    return center_pos;
}

/*-------------------------------------------------------------------*/
/*!

*/
boost::shared_ptr< const rcsc::FieldEvaluator >
Strategy::createFieldEvaluator() const
{
    return M_field_evaluator;
}

boost::shared_ptr< const rcsc::CooperativeActionCandidateGenerator >
Strategy::createActionGenerator() const
{
    return M_action_generator;
}

/*-------------------------------------------------------------------*/
/*!

*/
#include "default_field_evaluator.h"
#include "pass_test_field_evaluator.h"

boost::shared_ptr< const rcsc::FieldEvaluator >
Strategy::createFieldEvaluatorImpl( const rcsc::WorldModel & ) const
{
#if 1
    return boost::shared_ptr< const rcsc::FieldEvaluator >
           ( new DefaultFieldEvaluator() );
#else
    return boost::shared_ptr< const rcsc::FieldEvaluator >
           ( new PassTestFieldEvaluator() );
#endif
}

/*-------------------------------------------------------------------*/
/*!

*/
#include "actgen_recursive_level_filter.h"
#include "actgen_ball_pos_geometry_filter.h"
#include "actgen_direct_pass.h"
#include "actgen_lead_pass.h"
#include "actgen_voronoi_pass.h"
#include "actgen_through_pass2008.h"
#include "actgen_through_pass.h"
#include "actgen_through_pass_2.h"
#include "actgen_omnidir_dribble.h"
#include "actgen_dribble_table_omnidir_dribble.h"
#include "actgen_shoot.h"
#include "actgen_pass_table_pass.h"

boost::shared_ptr< const rcsc::CooperativeActionCandidateGenerator >
Strategy::createActionGeneratorImpl( const WorldModel & ) const
{
    rcsc::CompositeCooperativeActionCandidateGenerator *
        g = new rcsc::CompositeCooperativeActionCandidateGenerator();

#if 1
    //
    // pass table pass
    //
    g->addGenerator( new ActGen_MaxRecursiveLevelFilter
                     ( new ActGen_PassTablePass(), 1 ) );
#endif


    //
    // direct pass
    //
    g->addGenerator( new ActGen_RangeRecursiveLevelFilter
                     ( new ActGen_DirectPass(),
                       2, ActGen_RangeRecursiveLevelFilter::MAX ) );


#if 0
    //
    // lead pass
    //
    g->addGenerator( new ActGen_MaxRecursiveLevelFilter
                     ( new ActGen_LeadPass(), 1 ) );

    //
    // voronoi pass
    //
    g->addGenerator( new ActGen_MaxRecursiveLevelFilter
                     ( new ActGen_VoronoiPass(), 1 ) );
#endif


    //
    // dribble
    //
#if 0
    g->addGenerator( new ActGen_MaxRecursiveLevelFilter
                     ( new ActGen_OmnidirDribble(), 1 ) );
#else
    g->addGenerator( new ActGen_MaxRecursiveLevelFilter
                     ( new ActGen_DribbleTableOmnidirDribble(), 1 ) );
#endif

#if 0
    //
    // through pass
    //
    g->addGenerator( new ActGen_MaxRecursiveLevelFilter
                     ( new ActGen_ThroughPass(), 1 ) );
#endif

#if 0
    //
    // through pass 2
    //
    g->addGenerator( new ActGen_RangeRecursiveLevelFilter
                     ( new ActGen_ThroughPass2(),
                       2, 2 ) );
#endif

#if 1
    //
    // through pass
    //
    g->addGenerator( new ActGen_RangeRecursiveLevelFilter
                     ( new ActGen_ThroughPass2008(),
                       1, 2 ) );
#endif

    //
    // shoot
    //
    g->addGenerator( new ActGen_RangeRecursiveLevelFilter
                     ( new ActGen_Shoot(),
                       2, ActGen_RangeRecursiveLevelFilter::MAX ) );

    return boost::shared_ptr
           < const rcsc::CooperativeActionCandidateGenerator >( g );
}
